#ifndef FEMHOWTO_BASE_HH
#define FEMHOWTO_BASE_HH

#include <sstream>

#include <dune/common/misc.hh>
#include <dune/common/fvector.hh>

#include <dune/common/version.hh>

#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/misc/utility.hh>
#include <dune/fem/misc/l2error.hh>
#include <dune/fem/misc/femeoc.hh>
#include <dune/fem/misc/femtimer.hh>
#include <dune/fem/misc/gridwidth.hh>
#include <dune/fem/space/common/adaptmanager.hh>
#include <dune/fem/io/parameter.hh>
#include <dune/fem/io/file/dataoutput.hh>
#include <dune/fem/io/file/datawriter.hh>

inline void initializeFemEoc( const std::string &problemDescription ) 
{
  // output of error and eoc information
  std::string eocOutPath = Dune::Parameter::getValue<std::string>("femhowto.eocOutputPath",  /*@\label{base:param1}@*/
                                                                  std::string("."));
  std::string eocFile = eocOutPath + std::string("/eoc.tex");
  Dune::FemEoc::initialize(eocOutPath, "eoc", problemDescription); /*@\label{base:femeocInit}@*/
}

template< class HGridType > /*@LST0S@*/
Dune::GridPtr< HGridType > initialize ( const std::string &problemDescription )
{
  // ----- read in runtime parameters ------
  const std::string filekey  = Dune::IOInterface::defaultGridKey( HGridType::dimension );
  const std::string filename = Dune::Parameter::getValue< std::string >( filekey ); /*@\label{base:param0}@*/

  // initialize grid
  Dune::GridPtr< HGridType > gridptr(filename);
  gridptr->loadBalance();
  Dune::Parameter::appendDGF( filename );

  // initialize FemEoc 
  initializeFemEoc( problemDescription );

  // and refine the grid until the startLevel is reached
  const int startLevel = Dune::Parameter::getValue<int>("femhowto.startLevel", 0);
  for(int level=0; level < startLevel ; ++level)
    gridptr->globalRefine( 1 );
  
  return gridptr;
} /*@LST0E@*/

/////////////////////////////////////////////////////////////////////////////
//
//  compute algorithm 
//
/////////////////////////////////////////////////////////////////////////////
template <class Algorithm> /*@LST0S@*/
void compute(Algorithm& algorithm)
{
  typedef typename Algorithm::DiscreteFunctionType DiscreteFunctionType;
  typename Algorithm::DiscreteSpaceType& space = algorithm.space();
  typename Algorithm::GridPartType& gridPart = space.gridPart();
  typedef typename Algorithm::GridPartType::GridType HGridType;
  HGridType& grid = gridPart.grid();

  // solution function
  DiscreteFunctionType u("solution",space);

  // get some parameters
  const int eocSteps   = Dune::Parameter::getValue<int>("femhowto.eocSteps", 1);

  // Initialize the DataOutput that writes the solution on the harddisk in a
  // format readable by e.g. Paraview
  // in each loop for the eoc computation the results at
  // the final time is stored
  typedef Dune::tuple< DiscreteFunctionType* > IOTupleType;
  typedef Dune::DataWriter<HGridType, IOTupleType> DataOutputType;
  IOTupleType dataTup ( &u );
  DataOutputType dataOutput( grid, dataTup );

  const unsigned int femTimerId = Dune::FemTimer::addTo("timestep");
  for(int eocloop=0; eocloop < eocSteps; ++eocloop)
  {
    Dune::FemTimer :: start(femTimerId);

    // do one step
    algorithm(u); /*@\label{base:solveProblem}@*/

    double runTime = Dune::FemTimer::stop(femTimerId);

    Dune::FemTimer::printFile("./timer.out");
    Dune::FemTimer::reset(femTimerId);

    // Write solution to hd
    dataOutput.writeData(eocloop);

    // finalize algorithm (compute errors)
    algorithm.finalize(u); /*@\label{base:finalize}@*/

    // calculate grid width
    const double h = Dune::GridWidth::calcGridWidth(gridPart);

    if( Dune::Parameter :: verbose() )
      Dune::FemEoc::write(h,grid.size(0),runTime,0, std::cout);
    else
      Dune::FemEoc::write(h,grid.size(0),runTime,0);

    // Refine the grid for the next EOC Step. If the scheme uses adaptation,
    // the refinement level needs to be set in the algorithms' initialize method.
    if(eocloop < eocSteps-1)
    {
      Dune::GlobalRefine::apply(grid,Dune::DGFGridInfo<HGridType>::refineStepsForHalf()); /*@\label{base:globalRefine2}@*/
      grid.loadBalance();
    }
  } /***** END of EOC Loop *****/
  Dune::FemTimer::removeAll();
}/*@LST0E@*/
#endif
