#ifndef FEMHOWTO_BASEADAPTIVE_HH
#define FEMHOWTO_BASEADAPTIVE_HH

#include "base.hh"

#include <dune/fem/space/common/adaptmanager.hh>

// return true if adaptation is used 
template< class HGridType >
bool checkAdaptive(HGridType& grid)
{
  Dune::AdaptationMethod< HGridType > am( grid );
  return am.adaptive();
}

template <class Algorithm>
static void computeAdaptive(Algorithm & algorithm) 
{
  typedef typename Algorithm :: DiscreteFunctionType DiscreteFunctionType;
  typename Algorithm::DiscreteSpaceType& space = algorithm.space();
  typename Algorithm::GridPartType& gridPart = space.gridPart();
  typedef typename Algorithm::HGridType HGridType;
  HGridType& grid = gridPart.grid();

  // if adaptation is disabled fallback to normal compute 
  if ( ! checkAdaptive( grid )) 
  {
    std::cout << "Adaptivity disabled, fallback to EOC calculation! " << std::endl;
    compute(algorithm);
    return ;
  }

  // solution function
  DiscreteFunctionType u("solution",space);

  // get tolerance for adaptive problem 
  double tolerance = 1.0;  // default value 
  tolerance = Dune::Parameter::getValue("femhowto.tolerance", tolerance);

  // Initialize the DataOutput that writes the solution on the harddisk in a
  // format readable by e.g. Paraview in each adaption step
  typedef Dune::tuple< DiscreteFunctionType * > IOTupleType;
  typedef Dune::DataOutput< HGridType, IOTupleType > DataOutputType;
  IOTupleType dataTup ( &u );
  DataOutputType dataOutput( grid, dataTup );

  const unsigned int adaptStepTimer = Dune::FemTimer::addTo("totaltime (time of first step, max time of steps)", 2);
  Dune::FemTimer::start( adaptStepTimer );
  Dune::FemTimer::start( adaptStepTimer, 1 );
  algorithm(u);
  Dune::FemTimer::stop( adaptStepTimer, 1 );
  Dune::FemTimer::stop( adaptStepTimer, Dune::FemTimer::sum );

  algorithm.finalize(u);
  dataOutput.writeData(0);

  typedef typename Algorithm :: RestrictionProlongationType RestrictionProlongationType;
  typedef typename Algorithm :: AdaptationManagerType AdaptationManagerType;

  // create Restriction and Prolongation Operator 
  RestrictionProlongationType rpSolution( u );
  // create Adaptation Manager 
  AdaptationManagerType adaptManager( grid, rpSolution );

  int counter=1;
  while ( algorithm.estimateAndMark(u, tolerance) )
  {
    // adapt grid and restrict and prolong solution (see AdaptionManager)
    adaptManager.adapt();

    // re-calculate solution 
    Dune::FemTimer::start( adaptStepTimer );
    Dune::FemTimer::start( adaptStepTimer, 2 );
    algorithm(u);
    Dune::FemTimer::stop( adaptStepTimer, 2, Dune::FemTimer::max );
    Dune::FemTimer::stop( adaptStepTimer, Dune::FemTimer::sum );

    algorithm.finalize(u);
    // write data file for time and timestep (internal step number is increased)
    dataOutput.writeData(counter);
    ++counter;
  }
  Dune::FemTimer::printFile( "./timer.out" );
  Dune::FemTimer::removeAll();
}
#endif
