#ifndef FEMHOWTO_BASEEVOL_HH
#define FEMHOWTO_BASEEVOL_HH

#include "base.hh"

#include <dune/fem/io/file/datawriter.hh>
#include <dune/fem/space/common/adaptmanager.hh>

/////////////////////////////////////////////////////////////////////////////
//
//  EOC output parameter class 
//
/////////////////////////////////////////////////////////////////////////////
struct EocDataOutputParameters :   /*@LST1S@*/
       public Dune::LocalParameter<Dune::DataWriterParameters,EocDataOutputParameters> 
{
  std::string loop_;
  EocDataOutputParameters(int loop) 
  {
    std::stringstream ss;
    ss << "loop" << loop;
    loop_ = ss.str();
  }
  EocDataOutputParameters(const EocDataOutputParameters& other)
  : loop_(other.loop_) 
  {}
  std::string path() const { return loop_; }
};                              /*@LST1E@*/


/////////////////////////////////////////////////////////////////////////////
//
//  Base class for evolutionary problems 
//
/////////////////////////////////////////////////////////////////////////////
template <class TraitsImp> /*@LST0S@*/
class AlgorithmBase /*@LST0E@*/
{
  typedef TraitsImp Traits ;
public:
  // type of Grid
  typedef typename Traits :: GridType                  GridType;

  // choose a suitable GridView
  typedef typename Traits :: GridPartType              GridPartType;

  // the space type
  typedef typename Traits :: DiscreteSpaceType         DiscreteSpaceType;

  // the discrete function type
  typedef typename Traits :: DiscreteFunctionType      DiscreteFunctionType;

  // the restriction and prolongation projection 
  typedef typename Traits :: RestrictionProlongationType RestrictionProlongationType;

  // type of adaptation manager 
  typedef Dune::AdaptationManager< GridType, RestrictionProlongationType > AdaptationManagerType;

  // type of time provider organizing time for time loops 
  typedef Dune::GridTimeProvider< GridType >                 TimeProviderType;

  // type of input/output
  typedef Dune::tuple< DiscreteFunctionType* >               IOTupleType;    /*@\label{baseevol:datawriterdef0}@*/

  // type of the data writer
  typedef Dune::DataWriter< GridType, IOTupleType >          DataWriterType; /*@\label{baseevol:datawriterdef1}@*/

  //! constructor 
  AlgorithmBase(GridType& grid) 
   : grid_( grid ),
     gridPart_( grid_ ),
     space_( gridPart_ ),
     // Initialize Timer for CPU time measurements
     timeStepTimer_( Dune::FemTimer::addTo("max time/timestep") ),
     loop_( 0 )
  {
  }

  /*@LST0S@*/ //! return reference to space 
  DiscreteSpaceType& space()
  {
    return space_;
  }


  //! initialize grid, i.e. to start adaptation (overload to do something)
  //! Return true if grid should be adapted; this method is then called again on the
  //! new grid. Now that on the final grid, the data should be set in the method
  //! initializeTimeStep().
  virtual bool initializeGrid(DiscreteFunctionType & solution) {return false;}

  //! initialize method for time loop, i.e. L2-project initial data 
  virtual void initializeStep(TimeProviderType& tp, 
                              DiscreteFunctionType & solution) = 0;

  //! solve one time step 
  virtual void step(TimeProviderType& tp,
                    DiscreteFunctionType & solution) = 0;

  //! finalize problem, i.e. calculate EOC ...
  virtual void finalizeStep(TimeProviderType& tp,
                            DiscreteFunctionType & solution) = 0;

  //! restore all data from check point (overload to do something)
  virtual void restoreFromCheckPoint(TimeProviderType& tp, 
                                     DiscreteFunctionType & solution) {}

  //! write a check point (overload to do something)
  virtual void writeCheckPoint(TimeProviderType& tp, DiscreteFunctionType & solution ) const {}

  //! default time loop implementation, overload for changes 
  virtual void operator()(DiscreteFunctionType & solution)
  {
    const bool verbose = Dune::Parameter :: verbose ();
    int printCount = Dune::Parameter::getValue<int>("femhowto.printCount", -1);

    const double maxTimeStep =
      Dune::Parameter::getValue("femhowto.maxTimeStep", std::numeric_limits<double>::max());
    const double startTime = Dune::Parameter::getValue<double>("femhowto.startTime", 0.0);
    const double endTime   = Dune::Parameter::getValue<double>("femhowto.endTime", 0.9);

    // for statistics
    double maxdt     = 0.;
    double mindt     = 1.e10;
    double averagedt = 0.;

    // Initialize TimeProvider
    TimeProviderType tp(startTime, grid_ );  /*@\label{baseevol:timeprovider}@*/

    RestrictionProlongationType rp( solution );

    // set refine weight 
    rp.setFatherChildWeight( Dune::DGFGridInfo<GridType> :: refineWeight() );

    // create adaptation manager 
    AdaptationManagerType adaptManager( grid_ , rp );

    // restoreData if checkpointing is enabled (default is disabled)
    restoreFromCheckPoint( tp, solution );

    // create data writer after restore is done, otherwise tp is not set correctly  
    IOTupleType dataTup ( &solution );
    DataWriterType eocDataOutput( grid_, dataTup, tp, EocDataOutputParameters(loop_) );

    // adapt the grid to the initial data
    int startCount = 0;
    while ( initializeGrid(solution) )
    {
      eocDataOutput.write( tp );
      adaptManager.adapt();
      if(verbose) 
        std::cout << "start: " << startCount << " grid size: " << grid_.size(0)
                  << std::endl;
      ++startCount;
    }
    // set initial data (and create ode solver)
    initializeStep(tp, solution); /*@\label{baseevol:stepperInitialize}@*/

    // write initial data
    eocDataOutput.write( tp );

    //**********************************************                    /*@LST0S@*/
    //* Time Loop                                  *
    //**********************************************
    tp.provideTimeStepEstimate(maxTimeStep);
    for( tp.init() ; tp.time() < endTime ; tp.next() )    /*@\label{baseevol:timeLoop}@*/
    {
      tp.provideTimeStepEstimate(maxTimeStep);                                          /*@LST0E@*/
      const double tnow  = tp.time();
      const double ldt   = tp.deltaT();
      const int counter  = tp.timeStep();

      //************************************************        /*@LST0S@*/
      //* Compute an ODE timestep                      *
      //************************************************
      Dune::FemTimer::start(timeStepTimer_);
      step(tp, solution );                                      /*@\label{baseevol:step}@*/
      Dune::FemTimer::stop(timeStepTimer_,Dune::FemTimer::max);              /*@LST0E@*/

       // Check that no NAN have been generated
      if (! solution.dofsValid()) {
        std::cout << "Loop(" << loop_ << "): Invalid DOFs" << std::endl;
        eocDataOutput.write(tp);
        abort();
      }

      if( (printCount > 0) && (counter % printCount == 0) )
      {
        unsigned int gridSize = grid_.size( 0 );
        gridSize = grid_.comm().sum( gridSize );
        if( grid_.comm().rank() == 0 )
        {
          std::cout << "step: " << counter << "  time = " << tnow << ", dt = " << ldt
                    << " grid size: " << gridSize << std::endl;
        }
      }

      // adapt grid and solution (marking should be done in step() )
      adaptManager.adapt();

      // write data 
      eocDataOutput.write( tp );     /*@LST0S@*/ /*@LST0E@*/            /*@\label{dg:datawrite1}@*/

      // possibly write check point (default is disabled)
      writeCheckPoint( tp, solution );

      // some statistics
      mindt = (ldt<mindt)?ldt:mindt;
      maxdt = (ldt>maxdt)?ldt:maxdt;
      averagedt += ldt;
    } /****** END of time loop *****/ /*@LST0S@*/

    averagedt /= double(tp.timeStep());
    if(verbose)
    {
      std::cout << "Minimum dt: " << mindt
         << "\nMaximum dt: " << maxdt
         << "\nAverage dt: " << averagedt << std::endl;
    }

    // finalize time step 
    finalizeStep(tp, solution);                                   /*@\label{baseevol:finalize}@*/
    
    // increase loop counter
    ++loop_;
  }

  //! finalize problem, i.e. calculated EOC ...
  virtual void finalize(DiscreteFunctionType & solution) 
  {}

protected:      
  GridType&            grid_;
  GridPartType         gridPart_;
  DiscreteSpaceType    space_;
  unsigned int timeStepTimer_; 
  unsigned int loop_ ; 
}; /*@LST0E@*/

#endif
