#ifndef DUNE_PROBLEMINTERFACE_HH
#define DUNE_PROBLEMINTERFACE_HH

// dune-common includes
#include <dune/common/fmatrix.hh>

// dune-fem includes
#include <dune/fem/function/common/function.hh>

namespace Dune {

/**
 * @brief describes the interface for
 * initial and exact solution of the advection-diffusion model
 */
template <class FunctionSpaceImp, bool constantVelocity >
class EvolutionProblemInterface
: public Fem :: Function< FunctionSpaceImp,
                          EvolutionProblemInterface< FunctionSpaceImp, constantVelocity> >
{
public:
  typedef FunctionSpaceImp                                           FunctionSpaceType;
  typedef EvolutionProblemInterface< FunctionSpaceType,
                                     constantVelocity >              ThisType;
  typedef Fem :: Function< FunctionSpaceType, ThisType >             BaseType;

  enum { ConstantVelocity = constantVelocity };
  enum { dimDomain = FunctionSpaceType :: dimDomain };

  typedef typename FunctionSpaceType :: DomainType                   DomainType;
  typedef typename FunctionSpaceType :: RangeType                    RangeType;
  typedef typename FunctionSpaceType :: DomainFieldType              DomainFieldType;
  typedef typename FunctionSpaceType :: RangeFieldType               RangeFieldType;

  /**
   * @brief define problem parameters
   */
protected:
  EvolutionProblemInterface()
  { }

public:
  //! destructor
  virtual ~EvolutionProblemInterface() {}

  /** \brief start time of problem */
  virtual double startTime() const { return 0.0; }

  /** \brief diffusion coefficient of problem */
  virtual double epsilon() const { return 0.0; }

  /**
   * @brief getter for the velocity
   */
  virtual void velocity(const DomainType& x, DomainType& v) const {}

  /**
   * @brief evaluates \f$ u_0(x) \f$
   */
  virtual void evaluate(const DomainType& arg, RangeType& res) const {
    evaluate(arg, startTime(), res);
  }

  /**
   * @brief old version of the exact solution
   *
   * old version of evaluate(const DomainType& arg, double t, RangeType& res),
   * which is still needed by the DataWriter
   */
  virtual inline void evaluate(const double t,
                               const DomainType& arg, RangeType& res) const {
    evaluate(arg, t, res);
  }

  /**
   * @brief evaluate exact solution
   */
  virtual void evaluate(const DomainType& arg,
                        const double t, RangeType& res) const {};

  /**
   * @brief latex output for EocOutput, default is empty
   */
  virtual std::string description() const
  {
    return std::string("");
  }
};


template <class FunctionSpaceImp>
class ProblemInterface
{
public:
  typedef FunctionSpaceImp                                           FunctionSpaceType;
  typedef ProblemInterface< FunctionSpaceType >                      ThisType;

  enum { dimDomain = FunctionSpaceType :: dimDomain };

  typedef typename FunctionSpaceType :: DomainType                   DomainType;
  typedef typename FunctionSpaceType :: RangeType                    RangeType;
  typedef typename FunctionSpaceType :: JacobianRangeType            JacobianRangeType;
  typedef typename FunctionSpaceType :: DomainFieldType              DomainFieldType;
  typedef typename FunctionSpaceType :: RangeFieldType               RangeFieldType;

  typedef FieldMatrix< RangeFieldType, dimDomain, dimDomain >        DiffusionMatrixType;

public:

  //! destructor
  virtual ~ProblemInterface() {}

  //! evaluates the right hand side (Function like bahavior)
  inline void evaluate(const DomainType& x, RangeType& ret) const
  {
    f(x , ret);
  }

  //! the right hand side
  virtual void f(const DomainType& x, RangeType& ret) const = 0;

  //! the exact solution
  virtual void u(const DomainType& x, RangeType& ret) const = 0;

  //! the diffusion matrix
  virtual void K(const DomainType& x, DiffusionMatrixType& m) const = 0;

  //! returns true if diffusion coefficient is constant
  virtual bool constantK() const = 0;

  //! the Dirichlet boundary data function
  virtual void g(const DomainType& x, RangeType& ret) const
  {
    u( x, ret );
  }

  //! the Neumann boundary data function
  virtual void psi(const DomainType& x,
                   JacobianRangeType& dn) const
  {
    gradient( x, dn );
  }

  //! the gradient of the exact solution
  virtual void gradient(const DomainType& x,
                        JacobianRangeType& grad) const = 0;

  //! return whether boundary is Dirichlet (true) or Neumann (false)
  virtual bool dirichletBoundary(const int bndId, const DomainType& x) const
  {
    return true;
  }

protected:
  ProblemInterface() : exactSolution_( *this ) {}

  //! the exact solution to the problem for EOC calculation
  class ExactSolution
  : public Fem :: Function< FunctionSpaceType, ExactSolution >
  {
  private:
    typedef Fem :: Function< FunctionSpaceType, ExactSolution >             BaseType;

  protected:
    const ThisType    &data_;

  public:
    inline ExactSolution ( const ThisType& data )
    : data_( data )
    {
    }

    inline void evaluate ( const DomainType &x, RangeType &ret ) const
    {
      data_.u( x, ret );
    }

    inline void jacobian ( const DomainType &x, JacobianRangeType &ret ) const
    {
      data_.gradient( x, ret );
    }

    inline void evaluate (const DomainType &x,
                          const double time, RangeType &phi ) const
    {
      evaluate( x, phi );
    }
  }; // end class ExactSolution

public:
  //! type of function converter for exact solution and gradient
  typedef ExactSolution ExactSolutionType;

protected:
  ExactSolutionType exactSolution_;

public:
  const ExactSolutionType& exactSolution() const { return exactSolution_; }
};

}
#endif  /*DUNE_PROBLEMINTERFACE_HH*/
