// Include configuration (generated by configure)
#include <config.h>

// Include some standard libraries
#include <iostream>
#include <string>

// Include required dune-fem headers
#include <dune/fem/misc/double.hh>

#include <dune/fem/space/fvspace.hh>
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/solver/odesolver.hh>
#include <dune/fem/misc/l1error.hh>

// Include dune-fem-howto base headers
#include <dune/fem-howto/baseevolution.hh>

#if HAVE_DUNE_FEM_DG
//#define HIGHERORDER_FV 
#endif

#ifdef HIGHERORDER_FV
#include <dune/fem-dg/operator/limiter/limiter.hh>
#include "limitermodel.hh"
#endif

// Include problem data
#if PROBLEM == 1 
#include "problem.hh"
#elif PROBLEM == 2 
#include "rotatingpulse.hh"
#elif PROBLEM == 3 
#include "../../tutorial/finitevolume/problem.hh"
#endif

template< class DestinationType, class Problem >                                /*@LST0S@*/
struct FiniteVolumeScheme 
: public Dune::SpaceOperatorInterface< DestinationType >
{
  // type of discrete function space 
  typedef typename DestinationType::DiscreteFunctionSpaceType DiscreteSpaceType;

  // type of grid part
  typedef typename DiscreteSpaceType::GridPartType GridPartType;
  
  // type of hierarchic grid
  typedef typename GridPartType::GridType HGridType;

  // suitable iterator type over the codim 0 entities of the grid part
  typedef typename DiscreteSpaceType::IteratorType IteratorType;

  // type of codim 0 entity and entity pointer
  typedef typename IteratorType::Entity Entity;
  typedef typename Entity::EntityPointer EntityPointer;

  // type of geometry
  typedef typename Entity::Geometry Geometry;

  // type of intersections and intersection iterators
  typedef typename GridPartType::IntersectionIteratorType
    IntersectionIteratorType;
  typedef typename IntersectionIteratorType::Intersection IntersectionType;

  // type of local function
  typedef typename DestinationType::LocalFunctionType LocalFunctionType;        /*@LST0E@*/

  // type of global domain coordinates and range of problem
  typedef typename Problem::RangeType RangeType;
  typedef typename Problem::DomainType DomainType;

  // type of local coordinates for elements (codim 0)
  typedef typename Geometry::LocalCoordinate  LocalCoordinate;

  // type of field representing domain coordinates (e.g., double)
  typedef typename LocalCoordinate::field_type FieldType;

  // type of local coordinates for faces (codim 1)
  typedef Dune::FieldVector< FieldType, LocalCoordinate::dimension-1 >
    FaceLocalCoordinate;

  // type of reference element and reference element container for faces
  typedef Dune::GenericReferenceElement< FieldType, FaceLocalCoordinate::dimension >
    ReferenceElementType;
  typedef Dune::GenericReferenceElements< FieldType, FaceLocalCoordinate::dimension >
    ReferenceElementContainerType;

#ifdef HIGHERORDER_FV
  typedef Dune :: LimiterModel < GridPartType, Problem > LimiterModelType;
  typedef Dune :: Fem :: LimitedReconstruction< LimiterModelType, DestinationType >
    LinearReconstructionType;

  typedef typename LinearReconstructionType :: LocalFunctionType  LinearFunctionType;
  typedef typename LinearReconstructionType :: ReconstructionType ReconstructionType;
#endif

  // constructor
  FiniteVolumeScheme ( const DiscreteSpaceType &space, 
                       const Problem &problem,
                       const bool higherOrder = false )
  : space_( space ),
    problem_( problem ),
    indicator_("indicator",space),
    adaptive_( Dune::AdaptationMethod< HGridType > ( space.grid() ).adaptive() )
#ifdef HIGHERORDER_FV
    , model_( problem_ )
    , reconstruction_( model_, space_ )
#endif
    , higherOrder_( higherOrder )
    , lMax_( Dune::Parameter:: getValue<int> ("fem.adaptation.finestLevel" ) )
    , lMin_( Dune::Parameter:: getValue<int> ("fem.adaptation.coarsestLevel" ) )
  {}

  // return reference to space (needed by ode solvers)                          /*@LST0S@*/
  const DiscreteSpaceType &space () const 
  {
    return space_;
  } /*@LST0E@*/

  // initialize solution to initial data                                        /*@LST0S@*/ 
  void initialize ( DestinationType &solution )
  {
    // iterate over all entities
    const IteratorType end = space().end();
    for( IteratorType it = space().begin(); it != end; ++it )                   /*@\label{fv:initLoop}@*/
    {
      // obtain entity and geometry
      const Entity &entity = *it;
      const Geometry &geo = entity.geometry();

      // evaluate initial data in element's center
      RangeType initialValue;
      problem_.evaluate( geo.center(), initialValue );

      // obtain local function
      LocalFunctionType lf = solution.localFunction( entity );

      // directly manipulate the only local DoF (degree of freedom)
      lf[ 0 ] = initialValue[ 0 ];
    }

    // communicate the data to other processes (for parallel runs only)
    solution.communicate(); /*@\label{fv:solcomm}@*/  
  }                                                                             /*@LST0E@*/ 

  // application operator                                                       /*@LST0S@*/
  void operator () ( const DestinationType &solution,
                     DestinationType &update ) const
  {
#ifdef HIGHERORDER_FV
    // calculate linear reconstruction 
    reconstruction_.update( solution );
#endif

    // clear update (i.e., initialize to zero)
    update.clear();

    // initialize time step to infinity
    dt_ = std::numeric_limits< double >::infinity();

    // obtain grid part and index set references
    const GridPartType &gridPart = space().gridPart();
    const typename GridPartType::IndexSetType &indexSet = gridPart.indexSet();

    // iterate over all entities
    const IteratorType end = space().end();
    for( IteratorType it = space().begin(); it != end; ++it )                   /*@\label{fv:entityLoop}@*/
    {
      // obtain entity and geometry
      const Entity &entity = *it;                                               /*@\label{fv:entityInit0}@*/
      const Geometry &geo = entity.geometry();

      // assert CFL condition for quadrilateral grids
      const double timeFactor = (geo.type().isCube() ? 0.5 : 1.0);

      // get local function of update 
      LocalFunctionType lfUpdEn = update.localFunction( entity );

      // obtain local functions for solution and update
      RangeType uLeft (0);
#ifdef HIGHERORDER_FV
      LinearFunctionType reconEn = reconstruction_.localFunction( entity );
#endif
      // for first order evaluation like this is ok, 
      // for higher order this is wrong 
      if( ! higherOrder_ ) 
      {
        evaluateConstantFunction( solution, entity, uLeft );
      }

      // element volume
      const double enVolume = geo.volume();                                     /*@\label{fv:entityInit1}@*/

      // iterate over all intersections
      const IntersectionIteratorType iend = gridPart.iend( entity );
      for( IntersectionIteratorType iit = gridPart.ibegin( entity );
           iit != iend; ++iit )
      {
        // obtain intersection
        const IntersectionType &intersection = *iit;                            /*@\label{fv:intersectionInit0}@*/

        // obtain reference element for intersection
        const ReferenceElementType &referenceElement
          = ReferenceElementContainerType::general(intersection.type());

        // local coordinate of intersection's barycenter
        const FaceLocalCoordinate &x = referenceElement.position( 0, 0 );

        const double volumeIntersection = referenceElement.volume();            /*@\label{fv:intersectionInit1}@*/

        // obtain velocity on intersection center 
        DomainType velocity;
        problem_.velocity( intersection.geometry().center(), velocity );        /*@\label{fv:numFlux0}@*/

        if( intersection.neighbor() )                                          /*@\label{fv:interiorFace}@*/
        {
          // access neighbor
          const EntityPointer neighborPointer = intersection.outside();         /*@\label{fv:neighborInit0}@*/
          const Entity &neighbor = *neighborPointer;                            
          /*@\label{fv:neighborInit1}@*/

          // compute flux from one side only (except it's not an interior entity)
          if( (neighbor.partitionType() != Dune::InteriorEntity)                /*@\label{fv:calcFluxJustOnce}@*/
              || (indexSet.index( entity ) < indexSet.index( neighbor )) )      /*@LST0E@*/ 
          {
            // obtain neighbor volume
            const double nbVolume = neighbor.geometry().volume();

            // get integration outer normal
            DomainType normal = intersection.integrationOuterNormal( x );       /*@LST0S@*/ /*@LST0E@*/ 

            // obtain local functions on neighbor entity
            RangeType uRight;
#ifdef HIGHERORDER_FV
            if( higherOrder_ )
            {
              LinearFunctionType reconNb = reconstruction_.localFunction( neighbor );
              // check whether intersection is conforming
              if( intersection.conforming() )
                evaluateOnIntersection< true > ( intersection, reconEn, reconNb, uLeft, uRight);
              else
                evaluateOnIntersection< false >( intersection, reconEn, reconNb, uLeft, uRight);
            }
            else
#endif 
            {
              evaluateConstantFunction( solution, neighbor, uRight );
            }

            // obtain local update function on neighbor entity
            LocalFunctionType lfUpdNb = update.localFunction( neighbor );

            // calculate wave speed
            const double waveSpeed = (velocity * normal) * volumeIntersection;

            // compute upwind numerical flux
            RangeType flux;
            flux[ 0 ]
              = (waveSpeed > 0 ? uLeft[ 0 ] : uRight[ 0 ]) * waveSpeed;      /*@\label{fv:numFlux1}@*/

            // compute update for entity and neighbor
            lfUpdEn[ 0 ] -= flux[ 0 ] / enVolume;
            lfUpdNb[ 0 ] += flux[ 0 ] / nbVolume;

            // compute time step restriction
            const double dtLocal = timeFactor * std::min( enVolume, nbVolume ) / std::abs( waveSpeed );
            dt_ = std::min( dt_, dtLocal );
          }
        }
        else 
        {
          DomainType normal = intersection.integrationOuterNormal( x );

          // calculate wave speed
          const double waveSpeed = (velocity * normal) * volumeIntersection;

          // compute upwind numerical flux
          RangeType flux;
          if( waveSpeed < 0 )
          {                                                                     /*@\label{fv:inflowBoundary0}@*/
            const DomainType xBnd
              = geo.global( intersection.geometryInInside().global( x ) );

            RangeType uBnd;
            problem_.evaluate ( time_, xBnd, uBnd );
            flux[ 0 ] = uBnd[ 0 ] * waveSpeed;
          }                                                                     /*@\label{fv:inflowBoundary1}@*/
          else
          {
#ifdef HIGHERORDER_FV
            // evaluate only for higher order (for first order see above)
            if( higherOrder_ )
            {
              evaluateOnBoundary( intersection, reconEn, uLeft );
            }
#endif
            flux[ 0 ] = uLeft[ 0 ] * waveSpeed;
          }

          // compute update for entity
          lfUpdEn[ 0 ] -= flux[ 0 ] / enVolume;

          // compute time step restriction
          const double dtLocal = timeFactor * enVolume / std::abs( waveSpeed );
          dt_ = std::min( dt_, dtLocal );
        }
      }
    }

    // communicate the data to other processes (for parallel runs only)
    update.communicate(); /*@\label{fv:updatecomm}@*/ 
  }                                                                             /*@LST0E@*/

  // set current time of space operator
  void setTime ( const double time )
  {
    time_ = time;
  }

  // return maximum possible time step ensuring stability of Euler scheme
  double timeStepEstimate () const 
  {
    return dt_;  
  }

  // return true if higherOrder scheme is used 
  const bool higherOrder() const { return higherOrder_; }

  int lMax() const { return lMax_ ; }

  bool mark ( HGridType &grid, DestinationType &solution ) const
  {
    // if method no adaptation is used do nothing 
    if( ! adaptive_ ) return false;

    // tol value for refinement strategy
    const double refinetol  = 0.05;
    const double coarsentol = 0.001;

    // obtain grid part and index set references
    const GridPartType &gridPart = space().gridPart();
    const typename GridPartType::IndexSetType &indexSet = gridPart.indexSet();

    indicator_.clear();

    double globalmax = -1E100;
    double globalmin =  1E100;

    // iterate over all entities
    const IteratorType end = space().end();
    for( IteratorType it = space().begin(); it != end; ++it )                   
    {
      const Entity &entity = *it;                                              
      // obtain local functions for solution and update
      LocalFunctionType lfSolEn = solution.localFunction( entity );
      LocalFunctionType lfIndEn = indicator_.localFunction( entity );
 
      // global min/max
      globalmax = std::max(globalmax, lfSolEn[0]);
      globalmin = std::min(globalmin, lfSolEn[0]);

      // iterate over all intersections
      const IntersectionIteratorType iend = gridPart.iend( entity );
      for( IntersectionIteratorType iit = gridPart.ibegin( entity ); iit != iend; ++iit )
      {
        // obtain intersection
        const IntersectionType &intersection = *iit;                           
        if( ! intersection.neighbor() )
          continue;

        // access neighbor
        const EntityPointer neighborPointer = intersection.outside();         
        const Entity &neighbor = *neighborPointer;                            

        // handle face from one side only
        if( (neighbor.partitionType() != Dune::InteriorEntity)                
            || (indexSet.index( entity ) < indexSet.index( neighbor )) )       
        {
          LocalFunctionType lfSolNb = solution.localFunction( neighbor );
          LocalFunctionType lfIndNb = indicator_.localFunction( neighbor );
          double localdelta = std::abs(lfSolEn[0]-lfSolNb[0]);
          lfIndEn[0] = std::max(lfIndEn[0],localdelta);         
          lfIndNb[0] = std::max(lfIndNb[0],localdelta);         
        }
      }
    }

    // make max and min equal on each process 
    globalmax = grid.comm().max( globalmax );
    globalmin = grid.comm().min( globalmin );

    // mark cells for refinement/coarsening
    double globaldelta = globalmax-globalmin;
    int marked=0;
    for( IteratorType it = space().begin(); it != end; ++it )                   
    {
      const Entity &entity = *it;                                              
      LocalFunctionType lfIndEn = indicator_.localFunction( entity );
      if (lfIndEn[0] > refinetol*globaldelta 
          && entity.level() < lMax_ && entity.isRegular())
      {
        grid.mark( 1, entity );
        ++marked;
        /*
        // iterate over all intersections
        const IntersectionIteratorType iend = gridPart.iend( entity );
        for( IntersectionIteratorType iit = gridPart.ibegin( entity ); iit != iend; ++iit )
        {
          const IntersectionType &intersection = *iit;                           
          if( !intersection.neighbor() )
            continue;

          const EntityPointer neighborPointer = intersection.outside();         
          const Entity &neighbor = *neighborPointer;                            
          if( (neighbor.level() < lMax_) && neighbor.isRegular() )
            grid.mark( 1, neighbor );
        }
        */
      }
      else if ( lfIndEn[0] < coarsentol*globaldelta && entity.level() > lMin_ )
      {
        grid.mark( -1, entity );
        ++marked;
      }
      else 
        grid.mark(0, entity);
    }                                              /*@\label{fah:loop3}@*/
    // std::cout << "    elements marked: " << marked << std::endl;

    // make marked the same for all cores 
    marked = grid.comm().max( marked );

    return (marked>0);
  }


  void evaluateConstantFunction( const DestinationType& solution, 
                                 const Entity& entity, 
                                 RangeType& uValue) const  
  {
    LocalFunctionType lf = solution.localFunction( entity );
    assert( RangeType :: dimension == lf.numDofs() );
    for(int l=0; l< RangeType :: dimension; ++l) 
      uValue[ l ] = lf[ l ];
  }

#ifdef HIGHERORDER_FV
  ReconstructionType& reconstruction( DestinationType& u )
  {
    reconstruction_.update( u );
    return reconstruction_.function();
  }

protected:
  template < bool conforming >
  void evaluateOnIntersection( const IntersectionType& intersection,
                               const LinearFunctionType& reconEn,
                               const LinearFunctionType& reconNb,
                               RangeType& uLeft,
                               RangeType& uRight) const
  {
    // make sure we got the right conforming statement
    assert( intersection.conforming() == conforming );

    typedef Dune :: CachingQuadrature< GridPartType, 1 > FaceQuadratureType;
    // use IntersectionQuadrature to create appropriate face quadratures 
    typedef Dune :: IntersectionQuadrature< FaceQuadratureType, conforming > IntersectionQuadratureType;
    typedef typename IntersectionQuadratureType :: FaceQuadratureType QuadratureImp;

    // create intersection quadrature 
    IntersectionQuadratureType interQuad( space_.gridPart(), intersection, 0 );

    // get appropriate references 
    const QuadratureImp &faceQuadInner = interQuad.inside();
    const QuadratureImp &faceQuadOuter = interQuad.outside();

    // set neighbor to caller 
    assert( faceQuadInner.nop() == 1 );

    reconEn.evaluate( faceQuadInner[ 0 ], uLeft  );
    reconNb.evaluate( faceQuadOuter[ 0 ], uRight );
  }

  void evaluateOnBoundary( const IntersectionType& intersection,
                           const LinearFunctionType& reconEn,
                           RangeType& uLeft ) const
  {
    typedef Dune :: CachingQuadrature< GridPartType, 1 > FaceQuadratureType;
    FaceQuadratureType faceQuadInner( space_.gridPart(), intersection, 0, FaceQuadratureType::INSIDE);

    // set neighbor to caller 
    assert( faceQuadInner.nop() == 1 );
    reconEn.evaluate( faceQuadInner[ 0 ], uLeft  );
  }
#endif


protected:
  const DiscreteSpaceType &space_;
  const Problem &problem_;
  double time_;
  mutable double dt_;
  mutable DestinationType indicator_;
  const bool adaptive_ ;

#ifdef HIGHERORDER_FV
  const LimiterModelType model_;
  mutable LinearReconstructionType reconstruction_;
#endif
  const bool higherOrder_;
  const int lMax_;
  const int lMin_;
};


template <class GridImp> /*@LST0S@*/
struct FVStepperTraits 
{
  // type of grid 
  typedef GridImp GridType;

  // Initial data
  //typedef U0<GridType, Dune::Double> InitialData;
  typedef U0<GridType, double> InitialData;
  // Define the discrete function for the unkown solution
  // ... first the grid part ...
  typedef Dune::DGAdaptiveLeafGridPart<GridType> GridPartType;
  //typedef Dune::LeafGridPart<GridType> GridPartType;
  // ... next the function space ...
  typedef Dune::FunctionSpace
          <typename InitialData::DomainFieldType,
           typename InitialData::RangeFieldType,
           GridType::dimensionworld, InitialData::dimRange>
          FunctionSpaceType;
  // ... followed by the discrete function space type ...
  typedef Dune::FiniteVolumeSpace
          < FunctionSpaceType, GridPartType, 0, Dune::CachingStorage >
          DiscreteSpaceType;
  // ... and finally the type for the discrete function ...
  typedef Dune::AdaptiveDiscreteFunction
          < DiscreteSpaceType > 
          DiscreteFunctionType; /*@LST0E@*/

  // type of restriction/prolongation projection for adaptive simulations 
  typedef Dune::RestrictProlongDefault< DiscreteFunctionType >  RestrictionProlongationType;
};

template <class GridImp> /*@LST0S@*/
struct Stepper : public AlgorithmBase< FVStepperTraits< GridImp > >  
{
  // my traits class 
  typedef FVStepperTraits< GridImp > Traits;
  // base class 
  typedef AlgorithmBase< Traits > BaseType ;

  // type of grid 
  typedef typename Traits :: GridType GridType;

  // Initial data
  typedef typename Traits :: InitialData InitialData;
  // Define the discrete function for the unkown solution
  // ... first the grid part ...
  typedef typename Traits :: GridPartType GridPartType;
  // ... followed by the discrete function space type ...
  typedef typename Traits :: DiscreteSpaceType   DiscreteSpaceType;
  // ... and finally the type for the discrete function ...
  typedef typename Traits :: DiscreteFunctionType  DiscreteFunctionType; 

  // type of time provider 
  typedef typename BaseType :: TimeProviderType TimeProviderType;

  typedef FiniteVolumeScheme<DiscreteFunctionType,InitialData> SpaceSolverType;

  Stepper(GridType& grid) :
    BaseType( grid ),
    problem_(),
    spaceSolver_(this->space(), problem_ 
#ifdef HIGHERORDER_FV
        , Dune::Parameter::getValue<bool>("femhowto.higherorder", false) 
#endif
        ),
    odeSolver_(0),
    eocId_( Dune::FemEoc::addEntry(std::string("L1-error")) ),
    verbose_( Dune::Parameter :: verbose() )
  {
  }
  
  bool initializeGrid(DiscreteFunctionType & u)
  {
    const int lmax = spaceSolver_.lMax();
    int count = BaseType::grid_.maxLevel();

    count = BaseType::grid_.comm().max( count );
    if( count >= lmax ) return false ;

    spaceSolver_.initialize(u);
    bool adaptAgain = spaceSolver_.mark(BaseType::grid_,u);
#ifndef NDEBUG 
    // make sure every processes finished the above steps 
    BaseType::grid_.comm().barrier();
#endif
    return adaptAgain ;
  }

  void initializeStep(TimeProviderType& tp, DiscreteFunctionType& u)  
  {
    const int rkOrder = ( spaceSolver_.higherOrder() ) ? 2 : 1;
    odeSolver_ = new DuneODE::ExplicitRungeKuttaSolver<DiscreteFunctionType>(spaceSolver_,tp, rkOrder);
    spaceSolver_.initialize(u);
    odeSolver_->initialize(u);
  } /*@LST0E@*/ 
  
  void step(TimeProviderType& tp, DiscreteFunctionType& u) { /*@LST0S@*/ 
    assert(odeSolver_);
    odeSolver_->solve(u);
    spaceSolver_.mark(BaseType::grid_,u);
  } /*@LST0E@*/ 
  
  void finalizeStep(TimeProviderType& tp, DiscreteFunctionType& u) { /*@LST0S@*/ 
    typename InitialData::RangeType error;
#ifdef HIGHERORDER_FV
    if ( spaceSolver_.higherOrder() )
    {
      typedef typename SpaceSolverType :: ReconstructionType ReconstructionType;
      Dune::L1Error< ReconstructionType > L1err;
      // Compute L1 error of discretized solution ...
      error = L1err.norm(problem_, spaceSolver_.reconstruction( u ), tp.time());
    }
    else
#endif
    {
      Dune::L1Error<DiscreteFunctionType> L1err;
      // Compute L1 error of discretized solution ...
      error = L1err.norm(problem_, u, tp.time());
    }

    // ... and print the statistics out to the eocOutputPath file
    typename InitialData::RangeFieldType err = error.two_norm2();
    Dune::FemEoc::setErrors(eocId_,sqrt(err));
    delete odeSolver_;
    odeSolver_ = 0;
  } /*@LST0E@*/ 
  
  std::string description() 
  {
    return problem_.description();
  }
  private:
  // InitialDataType is a Dune::Operator that evaluates 
  // to $u_0$ and also has a method that gives you the exact solution.
  InitialData problem_;
  SpaceSolverType spaceSolver_;
  DuneODE::OdeSolverInterface<DiscreteFunctionType> *odeSolver_;
  unsigned int eocId_;
  const int verbose_;
  double cfl_;
};

int main ( int argc, char **argv )                                              /*@LST1S@*/
try
{ 
  typedef Dune::GridSelector::GridType GridType;
  
  // initialization
  Dune::MPIManager::initialize( argc, argv );
  Dune::Parameter::append( argc, argv );                                        /*@\label{fv:param0}@*/
  Dune::Parameter::append( argc >= 2 ? argv[ 1 ] : "parameter" );               /*@\label{fv:param1}@*/

  Dune::GridPtr< GridType > gridptr                                             /*@\label{fv:grid}@*/
    = initialize< GridType >( "Finite Volume Scheme" );
  GridType &grid = *gridptr;

  // EOC loop
  Stepper< GridType > stepper( grid );                                          /*@\label{fv:eocLoop0}@*/
  compute(stepper);                                                             /*@\label{fv:eocLoop1}@*/

  // dump parameters
  Dune::Parameter::write( "parameter.log" );

  return 0;
}
catch( const Dune::Exception &e )                                               /*@\label{fv:catch0}@*/
{
  std::cerr << e << std::endl;
  return 1;
}
catch (...)
{
  std::cerr << "Generic exception caught." << std::endl;
  return 2;
}                                                                               /*@\label{fv:catch1}@*/ /*@LST1E@*/
