#ifndef LIMITER_MODEL_HH
#define LIMITER_MODEL_HH

#include <cmath>
#include <dune/common/fvector.hh>
#include <dune/fem/misc/field.hh>

#include <dune/fem/storage/vector.hh>
#include <dune/fem/io/parameter.hh>

namespace Dune {

template <class GridPart,
          class Field,
          int dimRange2,
          int dimRange1 = dimRange2 * GridPart::GridType::dimensionworld>
class LimiterModelTraits {
 public:
  typedef Field RangeFieldType;
  typedef Field FieldType;
  typedef GridPart GridPartType;
  typedef typename GridPart::GridType GridType;
  enum { dimDomain = GridType::dimensionworld };
  enum { dimRange = dimRange2, dimGradRange = dimRange1 };
  typedef FieldVector<typename GridType::ctype, dimDomain> DomainType;
  typedef FieldVector<typename GridType::ctype, dimDomain-1> FaceDomainType;
  typedef FieldVector<RangeFieldType, dimRange> RangeType;
  typedef FieldVector<RangeFieldType, dimGradRange> GradientType;
  typedef typename GridPart::IntersectionIteratorType IntersectionIteratorType;
  typedef typename IntersectionIteratorType :: Intersection IntersectionType;
  typedef typename GridType::template Codim<0>::Entity EntityType;

  typedef MinModLimiter< FieldType > LimiterFunctionType ;
  //typedef SuperBeeLimiter< FieldType > LimiterFunctionType ;
  //typedef VanLeerLimiter< FieldType > LimiterFunctionType ;
};

// ************************************************
template <class GridPartType,class ProblemType>
class LimiterModel 
{
public:
  enum { dimDomain = GridPartType::GridType::dimensionworld };
  enum { dimRange = 1 };
  typedef LimiterModelTraits<GridPartType, double, dimRange> Traits;
  typedef typename Traits::FieldType FieldType;
  typedef typename Traits::GridType GridType;
  typedef typename Traits::RangeType RangeType;
  typedef typename Traits::DomainType DomainType;
  typedef typename Traits::FaceDomainType FaceDomainType;
  typedef typename GridPartType::IntersectionIteratorType IntersectionIteratorType;
  typedef typename IntersectionIteratorType :: Intersection IntersectionType;
  typedef typename Traits::EntityType EntityType;

protected:
  typedef LimiterModel<GridPartType,ProblemType> ThisType;

public:
  LimiterModel(const ProblemType& problem) :
    problem_(problem)
  {}

  // has flux 
  bool hasFlux() const { return true; }
  
  // has flux 
  bool hasSource() const { return false; }

  // needed for limitation
  inline void velocity(
             const EntityType& en,
             const double time,
             const DomainType& x,
             const RangeType& u,
             DomainType& velocity) const
  {
    // copy constant velocity vector 
    problem_.velocity( en.geometry().global( x ), velocity );

    // the Euler case 
    /*
    for(int i=0; i<dimDomain; ++i) 
    {
      // we store \rho u but do not need to divide by \rho here since only
      // sign is needed.
      velocity[i] = u[i+1];
    }
    */
  }

  // we have physical check for this model 
  bool hasPhysical() const                   
  {
    return true;
  }

  // calculate jump between left and right value
  inline  bool physical(  const EntityType& entity,
                          const DomainType& xLocal,
                          const RangeType& u ) const
  {
    // the set of states in R in this case 
    return true ;
  }

  // adjust average value if necessary 
  // (e.g. transform from conservative to primitive variables )
  void adjustAverageValue( const EntityType& entity, 
                           const DomainType& xLocal,
                           RangeType& u ) const
  {
    // nothing to be done here for this test case 
  }

  // calculate jump between left and right value 
  inline void jump( 
        const IntersectionType& it, 
        const double time, 
        const FaceDomainType& x,
        const RangeType& uLeft,
        const RangeType& uRight,
        RangeType &shockIndicator ) const 
  {
    shockIndicator  = (uLeft[0] - uRight[0])/(0.5 * (uLeft[0] + uRight[0]));
    /*
    // take pressure as shock detection values 
    const FieldType pl = pressure( uLeft );
    const FieldType pr = pressure( uRight );
    jump  = (pl-pr)/(0.5*(pl+pr));
    */
  }

  // calculate jump between left and right value 
  inline void adaptationIndicator(
                   const IntersectionType& it,
                   const double time,
                   const FaceDomainType& x,
                   const RangeType& uLeft,
                   const RangeType& uRight,
                   RangeType& indicator) const 
  {
    // take density as shock detection values 
    indicator = (uLeft[0] - uRight[0])/(0.5 * (uLeft[0]+uRight[0]));
    
    /*
    const DomainType unitNormal = it.unitOuterNormal(x);  
    RangeType ul(uLeft),ur(uRight);
    rot_.rotateForth( ul , unitNormal );
    rot_.rotateForth( ur , unitNormal );
    for (int i=1; i<dimDomain+1; ++i) 
    {
      indicator += ((ul[i+1]/ul[0] - ur[i+1]/ur[0])/problem_.V());
    }
    */
  }

  inline bool hasBoundaryValue(const IntersectionType& it,
                               const double time, 
                               const FaceDomainType& x) const 
  {
    return true;
  }
  
  inline  void boundaryValue(
            const IntersectionType& intersection,
            const double time, 
            const FaceDomainType& x,
            const RangeType& uLeft, 
            RangeType& uRight) const 
  {
    const DomainType normal = intersection.integrationOuterNormal( x );

    DomainType xGlobal = intersection.geometry().global( x );
    // obtain velocity
    DomainType velocity;
    problem_.velocity( xGlobal, velocity );

    // calculate upwind 
    const double upwind = (velocity * normal);
    
    if( upwind < 0 ) 
    {
      problem_.evaluate( xGlobal, time, uRight );
    }
    else 
    {
      uRight = uLeft;
    } 

    //if ( it.boundaryId() == 2 ) 
    /*
    else if ( it.boundaryId() == 3 ) 
    {
      DomainType xgl = intersectionGeometry( it ).global(x);
      problem_.evaluate(time,xgl,uRight);
    }
    else {
      abort();
      assert( false );
    }
    */
  }

 protected:
  const ProblemType& problem_;
};

}
#endif // header 
