#ifndef  DUNE_PROBLEM_HH__
#define  DUNE_PROBLEM_HH__


/**
 * @brief describes the initial and exact solution of the advection-diffusion model
 * taken from dune-grid-howto
 *
 */
template <class GridType, class Field = double >
class U0 : public Dune::Fem::Function<
            Dune::FunctionSpace<typename GridType :: ctype, Field,
                                GridType::dimensionworld, 1>, U0<GridType, Field> > {
 public:
  enum { ConstantVelocity = true };
  enum { dimDomain = GridType::dimensionworld };  
  enum { dimRange = 1 };  
  typedef typename GridType :: ctype DomainFieldType;
  typedef Field RangeFieldType;

  typedef Dune::FieldVector<DomainFieldType,dimDomain> DomainType;
  typedef Dune::FieldVector<RangeFieldType ,dimRange>  RangeType;


  /**
   * @brief define problem parameters
   */
  U0():
    velocity_(0),
    startTime_(Dune::Parameter::getValue<double>("femhowto.startTime",0.0))
  {
    velocity_[ 0 ] = 1; 
  }

  /**
   * @brief getter for the velocity
   */
  const DomainType &velocity() const {
    return velocity_;
  }

  /**
   * @brief the velocity method
   */
  void velocity(const DomainType& x, DomainType& velocity ) const {
    velocity = velocity_;
  }

  /**
   * @brief evaluates \f$ u_0(x) \f$
   */
  void evaluate(const DomainType& arg, RangeType& res) const {             /*@LST0S@@LST0E@*/
    evaluate(arg, startTime_, res);
  }
  
  /**
   * @brief old version of the exact solution
   * 
   * old version of evaluate(const DomainType& arg, double t, RangeType& res),
   * which is still needed by the DataWriter
   */
  inline void evaluate(double t,  const DomainType& arg, RangeType& res) const { 
    evaluate(arg, t, res);
  }
    
  /**
   * @brief evaluate exact solution
   */ 
  void evaluate(const DomainType& arg, double t, RangeType& res) const {  /*@LST0S@@LST0E@*/
    DomainType x(arg);
    x.axpy( -t, velocity_ );

    //if (x.two_norm()>0.125 && x.two_norm()<0.5 && x[0]>1e-8 && x[1]>1e-8)
    if (x[ 0 ] < 0.25 )
	    res[0]=1.0;
    else
	    res[0]=0.0;
  }
  
  /**
   * @brief latex output for EocOutput
   */
  std::string description()
  {
    std::ostringstream ofs;
    ofs << "Problem: " << myName 
        << " Velocity: " << velocity_
        << "\n\n";
    return ofs.str();
  }
 private:    
  DomainType velocity_;
  double startTime_;
 public:
  std::string myName;
};

#endif  /*DUNE_PROBLEM_HH__*/
