#ifndef DUNE_FEM_HOWTO_CHECKPOINTING_HH
#define DUNE_FEM_HOWTO_CHECKPOINTING_HH


#if HAVE_SIONLIB && WANT_SIONLIB
#include <dune/fem/io/streams/sionlibstreams.hh>
namespace Dune {

struct PersistenceManagerTraits 
{
  typedef Fem :: SIONlibOutStream  BackupStreamType ;
  typedef Fem :: SIONlibInStream   RestoreStreamType ;
  static const bool singleBackupRestoreFile = true ;
};
#define FEM_PERSISTENCEMANAGERSTREAMTRAITS  PersistenceManagerTraits
}
#endif

// include std libs
#include <iostream>
#include <string>

// Dune includes
#include <dune/fem/misc/l2error.hh>
#include <dune/fem/operator/projection/l2projection.hh>
#include <dune/fem/solver/odesolver.hh>

// space and function 
#include <dune/fem/space/dgspace.hh>
#include <dune/fem/gridpart/adaptiveleafgridpart.hh>
#include <dune/fem/function/adaptivefunction.hh>

// include local header files
#include <dune/fem-howto/baseevolution.hh>
#include "problem.hh" 

// approximation order
const int order   = POLORDER;

template <class GridImp, class InitialDataType>               /*@LST1S@*/
struct StepperTraits {
  // type of Grid
  typedef GridImp                                                    GridType;
  // Choose a suitable GridView
  typedef Dune::DGAdaptiveLeafGridPart< GridType >                       GridPartType;
  //typedef AdaptiveLeafGridPart< GridType >                         GridPartType;
  //typedef IdBasedLeafGridPart< GridType >                         GridPartType;

  // type of function space 
  typedef Dune::FunctionSpace< typename  GridType :: ctype, double, 
                               GridType :: dimensionworld, 1 >  FunctionSpaceType;

  // ... as well as the Space type
  typedef Dune::DiscontinuousGalerkinSpace < FunctionSpaceType, GridPartType, order> DiscreteSpaceType;

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef Dune::AdaptiveDiscreteFunction< DiscreteSpaceType >           DiscreteFunctionType;

  // type of restriction/prolongation projection for adaptive simulations 
  typedef Dune::RestrictProlongDefault< DiscreteFunctionType >  RestrictionProlongationType;
};

template <class GridImp, class InitialDataType>               /*@LST1S@*/
struct Stepper : public AlgorithmBase< StepperTraits< GridImp, InitialDataType> >
{
  // my traits class 
  typedef StepperTraits< GridImp, InitialDataType> Traits ;

  // my base class 
  typedef AlgorithmBase < Traits > BaseType;

  // type of Grid
  typedef typename Traits :: GridType                  GridType;

  // Choose a suitable GridView
  typedef typename Traits :: GridPartType              GridPartType;

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef typename Traits :: DiscreteFunctionType      DiscreteFunctionType;

  // ... as well as the Space type
  typedef typename Traits :: DiscreteSpaceType         DiscreteSpaceType;

  typedef typename BaseType :: TimeProviderType   TimeProviderType;

  // type of most simple check pointer 
  typedef Dune::CheckPointer< GridType >   CheckPointerType;

  using BaseType :: grid_;

  Stepper(GridType& grid, const InitialDataType& problem, const char* checkFile) :
    BaseType ( grid ),
    problem_(problem),
    eocId_( Dune::FemEoc::addEntry(std::string("$L^2$-error")) ),
    checkPointer_( 0 ),
    checkFile_( checkFile )
  {
    Dune::persistenceManager << error_;
  }                                                                        /*@LST1E@*/

  ~Stepper() 
  {
    removeObj();
  }

  void removeObj() 
  {
    delete checkPointer_;
    checkPointer_ = 0;
  }

  void finalize(DiscreteFunctionType&) 
  {
    removeObj();
  }

  CheckPointerType& checkPointer( TimeProviderType& tp ) const
  {
    // create check point if not exsistent 
    if( ! checkPointer_ ) 
      checkPointer_ = new CheckPointerType( grid_, tp );

    return *checkPointer_;
  }

  // restore data 
  void restoreFromCheckPoint(TimeProviderType& tp,
                             DiscreteFunctionType& solution) 
  {
    // add solution to persistence manager for check pointing 
    Dune::persistenceManager << solution ;

    // if check file is non-zero a restart is performed 
    if( checkFile_ ) 
    {
      // restore data 
      checkPointer( tp ).restoreData( grid_, checkFile_ );
  
      // check consistency of check point 
      consistencyCheck( tp, solution );
    }
  }

  // backup data  
  void writeCheckPoint(TimeProviderType& tp, DiscreteFunctionType & solution ) const 
  {
    std::cout << "Try to write checkpoint" << std::endl;
    checkPointer( tp ).write( tp );
  }

  // before first step, do data initialization 
  void initializeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    Dune::Fem::L2Projection<                          /*@\label{dg:l2pro0}@*/
                        InitialDataType, DiscreteFunctionType > l2pro;
    l2pro(problem_, u);                                   /*@\label{dg:l2pro1}@*/
  }

  // solve ODE for one time step 
  void step(TimeProviderType& tp, DiscreteFunctionType& u) 
  {
    typedef Dune::TimeDependentFunction< InitialDataType > FunctionType;
    FunctionType function( problem_, tp.time() );
    Dune::Fem::L2Projection<                         /*@\label{dg:l2pro0}@*/
                        FunctionType, DiscreteFunctionType > l2pro;
    l2pro(function, u);                                   /*@\label{dg:l2pro1}@*/

    // compute error for backup and restore 
    error_ = computeError(tp, u );
  }

  double computeError(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    Dune::L2Error<DiscreteFunctionType> L2err;
    // Compute L2 error of discretized solution ...
    typedef typename DiscreteFunctionType :: RangeType RangeType;
    RangeType error = L2err.norm(problem_, u, tp.time());
    return error.two_norm();
  }

  // after last step, do EOC calculation 
  void finalizeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    // ... and print the statistics out to the eocOutputPath file
    Dune::FemEoc::setErrors(eocId_, computeError(tp, u ) );
  }                                                       /*@LST1E@*/

  // after last step, do EOC calculation 
  void consistencyCheck(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    Dune::L2Error<DiscreteFunctionType> L2err;
    // Compute L2 error of discretized solution ...
    typename DiscreteFunctionType :: RangeType error;
    error = L2err.norm(problem_, u, tp.time());

    std::cout << "Stepper::consistencyCheck: L2-error after restore: " << error.two_norm() 
              << "  stored value: " << error_ << std::endl;
    if( std::abs( error.two_norm() - error_ ) > 1e-14 ) 
    {
      DUNE_THROW(Dune::InvalidStateException, "Error in backup/restore" );
    }
  }                                                       /*@LST1E@*/
private:
  // InitialDataType is a Dune::Operator that evaluates to $u_0$ and also has a
  // method that gives you the exact solution.
  const InitialDataType  &problem_;
  // Initial flux for advection discretization (UpwindFlux)
  const unsigned int      eocId_;

  // check point writer 
  mutable CheckPointerType* checkPointer_;
  // name of checkpoint file 
  const char* checkFile_;

  double error_;
};

#endif // #ifndef DUNE_FEM_HOWTO_CHECKPOINTING_HH
