// include configurations options
#include <config.h>


// Includes from the IOStream Library
// ----------------------------------

#include <iostream>
#include <sstream>


// Includes from Core Modules
// --------------------------


// Includes from DUNE-FEM
// ----------------------

// include basic grid parts
#include <dune/fem/gridpart/common/gridpart.hh>

// include Lagrange discrete function space
#include <dune/fem/space/lagrangespace.hh>

#include <dune/fem/function/adaptivefunction.hh>

#include <dune/fem/solver/inverseoperators.hh>

#include <dune/fem/misc/l2norm.hh>
#include <dune/fem/misc/h1norm.hh>

#include <dune/fem-howto/base.hh>

#include "massoperator.hh"



// Global Type Definitions
// -----------------------

#if defined POLORDER
const int polOrder = POLORDER;
#else
const int polOrder = 1;
#endif



// Function to Project
// -------------------

template< class FunctionSpace >
struct Function
{
  typedef FunctionSpace FunctionSpaceType;

  typedef typename FunctionSpaceType::DomainType DomainType;
  typedef typename FunctionSpaceType::RangeType RangeType;
  typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;

  void evaluate ( const DomainType &x, RangeType &value ) const
  {
    value = 1.;
    for( int k = 0; k < FunctionSpace::dimDomain; ++k )
      value *= sin( M_PI * x[k] );
  }

  void jacobian ( const DomainType &x, JacobianRangeType &jacobian ) const
  {
    for( int j = 0; j < FunctionSpace::dimDomain; ++j )
    {
      // jacobian has only one row, calc j-th column      
      jacobian[0][j] = M_PI;
      for( int k = 0; k < FunctionSpace::dimDomain; ++k )
        jacobian[0][j] *= (j == k ? cos( M_PI*x[k] ) : sin( M_PI*x[k] ));
    }
  }
};



// Algorithm
// ---------

struct Algorithm
{
  typedef Dune::GridSelector :: GridType GridType;
  typedef Dune::LeafGridPart< GridType > GridPartType;
  typedef Dune::FunctionSpace< double, double, GridType::dimensionworld, 1 > SpaceType;
  typedef Dune::LagrangeDiscreteFunctionSpace< SpaceType, GridPartType, polOrder >
    DiscreteSpaceType;
  typedef Dune::AdaptiveDiscreteFunction< DiscreteSpaceType > DiscreteFunctionType;

  typedef ::Function< SpaceType > FunctionType;

  explicit Algorithm ( GridType &grid );
  void operator() ( DiscreteFunctionType &u );
  void finalize ( DiscreteFunctionType &u );
  DiscreteSpaceType &space ();

private:
  GridPartType gridPart_;
  DiscreteSpaceType dfSpace_;
  FunctionType function_;
  int eocId_;
};


inline Algorithm::Algorithm ( GridType &grid )
: gridPart_( grid ),
  dfSpace_( gridPart_ )
{
  std::vector< std::string > eocHeaders( 2 );
  eocHeaders[ 0 ] = "$L^2$-Error";
  eocHeaders[ 1 ] = "$H^1$-Error";
  eocId_ = Dune::FemEoc::addEntry( eocHeaders );
}


inline void Algorithm::operator() ( DiscreteFunctionType &solution )
{
  // get operator
  typedef MassOperator< DiscreteFunctionType > MassOperatorType;
  MassOperatorType massOperator( dfSpace_ );

  // assemble RHS
  DiscreteFunctionType rhs( "rhs", dfSpace_ );
  massOperator.assembleRHS( function_, rhs );

  // get solver
  //typedef Dune::CGInverseOp< DiscreteFunctionType, MassOperatorType > InverseOperator;  
  typedef Dune::CGInverseOperator< DiscreteFunctionType > InverseOperator;  
  InverseOperator inverseOperator( massOperator, 1e-10, 1e-10, 50000 );

  // apply solver
  solution.clear();
  inverseOperator( rhs, solution );
}


inline void Algorithm::finalize ( DiscreteFunctionType &solution )
{
  typedef Dune::GridFunctionAdapter< FunctionType, GridPartType > GridFunctionType;

  const int order = DiscreteSpaceType::polynomialOrder+1;
  GridFunctionType gridFunction( "exact solution", function_, gridPart_, order );

  Dune::L2Norm< GridPartType > l2norm( gridPart_ );
  Dune::H1Norm< GridPartType > h1norm( gridPart_ );

  std::vector< double > errors( 2 );
  errors[ 0 ] = l2norm.distance( gridFunction, solution );
  errors[ 1 ] = h1norm.distance( gridFunction, solution );

  Dune::FemEoc::setErrors( eocId_, errors );
}


inline Algorithm::DiscreteSpaceType &Algorithm::space ()
{
  return dfSpace_;
}



// Main Program
// ------------

int main ( int argc, char **argv )
try
{
  typedef Dune::GridSelector::GridType GridType;
  // initialize MPI manager
  Dune::MPIManager::initialize( argc, argv );
  // add command line parameters to global parameter table
  Dune::Parameter::append( argc, argv );
  // append parameters from the parameter file
  Dune::Parameter::append( (argc < 2) ? "parameter" : argv[ 1 ] );

  // create the grid
  Dune::GridPtr< GridType > gridptr = initialize< GridType >( std::string( "L2 projection" ) );

  Algorithm algorithm( *gridptr );
  compute( algorithm );

  Dune::Parameter::write( "parameter.log" );

  return 0;
}
catch( const Dune::Exception &exception )
{
  // display the exception message on the console
  std::cerr << exception << std::endl;
  return 1;
}
