#ifndef MASSOPERATOR_HH
#define MASSOPERATOR_HH

#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/operator/matrix/spmatrix.hh>
#include <dune/fem/operator/2order/lagrangematrixsetup.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>

template< class DiscreteFunction >
class MassOperator
: public Dune::Operator< typename DiscreteFunction::RangeFieldType,
                         typename DiscreteFunction::RangeFieldType,
                         DiscreteFunction, DiscreteFunction >
{
  typedef MassOperator< DiscreteFunction > ThisType;

  struct MatrixTraits;

public:
  typedef DiscreteFunction DiscreteFunctionType;
  typedef typename DiscreteFunctionType::DiscreteFunctionSpaceType
    DiscreteFunctionSpaceType;
  typedef Dune::DofManager< typename DiscreteFunctionSpaceType::GridType > DofManagerType;
  
  typedef Dune::CachingQuadrature< typename DiscreteFunctionSpaceType::GridPartType, 0 >
    QuadratureType;

  typedef Dune::SparseRowMatrixObject
    < DiscreteFunctionSpaceType, DiscreteFunctionSpaceType, MatrixTraits >
    SystemMatrixType;

  explicit MassOperator ( const DiscreteFunctionSpaceType &dfSpace )
  : dfSpace_( dfSpace ),
    dofManager_( DofManagerType::instance( dfSpace.grid() ) ),
    systemMatrix_( dfSpace, dfSpace ),
    sequence_( -1 )
  {}

  virtual void
  operator() ( const DiscreteFunctionType &u, DiscreteFunctionType &w ) const
  {
    systemMatrix().apply( u, w );
  }

  template< class Function >
  void assembleRHS( const Function &u, DiscreteFunctionType &w ) const;

  const SystemMatrixType &systemMatrix () const;

private:
  const DiscreteFunctionSpaceType &dfSpace_;
  const DofManagerType &dofManager_;
  mutable SystemMatrixType systemMatrix_;
  mutable int sequence_;
};


template< class DiscreteFunction >
struct MassOperator< DiscreteFunction >::MatrixTraits
{
  typedef DiscreteFunctionSpaceType RowSpaceType;
  typedef DiscreteFunctionSpaceType ColumnSpaceType;

  typedef Dune::LagrangeMatrixSetup< false > StencilType;

  typedef Dune::ParallelScalarProduct< DiscreteFunctionSpaceType >
    ParallelScalarProductType;
};


template< class DiscreteFunction >
template< class Function >
void MassOperator< DiscreteFunction >
  ::assembleRHS( const Function &u, DiscreteFunctionType &w ) const
{
  typedef typename DiscreteFunctionSpaceType::IteratorType IteratorType;
  typedef typename DiscreteFunctionSpaceType::BaseFunctionSetType BaseFunctionSetType;
  typedef typename DiscreteFunctionSpaceType::RangeFieldType FieldType;
  typedef typename DiscreteFunctionSpaceType::RangeType      RangeType;

  typedef typename DiscreteFunctionType::LocalFunctionType LocalFunctionType;
  typedef typename IteratorType::Entity EntityType;
  typedef typename EntityType::Geometry GeometryType;

  w.clear();

  // run over entities
  const IteratorType end = dfSpace_.end();
  for( IteratorType it = dfSpace_.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;
    const GeometryType &geometry = entity.geometry();

    LocalFunctionType localFunction = w.localFunction( entity );

    // run over quadrature points
    QuadratureType quadrature( entity, 2*dfSpace_.order()+1 );
    const unsigned int numQuadraturePoints = quadrature.nop();
    for( unsigned int qp = 0; qp < numQuadraturePoints; ++qp )
    {
      // evaluate u
      const typename QuadratureType::CoordinateType &x = quadrature.point( qp );        

      RangeType uValue;
      u.evaluate( geometry.global( x ), uValue );

      // put all things together and don't forget quadrature weights
      const FieldType weight = quadrature.weight( qp )*geometry.integrationElement( x );

      // apply weight 
      uValue *= weight; 

      // add to local function 
      localFunction.axpy( quadrature[ qp ], uValue );
    }
  }

  w.communicate();
}


template< class DiscreteFunction >
const typename MassOperator< DiscreteFunction >::SystemMatrixType &
MassOperator< DiscreteFunction >::systemMatrix () const
{
  typedef typename DiscreteFunctionSpaceType::IteratorType IteratorType;
  typedef typename DiscreteFunctionSpaceType::BaseFunctionSetType BaseFunctionSetType;
  typedef typename DiscreteFunctionSpaceType::RangeFieldType FieldType;

  typedef typename SystemMatrixType::LocalMatrixType LocalMatrixType;
  typedef typename IteratorType::Entity EntityType;
  typedef typename EntityType::Geometry GeometryType;

  if( sequence_ != dofManager_.sequence() )
  {
    systemMatrix_.reserve();
    systemMatrix_.clear();

    std::vector< typename DiscreteFunctionSpaceType::RangeType > values;

    // run over entities
    const IteratorType end = dfSpace_.end();
    for( IteratorType it = dfSpace_.begin(); it != end; ++it )
    {
      const EntityType &entity = *it;
      const GeometryType &geometry = entity.geometry();

      LocalMatrixType localMatrix = systemMatrix_.localMatrix( entity, entity );

      const BaseFunctionSetType &basis = localMatrix.domainBaseFunctionSet();
      const unsigned int numBasisFunctions = basis.size();
      values.resize( numBasisFunctions );

      // run over quadrature points
      QuadratureType quadrature( entity, 2*dfSpace_.order() );
      const unsigned int numQuadraturePoints = quadrature.nop();
      for( unsigned int qp = 0; qp < numQuadraturePoints; ++qp )
      {
        // evaluate base functions
        basis.evaluateAll( quadrature[ qp ], values );

        // get quadrature weight
        const typename QuadratureType::CoordinateType &x = quadrature.point( qp );
        const FieldType weight = quadrature.weight( qp )
                                    * geometry.integrationElement( x );

        // update system matrix
        for( unsigned int i = 0; i < numBasisFunctions; ++i )
        {
          // add diagonal entries
          localMatrix.add( i, i, weight * (values[ i ] * values[ i ]) );

          // add non-diagonal entries, use symmetric structure of the matrix
          for( unsigned int j = 0; j < i; ++j )
          {
            const FieldType value = weight * (values[ i ] * values[ j ]);
            localMatrix.add( i, j, value );
            localMatrix.add( j, i, value );
          }
        }
      }
    }
    sequence_ = dofManager_.sequence();
  }
  return systemMatrix_;
}

#endif
