// include host specific macros set by configure script   /*@LST0S@*/
#include <config.h>

// include std libs
#include <iostream>
#include <string>

// Dune includes
#include <dune/fem/space/common/arrays.hh>
#include <dune/fem/solver/timeprovider.hh>
#include <dune/fem/solver/odesolver.hh>

#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/io/parameter.hh>
#include <dune/fem/io/streams/asciistreams.hh>

#include "p.hh"


//============= EXECUTE ME FOR 1,2 and 3 respectively ========================
#define PRE 4 //1 = full, 2 = first, 3 = with correction
               //4 = H26C mechanism
//============================================================================



// a strange 1D diffusion which depends on the location, cf. eq. (33), p.251
template<class T> 
class StrangeDiffusion{
public:
  StrangeDiffusion(const T& D1, const T& e):D1_(D1),e_(e){}

  T operator()(const T& x){
    return D1_ + e_*x;
  }

private:
  const T& D1_;                //store only references
  const T& e_;
};




// approximation order
const int rkSteps = 3;

typedef Dune::StaticArray < double > DestinationType;
typedef Dune::MutableArray< double > MutableDestinationType;

#if PRE==1 //the full system, i.e. two species
int const dimRange = Constants::R_P_number_of_species*Constants::R_P_space; 
#endif

#if (PRE==2 || PRE==3)   //only one species involved
int const dimRange = Constants::R_P_space; //only one species to be considered
#endif 

#if PRE==4
int const dimRange = 6;  //the number of species in the H2C6 mechanism 
#endif


class RightHandSide : 
  public Dune::PARDGSpaceOperatorInterface< DestinationType >
{
protected:  
  double time_;
  mutable double dtEst_;
  const int size_;
public:  
  RightHandSide (const int size) 
    : time_( 0.0 ), 
      dtEst_( std::numeric_limits<double>::max() ),
      size_( size ) 
  {} 


  enum{
    numOfSpecies = Constants::R_P_number_of_species,
    space = Constants::R_P_space,
    domainDim = numOfSpecies*space
  };

  
  // return number of unknowns 
  int size() const { return size_; }

  void initializeTimeStepSize( const DestinationType& U0 ) const 
  {
    // initialize time step size 
    //assert( false );
    //abort();
    dtEst_ = 0.0012;               //DEFINE TIME STEP ESTIMATE
  }

  // identity 
  void operator() (const DestinationType& arg, DestinationType& dest) const
  {
    // implement right hand side 
    //dest = arg ;
    /*dest[0] = -0.04*arg[0] + 1e4*arg[1]*arg[2];
    dest[1] =  0.04*arg[0] - 1e4*arg[1]*arg[2] - Dune::SQR(arg[1]);
    dest[2] =  3e7*Dune::SQR(arg[1]);
    */

#if (PRE == 1 || PRE == 2 || PRE == 3)    
    const double h = 1./(space-1),   //spatially uniform step size
      a = Constants::R_P_a,
      b = Constants::R_P_b,
      c = Constants::R_P_c,
      d = Constants::R_P_d,
      e = Constants::R_P_e,
      D1 = Constants::R_P_D1,
      eps = Constants::R_P_eps;

    
    StrangeDiffusion<double> D2(D1,e);
#endif

    //apply method of lines to discretise spatial derivatives     
#if PRE==1    //FULL PROBLEM
    //Since the bdy values are constants, dy/dt = 0 at the boundary!!
    


    dest[0] = 0.;                     dest[space-1] = 0.; 
    dest[space] = 0.;                 dest[2*space-1] = 0.; 

    double z1_z2 = 0., x = 0.;
    for(int i = 1; i < space-1; ++i){
      z1_z2 = arg[space+i] - arg[i]/(1+a*arg[i]);

	x = i*h;

	dest[i] = c/eps*z1_z2 - d*arg[i] + D1*(arg[i+1] - 2*arg[i] + arg[i-1])/(Dune::SQR(h));

	dest[space+i] = -1./eps*z1_z2 - arg[i]/(Dune::SQR(1. + b*arg[i])) +   ( D2(x+0.5*h)*(arg[space+i+1] - arg[space+i]) - D2(x-0.5*h)*(arg[space+i] - arg[space+i-1]) )/(Dune::SQR(h));
      
    }
#endif 

#if PRE==2  //1st APPROXIMATION
    dest[0] = 0.; dest[space-1] = 0.;  

    for(int i = 1; i < space-1; ++i){
      dest[i] = -d*arg[i] + D1*(arg[i+1]- 2.*arg[i] + arg[i-1])/(Dune::SQR(h));
    }
#endif

#if PRE==3  //with CORRECTION TERM
   
    dest[0] = 0.; dest[space-1] = 0.;  

    double fprime = 0.,
      diffuse1 = 0.,        
      x = 0.;
    
    for(int i = 1; i < space-1; ++i){
      fprime = 1./(Dune::SQR(1 + a*arg[i]));
      
      diffuse1 = D1*(arg[i+1]- 2.*arg[i] + arg[i-1])/(Dune::SQR(h));
      
      x = i*h;
      
      dest[i] = -d*arg[i] +  diffuse1 +   //1st approx.
	c/(c*fprime + 1)*( d*arg[i]*fprime - arg[i]/(Dune::SQR(1+b*arg[i])) - fprime*diffuse1 + ( D2(x+0.5*h)*(arg[i+1]/(1.+a*arg[i+1]) - arg[i]/(1.+a*arg[i])) - D2(x-0.5*h)*(arg[i]/(1.+a*arg[i]) - arg[i-1]/(1.+a*arg[i-1])) )/(Dune::SQR(h)) ); //correction
      
    }
#endif


#if PRE==4
    const size_t nor = 12;   //number of reactions involved
    
    double pk[nor], //reaction rates 
      rV[nor];      //reaction velocities
    
    pk[0] = 2.0;
    pk[1] = 1.0;
    pk[2] = 1.0;
    pk[3] = 1000.0;
    pk[4] = 1000.0;
    pk[5] = 100.0;
    pk[6] = 216.0;
    pk[7] = 337.5;
    pk[8] = 1400.0;
    pk[9] = 10800.0;
    pk[10] = 33750.0;
    pk[11] = 0.7714;
    

    rV[0]  =  pk[0]*arg[0];
    rV[1]  =  pk[6]*arg[1]*arg[1];
    rV[2]  =  pk[1]*arg[2];
    rV[3]  =  pk[7]*arg[3]*arg[3];
    rV[4]  =  pk[2]*arg[4];
    rV[5]  =  pk[8]*arg[1]*arg[5];
    rV[6]  =  pk[3]*arg[0]*arg[3];
    rV[7]  =  pk[9]*arg[1]*arg[5];
    rV[8]  =  pk[4]*arg[2]*arg[1];
    rV[9]  =  pk[10]*arg[3]*arg[5];
    rV[10] =  pk[5]*arg[0]*arg[3];
    rV[11] =  pk[11]*arg[4];
   

     //the 6 species involved in the reaction
    // H2
    dest[0] = - rV[0]  + rV[1]
      - rV[6]  + rV[7]
      - rV[10] + rV[11];
    
    // H
    dest[1] = 2.0*rV[0] - 2.0*rV[1]
      + rV[4] - rV[5]
      + rV[6] - rV[7]
      - rV[8] + rV[9];
    
    // O2
    dest[2] = - rV[2] + rV[3]
      - rV[8] + rV[9];
    
    //  O
    dest[3] = + 2.0*rV[2] - 2.0*rV[3]
      - rV[6]     + rV[7]
      + rV[8]     - rV[9]
      - rV[10]    + rV[11];
    
    // H20
    dest[4] = - rV[4]  + rV[5]
      + rV[10] - rV[11];
    
    //  OH
    dest[5] = rV[4] - rV[5]
      + rV[6] - rV[7]
      + rV[8] - rV[9];
    

#endif


    dtEst_ = 0.0012;         //DEFINE TIME STEP ESTIMATE ONCE AGAIN!!
  }

  void operator () (const double* u, double *f ) const 
  {
    const DestinationType U (size_, u);
    DestinationType F (size_, f);
    this->operator()( U, F );
  }

  // store current time 
  void setTime( const double time )
  {
    time_ = time ;
  }

  // return time step size restriction for this problem 
  virtual double timeStepEstimate () const
  {
    // implement time step size restriction
    //assert( false );
    //abort();
    return dtEst_;
  }
};


/**
 * @brief main function for the LocalDG Advection-Diffusion application
 *
 * \c main starts the Simulation of an advection-diffusion pde with
 * the localdg method with EOC analysis and visual output to grape, paraview or
 * gnuplot.
 * \attention The localdg implementation uses the \c Dune::Pass
 * concept.
 *
 * @param argc number of arguments from command line
 * @param argv array of arguments from command line
 * @param envp array of environmental variables
 * @return 0 we don't program bugs. :)
 */
int main(int argc, char ** argv, char ** envp) {          /*@LST0S@*/

  /* Initialize MPI (always do this even if you are not using MPI) */
  Dune::MPIManager :: initialize( argc, argv );

  try {

    // *** Initialization
    Dune::Parameter::append(argc,argv);                           /*@\label{dg:param0}@*/
    if (argc == 2) {
      Dune::Parameter::append(argv[1]);
    } 
    else 
    {
      Dune::Parameter::append("parameter");                       /*@\label{dg:paramfile}@*/
    }                                                       /*@\label{dg:param1}@*/

    Dune::TimeProvider< Dune::DefaultCollectiveCommunicationType >  tp;

    // implementation of right hand side 
    RightHandSide spaceOperator( dimRange ); 

    // The ODE Solvers                                                         /*@LST1S@*/
    typedef DuneODE :: OdeSolverInterface< DestinationType >      OdeSolverInterfaceType;
    typedef DuneODE :: ImplicitOdeSolver < DestinationType >      ImplicitOdeSolverType; /*@\label{dg:}@*/
    typedef DuneODE :: ExplicitOdeSolver < DestinationType >      ExplicitOdeSolverType; /*@\label{}@*/

    // choice of explicit or implicit ode solver
    static const std::string odeSolverStr[]
      = { "explicit", "implicit" };
    int odeSolve = Dune::Parameter::getEnum( "femhowto.odesolver", odeSolverStr, 0 );

    OdeSolverInterfaceType* odeSolver = 0;  

    // create implicit or explicit ode solver
    if( odeSolve == 0 )
      odeSolver = new ExplicitOdeSolverType(spaceOperator, tp, rkSteps);   /*@\label{dg:odeexpldef}@*/
    else
      odeSolver = new ImplicitOdeSolverType(spaceOperator, tp, rkSteps);   /*@\label{dg:odeimpldef}@*/

    assert( odeSolver );

    MutableDestinationType U( dimRange );

    // ============= initial conditions ===============================
    U.clear();
    
#if (PRE==1 || PRE==2 || PRE==3)
    const int space = Constants::R_P_space;
    const double hx = 1./(space-1);
#endif

#if PRE==1  //FULL SYSTEM, cf. p. 251 left column, last block
    const double a = Constants::R_P_a;

    
    for(int i = 0; i < space; ++i){
      U[i] = i*hx;
      
      U[space+i] = U[i]/(1. + a*U[i]);
    }
#endif
    

#if (PRE==2 || PRE==3)  //reduced system for 1st and correction, resp.
    for(int i = 0; i < space; ++i){
      U[i] = i*hx;
    }
#endif


#if PRE==4  //taken from "JOCHEN/H2C6/START/x_init.dat"
    U[0] = 3.1964e-01;
    U[1] = 4.9003e-02;
    U[2] = 1.5940e-01;
    U[3] = 1.9500e-02;
    U[4] = 6.5000e-01;
    U[5] = 1.1708e-02;
#endif

    /* U[ 0 ] = 1;
    //U[ 1 ] = 0;
    //U[ 2 ] = 0;
    /// ......
    */
    //=================================================================

    // ode.initialize applies the DG Operator once to get an estimate for dt.
    odeSolver->initialize( U );                               /*@\label{dg:odeinit}@*/

    
    const double T = Dune::Parameter :: getValue< double > ("femhowto.endTime");
    // do time loop
   
    std::ofstream horizon("Horizon.dat", std::ios_base::out);
    horizon << "#Time t               Vector U(t)"<<std::endl;

    for( tp.init() ; tp.time() < T ; tp.next() ) //solve over time interval
    {
      odeSolver->solve( U );   //U will be overwritten to hold the solution
      std::cout << "Time = " << tp.time() << "  dt = " << tp.deltaT() << std::endl;
      horizon << " "<<tp.time() << "     ";
      for( size_t i = 0; i < U.size(); ++i )
	horizon << U[ i ] << "  ";
      horizon << std::endl;

    }//end loop over time intervall
    horizon.close();

    // write vector U (solution at final time t = T) to file out.txt 
    Dune::ASCIIOutStream out ( "out.txt" );
    U.write( out );

#if PRE==1
    std::cout << std::endl<< "full system:"<<std::endl;
    std::cout <<             "------------"<<std::endl;
#elif PRE==2
    std::cout << std::endl<< "1st approximation:"<<std::endl;
    std::cout <<             "------------------"<<std::endl;
#elif PRE==3
    std::cout << std::endl<< "with correction:"<<std::endl;
    std::cout <<             "----------------"<<std::endl;
#elif PRE==4
    std::cout << std::endl<< "H2C6 mechanism:"<<std::endl;
    std::cout <<             "---------------"<<std::endl;
#else 
    std::cout << std::endl<< "Nothing defined so far, Marc."<<std::endl;
#endif 

    std::cout<< std::endl<< "Solution U(t = "<<T<<") = "  <<std::endl;

    for( size_t i = 0; i < U.size(); ++i )
      std::cout << U[ i ] << " ";
    std::cout << std::endl; 

    //my output file
    std::string filename;

#if PRE==1 
    filename = "Full.dat";
#endif
    
#if PRE==2
    filename = "1stApprox.dat";
#endif

#if PRE==3
    filename = "withCorrection.dat";
#endif


#if(PRE==1 || PRE==2 || PRE==3)
    std::ofstream tofile(filename.c_str(), std::ios_base::out);

    for(int i = 0; i < space; ++i){
      tofile  << " "<< i*hx << "    "<<U[i] << std::endl;// "    "<<U[i+space] <<std::endl;
    }
    tofile.close();
#endif


    /*#if PRE==4
    filename= "H2.dat";
     std::ofstream tof(filename.c_str(), std::ios_base::out);
     
     tof.close();
#endif
    */

    //close file 
    Dune::Parameter::write("parameter.log");

    delete odeSolver;
  }
  catch (Dune::Exception &e) {                            /*@\label{dg:catch0}@*/
    std::cerr << e << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Generic exception!" << std::endl;
    return 2;
  }                                                      /*@\label{dg:catch1}@*/

  return 0;
}

