#include <config.h>

#include <iostream>
#include <string>

#include <dune/fem/gridpart/common/gridpart.hh>
#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/operator/matrix/spmatrix.hh>
#include <dune/fem/operator/matrix/blockmatrix.hh>
#include <dune/fem/operator/matrix/ontheflymatrix.hh>
#include <dune/fem/operator/matrix/istlmatrix.hh>
#include <dune/fem/solver/inverseoperators.hh>
#include <dune/fem/solver/oemsolver/oemsolver.hh>
#include <dune/fem/solver/istlsolver.hh>
#include <dune/fem/space/common/restrictprolonginterface.hh>

#include <dune/fem-howto/baseevolution.hh>

#include <dune/fem/misc/l2norm.hh>
#include <dune/fem/misc/h1norm.hh>

#include "problemdata.hh"
#include "laplaceoperator.hh"

#ifndef POLORDER 
const int polynomialOrder = 1;
#else 
const int polynomialOrder = POLORDER;
#endif

namespace Dune
{

  template< class DiscreteFunction, class MatrixTraits >
  class LaplaceFEOp;

  struct AlgorithmDataOutputParameters
  : public Dune::LocalParameter< Dune::DataOutputParameters, AlgorithmDataOutputParameters >
  {
    explicit AlgorithmDataOutputParameters ( int loop )
    {
      std::stringstream ss;
      ss << "iter" << loop << "/";
      loop_ = ss.str();
    }

    AlgorithmDataOutputParameters ( const AlgorithmDataOutputParameters &other )
    : loop_( other.loop_ )
    {}

    std::string path () const { return loop_; }

  private:
    std::string loop_;
  };

  template< class GridImp, class ProblemImp, int polOrder >
  struct OptimalControlStepperTraits
  {
    // grid type
    typedef GridImp GridType;

    // grid part type
    typedef Dune::LeafGridPart<GridType> GridPartType;

    // problem type
    typedef ProblemImp ProblemType;

    // function space type
    typedef Dune::FunctionSpace< double, double, GridType::dimensionworld, 1 > 
      FunctionSpaceType;

    // discrete space type
    typedef Dune::LagrangeDiscreteFunctionSpace
            < FunctionSpaceType, GridPartType, polOrder, Dune::CachingStorage >
      DiscreteSpaceType;

    // discrete function type
    typedef Dune::AdaptiveDiscreteFunction< DiscreteSpaceType > DiscreteFunctionType;

    // type of restriction/prolongation projection for adaptive simulations
    typedef Dune::RestrictProlongDefault< DiscreteFunctionType > RestrictionProlongationType;

    // time provider type
    typedef Dune::TimeProviderBase TimeProviderType;
      
    // io tuple
    typedef Dune::tuple< DiscreteFunctionType *, DiscreteFunctionType *, DiscreteFunctionType *, DiscreteFunctionType * >
      IOTupleType;

    // data writer type
    typedef Dune::DataOutput< GridType, IOTupleType > DataWriterType;

  }; // end struct OptimalControlStepperTraits


  template< class GridImp, class ProblemImp, int polOrder >
  struct OptimalControlStepper
  : public AlgorithmBase< OptimalControlStepperTraits< GridImp, ProblemImp, polOrder > >
  {
    // traits class
    typedef OptimalControlStepperTraits< GridImp, ProblemImp, polOrder > Traits;

    // base class
    typedef AlgorithmBase< Traits > BaseType;
   
    // get types
    typedef typename Traits::ProblemType ProblemType;
    typedef typename Traits::GridType GridType;
    typedef typename Traits::GridPartType GridPartType;
    typedef typename Traits::FunctionSpaceType FunctionSpaceType;
    typedef typename Traits::DiscreteSpaceType DiscreteSpaceType;
    typedef typename Traits::DiscreteFunctionType DiscreteFunctionType;
    typedef typename BaseType::TimeProviderType TimeProviderType;

  private:
    typedef Dune::Fem::SparseRowMatrixTraits < DiscreteSpaceType, DiscreteSpaceType >
      MatrixObjectTraits;

    typedef Dune::LaplaceFEOp< DiscreteFunctionType, MatrixObjectTraits >
      LaplaceOperatorType;

    typedef typename ProblemType::ExactSolutionType ExactSolutionType;

    typedef Dune::Fem::GridFunctionAdapter< ExactSolutionType, GridPartType >
      GridExactSolutionType;

    typedef Dune::Fem::CGInverseOperator< DiscreteFunctionType, LaplaceOperatorType >
      InverseOperatorType;

  public:
    OptimalControlStepper ( GridType & grid, const ProblemType & problem )
    : BaseType( grid ),
    laplace_( space_, problem ),
    y_( "y", space_ ),
    p_( "p", space_ ),
    v_( "v", space_ ),
    w_( "w", space_ ),
    rhs_( "rhs", space_ ),
    dataTup_(&y_,&p_,&v_,&w_),
    dataOutput_(0),
    loop_( 0 ),
    solverEpsilon_( Dune::Parameter::getValue( "femhowto.solvereps", 1.0e-8 ) )
    {
      // add entries to eoc calculation 
      std::vector<std::string> femEocHeaders;

      // we want to calculate L2 and H1 error (and EOC)
      femEocHeaders.push_back("$L^2$-error");
      femEocHeaders.push_back("$H^1$-error");

      // get eoc id 
      eocId_ = Dune::FemEoc::addEntry(femEocHeaders);

      // distribute grid (only for parallel runs, in serial nothing is done here)
      grid.loadBalance();
    }

    ~OptimalControlStepper ()
    {
      delete dataOutput_;
    }
    
    //! initialize method for time loop, i.e. L2-project initial data 
    void initializeStep ( TimeProviderType & tp, DiscreteFunctionType & u )
    {
      delete dataOutput_;
      dataOutput_ = new typename Traits::DataWriterType(grid_,dataTup_,tp,AlgorithmDataOutputParameters(loop_));
      ++loop_;
      tp.provideTimeStepEstimate( 1 );

      u.clear();
      y_.clear();
      v_.clear();
      w_.clear();
      p_.clear();

      // ExactSolutionType uexact( laplace_.problem() ); 
      // Dune::LagrangeInterpolation< DiscreteFunctionType >
      //  :: interpolateFunction( uexact, u );

      RHSPrimalLocal< true > localf( u, laplace_.problem() );
      Dune::LocalFunctionAdapter< RHSPrimalLocal< true > > f( "primal", localf, space_.gridPart() );
      rightHandSide( f );
      rhs_.dofsValid();
      solve( y_ );
      y_.dofsValid();
    }

    //! solve one time step 
    void step ( TimeProviderType & tp, DiscreteFunctionType & u )
    {
      tp.provideTimeStepEstimate( 1 );
      {
        RHSDualLocal localf(y_,laplace_.problem());
        Dune::LocalFunctionAdapter< RHSDualLocal > f( "dual", localf, space_.gridPart() );
        rightHandSide( f );
        rhs_.dofsValid();
        solve( p_ );
        p_.dofsValid();
      }

      direction( u, laplace_.problem() );
      v_.dofsValid();

      {
        RHSPrimalLocal<false> localf(v_,laplace_.problem());
        Dune::LocalFunctionAdapter< RHSPrimalLocal<false> > f("primal",localf,space_.gridPart());
        rightHandSide( f );
        rhs_.dofsValid();
        solve( w_ );
        w_.dofsValid();
      }

      dataOutput_->write(tp);

      double s = stepSize( u, laplace_.problem() );

      u.addScaled( v_, s );
      y_.addScaled( w_, s );
    }
    
    //! finalize problem, i.e. calculate EOC ...
    void finalizeStep ( TimeProviderType & tp, DiscreteFunctionType & solution )
    {
      // create exact solution 
      ExactSolutionType uexact( laplace_.problem() );

      // create grid function adapter 
      GridExactSolutionType ugrid( "exact solution", uexact, gridPart_, DiscreteSpaceType::polynomialOrder + 1 );

      // create L2 - Norm 
      Dune::L2Norm< GridPartType > l2norm( gridPart_ );
      // calculate L2 - Norm 
      const double l2error = l2norm.distance( ugrid, solution );      /*@LST0E@*/

      // create H1 - Norm 
      Dune::H1Norm< GridPartType > h1norm( gridPart_ );
      // calculate H1 - Norm 
      const double h1error = h1norm.distance( ugrid, solution );

      // store values 
      std::vector<double> errors;
      errors.push_back( l2error );
      errors.push_back( h1error );

      // submit error to the FEM EOC calculator 
      Dune::FemEoc::setErrors( eocId_, errors );
    }

#if 0
    //! restore all data from check point (overload to do something)
    void restoreFromCheckPoint( TimeProviderType & tp, DiscreteFunctionType & solution )
    { 

    }
        
    //! write a check point (overload to do something)
    void writeCheckPoint ( TimeProviderType & tp ) const 
    {
      
    }

    //! default time loop implementation, overload for changes 
    void operator () ( DiscreteFunctionType & solution )
    {

    }

    //! finalize problem, i.e. calculated EOC ...
    void finalize ( DiscreteFunctionType & solution )
    {

    }
#endif

  protected:
    using BaseType::grid_;
    using BaseType::gridPart_;
    using BaseType::space_;

  private:
    LaplaceOperatorType laplace_;
    DiscreteFunctionType y_,p_,v_,w_,rhs_;
    typename Traits::IOTupleType dataTup_;
    typename Traits::DataWriterType *dataOutput_;
    int eocId_;
    int loop_;
    double solverEpsilon_;

private:              
    void direction( const DiscreteFunctionType &u, const ProblemType &problem )
    {
      const double eps = 0.0001;
      typedef typename DiscreteFunctionType::ConstDofIteratorType ConstDofIteratorType;
      ConstDofIteratorType uIt = u.dbegin();
      typedef typename DiscreteFunctionType::DofIteratorType DofIteratorType;
      DofIteratorType vEnd = v_.dend();
      for ( DofIteratorType vIt = v_.dbegin(),
            pIt = p_.dbegin();
            vIt != vEnd; ++vIt, ++pIt, ++uIt )
      {
        double d = problem.lambda()*(*uIt) + problem.beta()*(*pIt);
        //*vIt = (d>=0)?problem.ua():problem.ub();
        *vIt = 0.5*(tanh(d/eps) + 1.)*(problem.ua()-problem.ub())+problem.ub();
        *vIt -= *uIt;
      }
    }

    double stepSize( const DiscreteFunctionType &u, const ProblemType &problem )
    {
      typedef typename DiscreteSpaceType::IteratorType IteratorType;
      typedef typename DiscreteFunctionType::LocalFunctionType LocalFunctionType;
      typedef typename LocalFunctionType :: RangeType RangeType;
      typedef typename IteratorType::Entity EntityType;
      typedef typename EntityType::Geometry GeometryType;

      double s0=0;
      double s1=0;
      double s2=0;

      const IteratorType end = space_.end();
      for( IteratorType it = space_.begin(); it != end; ++it )
      {
        const EntityType &entity = *it;
        const GeometryType &geometry = entity.geometry();

        LocalFunctionType localv = v_.localFunction( entity );
        LocalFunctionType localw = w_.localFunction( entity );
        LocalFunctionType localu = u.localFunction( entity );
        LocalFunctionType localy = y_.localFunction( entity );

        typedef Dune::CachingQuadrature< typename DiscreteSpaceType::GridPartType, 0 >
                      QuadratureType;
        QuadratureType quad( entity, 2*space_.order()+1 );
        const unsigned int numQuadraturePoints = quad.nop();
        for( unsigned int qp = 0; qp < numQuadraturePoints; ++qp )
        {
          RangeType ly,lw,lu,lv;
          const typename QuadratureType::CoordinateType &x = quad.point( qp );
          localy.evaluate( quad[qp], ly );
          localw.evaluate( quad[qp], lw );
          localu.evaluate( quad[qp], lu );
          localv.evaluate( quad[qp], lv );
          ly -= problem.yOmega( geometry.global(x) );
          double val0 = ly*ly + problem.lambda()*lu*lu;
          val0 *= quad.weight(qp)*geometry.integrationElement(x);
          s0 += val0;
          double val1 = lw*ly + lv*lu*problem.lambda();
          val1 *= quad.weight(qp)*geometry.integrationElement(x);
          s1 += val1;
          double val2 = lw*lw + lv*lv*problem.lambda();
          val2 *= quad.weight(qp)*geometry.integrationElement(x);
          s2 += val2;
        }
      }
      double s = std::max(0.,std::min(-s1/s2,1.));
      if (s2<1e-10) 
        s=1;
      std::cout << "s0: " << 0.5*s0 << " s1: " << s1 << " s2: " << 0.5*s2 
                << " g(s) = " << 0.5*s0+s1*s+0.5*s2*s*s
                << std::endl;
      return s;
    }

    template <bool source>
    struct RHSPrimalLocal
    {
      typedef typename GridType::Traits::template Codim<0>::Entity EntityType;
      typedef typename OptimalControlStepper::FunctionSpaceType FunctionSpaceType;
      typedef typename OptimalControlStepper::GridPartType GridPartType;
      typedef typename DiscreteFunctionType :: LocalFunctionType LocalFunctionType;
      RHSPrimalLocal( const DiscreteFunctionType &v, const ProblemType &problem )
      : v_(v), problem_(problem),
        localv_(v)
      {}
      template <class PointType>
      void evaluate(const PointType& x, typename DiscreteFunctionType::RangeType& val)
      {
        localv_.evaluate(x,val);
        val[0] *= problem_.beta();
        // val[0] = problem_.control(localv_.entity().geometry().global( coordinate(x)));
        if (source) 
          val[0] += problem_.f(localv_.entity().geometry().global( coordinate(x) ));
      }
      void init(const EntityType& en)
      {
        localv_.init(en);
      }
      private:
      const DiscreteFunctionType &v_;
      const ProblemType &problem_;
      LocalFunctionType localv_;
    };

    struct RHSDualLocal
    {
      typedef typename GridType::Traits::template Codim<0>::Entity EntityType;
      typedef typename OptimalControlStepper::FunctionSpaceType FunctionSpaceType;
      typedef typename OptimalControlStepper::GridPartType GridPartType;
      typedef typename DiscreteFunctionType :: LocalFunctionType LocalFunctionType;
      RHSDualLocal( const DiscreteFunctionType &y, const ProblemType &problem )
      : y_(y), problem_(problem),
        localy_(y)
      {}
      template< class PointType >
      void evaluate(const PointType& x, typename DiscreteFunctionType::RangeType& val)
      {
        localy_.evaluate(x,val);
        val[0] -= problem_.yOmega(localy_.entity().geometry().global( coordinate(x) ));
      }
      void init(const EntityType& en)
      {
        localy_.init(en);
      }
      private:
      const DiscreteFunctionType &y_;
      const ProblemType &problem_;
      LocalFunctionType localy_;
    };

    template <class RHS>
    void rightHandSide( const RHS &f )
    {
      // initialize as zero 
      rhs_.clear();

      // setup right hand side 
      const int quadratureOrder = 2 * space_.order() + 1;
      Dune::RightHandSideAssembler< DiscreteFunctionType >
        ::assemble ( f , rhs_ , quadratureOrder ); 

      // set Dirichlet Boundary to exact solution  
      typedef typename DiscreteSpaceType::IteratorType IteratorType;
      typedef typename IteratorType::Entity EntityType;
      const IteratorType endit = space_.end();
      for( IteratorType it = space_.begin(); it != endit; ++it )
      {
        const EntityType &entity = *it;
        // in entity has intersections with the boundary adjust dirichlet nodes 
        if( entity.hasBoundaryIntersections() )
          boundaryTreatment( entity );
      }                                                               /*@\label{poi:rhsInit1}@*/
    }
    //! set the dirichlet points to exact values
    template< class EntityType >
    void boundaryTreatment( const EntityType &entity )
    {                                                                 /*@LST0E@*/
      typedef typename DiscreteSpaceType :: GridPartType :: IntersectionIteratorType
        IntersectionIteratorType;

      typedef typename DiscreteSpaceType :: LagrangePointSetType
        LagrangePointSetType;
      typedef typename LagrangePointSetType
        :: template Codim< 1 > :: SubEntityIteratorType
        FaceDofIteratorType;

      // get local functions of data 
      typename DiscreteFunctionType :: LocalFunctionType rhsLocal = rhs_.localFunction( entity );

      const LagrangePointSetType &lagrangePointSet = space_.lagrangePointSet( entity );

      IntersectionIteratorType it = space_.gridPart().ibegin( entity );
      const IntersectionIteratorType endit = space_.gridPart().iend( entity );
      for( ; it != endit; ++it )
      {
        typedef typename IntersectionIteratorType :: Intersection IntersectionType;
        const IntersectionType& intersection = *it;

        // if intersection is with boundary, adjust data  
        if( intersection.boundary() )
        {
          // get face number of boundary intersection 
          const int face = intersection.indexInInside();
          // get dof iterators 
          FaceDofIteratorType faceIt
            = lagrangePointSet.template beginSubEntity< 1 >( face );
          const FaceDofIteratorType faceEndIt
            = lagrangePointSet.template endSubEntity< 1 >( face );
          for( ; faceIt != faceEndIt; ++faceIt )
            rhsLocal[ *faceIt ] = 0;
        }
      }
    }                                                                 /*@LST0S@*/

    //! solve the resulting linear system
    void solve ( DiscreteFunctionType &solution )
    {                                                                /*@LST0E@*/
      // create timer (also stops time) 
      Dune::Timer timer;

      // solve the linear system (with CG)
      const double dummy = 12345.67890;

      // get verbosity information from Parameter class 
      const bool verbose = Dune::Parameter :: verbose();

      // create inverse operator                                     /*@LST0S@*/
      InverseOperatorType cg( laplace_, dummy, solverEpsilon_, 20000, verbose );/*@\label{poi:cgInit}@*/

      // solve the system 
      cg( rhs_, solution );                                            /*@\label{poi:cgExec}@*//*@LST0E@*/

      // get elapsed time 
      const double solveTime = timer.elapsed();

      // output in only in verbose mode (parameter fem.verbose)
      if( verbose ) 
        std :: cout << "Time needed by solver: " << solveTime << "s" << std :: endl;
    }                                                                /*@LST0S@*/

  }; // end struct OptimalControlStepper

} // end namespace Dune

int main ( int argc, char **argv )
{
  // initialize MPI 
  Dune::MPIManager::initialize( argc, argv );
  const int rank = Dune::MPIManager::rank ();
  
  try
  {
    // grid type
    typedef Dune::GridSelector::GridType GridType;

    // type of problem
    typedef Dune::ProblemInterface< Dune::FunctionSpace< double, double, Dune::GridSelector::dimworld, 1 > > ProblemType;

    // append parameters from the comand line 
    Dune::Parameter::append( argc, argv );

    // append parameters from the parameter file  
    Dune::Parameter::append( (argc < 2) ? "parameter" : argv[ 1 ] );

    // generate GridPointer holding grid instance
    Dune::GridPtr< GridType > gridptr = initialize< GridType >(std::string("Optimal control"));

    // get grid reference 
    GridType& grid = *gridptr;
      
    // create problem 
    ProblemType* problem = Dune::createProblem< GridType > ();
    assert( problem );

    // create stepper class 
    typedef Dune::OptimalControlStepper< GridType, ProblemType, polynomialOrder > StepperType;
    StepperType stepper( grid, *problem );

    //Algorithm< GridType, polynomialOrder > algorithm(grid, *problem);
    // compute solution
    compute( stepper );

    // write parameter logfile 
    Dune::Parameter :: write("parameter.log");

    // remove problem 
    delete problem;

    return 0;
  }
  catch( const Dune::Exception &exception )
  {
    if( rank == 0 )
      std :: cerr << exception << std :: endl;
    return 1;
  }
}                                                                   /*@LST0E@*/
