#ifndef DUNE_FEMHOWTO_EXAMPLES_OPTIMALCONTROL_PROBLEMDATA_HH
#define DUNE_FEMHOWTO_EXAMPLES_OPTIMALCONTROL_PROBLEMDATA_HH

#include <cassert>
#include <cmath>

#include <dune/common/array.hh>

#include <dune/fem/space/common/functionspace.hh>
#include <dune/fem/io/parameter.hh>

// #include <dune/fem-howto/probleminterfaces.hh>
#include "probleminterfaces.hh"

namespace Dune
{
  //! sinus product problem 
  template< class Grid >
  class BangBangProblem 
  : public ProblemInterface<FunctionSpace< double, double, Grid::dimensionworld, 1 > >
  {
  public:
    typedef FunctionSpace< double, double, Grid::dimensionworld, 1 > FunctionSpaceType;
    typedef ProblemInterface< FunctionSpaceType > BaseType;

    enum { dimDomain = FunctionSpaceType::dimDomain };

    typedef typename BaseType::DomainType DomainType;
    typedef typename BaseType::RangeType  RangeType;
    typedef typename BaseType::JacobianRangeType  JacobianRangeType;
    typedef typename BaseType::DomainFieldType DomainFieldType;
    typedef typename BaseType::RangeFieldType  RangeFieldType;

    typedef typename BaseType::DiffusionMatrixType DiffusionMatrixType;

  public:
    BangBangProblem ()
    : lambda_( Parameter::getValue( "femhowto.lambda", 0.0 ) )
    {}

    double f ( const DomainType &x ) const
    {
      double f = dimDomain * (M_PI*M_PI);
      for( int i = 0; i < dimDomain; ++i )
        f *= sin( M_PI * x[ i ] );
      return f - control( x );
    }

    double yOmega( const DomainType &x ) const 
    {
      double y = 1.0;
      for( int i = 0; i < dimDomain; ++i )
        y *= sin( M_PI * x[ i ] );
      return y + laplace_p( x );
    }

    double p ( const DomainType &x ) const
    {
      double ret = -1.0 / (dimDomain * 64 * (M_PI*M_PI));
      for( int i = 0; i < dimDomain; ++i )
        ret *= sin( 8 * M_PI * x[ i ] );
      return ret;
    }

    double laplace_p ( const DomainType &x ) const
    {
      double ret = 1.0;
      for( int i = 0; i < dimDomain; ++i )
        ret *= sin( 8 * M_PI * x[ i ] );
      return ret;
    }

    double control ( const DomainType &x ) const
    {
      const double beta_p = beta() * p( x );
      if( lambda() > 0.0 )
        return std::max( ua(), std::min( ub(), -beta_p / lambda() ) );
      else
        return (beta_p >= 0.0 ? ua() : ub());
    }

    double beta () const
    {
      return 1.0;
    }

    double lambda () const
    {
      return lambda_;
    }

    double ua () const
    {
      return -1.0;
    }

    double ub () const
    {
      return 1.0;
    }

    //! the exact solution 
    void u ( const DomainType &x, RangeType &ret ) const
    {
      ret[ 0 ] = control( x );
    }

    //! the diffusion matrix 
    void K(const DomainType& x, DiffusionMatrixType& m) const 
    {
      m = 0;
      for( int i = 0; i < dimDomain; ++i )
        m[ i ][ i ] = 1;
    }

    bool constantK() const
    {
      return true;
    }

    //! the gradient of the exact solution 
    virtual void gradient( const DomainType &x, JacobianRangeType &grad ) const
    {
      grad = 0;
    }

  private:
    double lambda_;
  };


  template <class GridImp> 
  static ProblemInterface<FunctionSpace< double, double, GridImp :: dimensionworld, 1 > >* 
  createProblem()
  {
    int problem = 0; // default value 
    const std::string problemTable[] = { "bangbang" };
    // read problem from parameter file 
    problem =  Parameter::getEnum( "femhowto.poissonproblem", problemTable, problem );

    // see problem.hh for available problems 
    switch (problem)
    {
      case 0: return new Dune::BangBangProblem< GridImp > ();
      default: std::cerr << "Wrong problem value, bye, bye!" << std::endl;
               abort();
    }
    return 0;
  }

}

#endif
