#ifndef  DUNE_PROBLEM_HH__
#define  DUNE_PROBLEM_HH__


/**
 * @brief describes the initial and exact solution of the advection-diffusion model
 *
 * \f[u(x,y,z,t):=\displaystyle{\sum_{i=0}^{1}} v_i(t) \cdot \mu_i(x) \cdot
 * \nu_i(y) \cdot \omega_i(z)\f]
 *
 * with
 *
 * \f{eqnarray*}{
 * v_0(t)&:=&e^{-\varepsilon t \pi^2 (2^2 + 1^2 + 1.3^2 )} \\
 * \mu_0(x)&:=&0.6\cdot \cos(2\pi (x-at)) + 0.8\cdot \sin(2\pi (x-at)) \\
 * \nu_0(y)&:=&1.2\cdot \cos(1\pi (y-at)) + 0.4\cdot \sin(1\pi (y-at)) \\
 * v_1(t)&:=&e^{-\varepsilon t \pi^2 (0.7^2 + 0.5^2 + 0.1^2 )} \\
 * \mu_1(x)&:=&0.9\cdot \cos(0.7\pi (x-at)) + 0.2\cdot \sin(0.7\pi (x-at)) \\
 * \nu_1(y)&:=&0.3\cdot \cos(0.5\pi (y-at)) + 0.1\cdot \sin(0.5\pi (y-at))
 * \f}
 *
 * This is a solution of the AdvectionDiffusionModel for \f$g_D = u|_{\partial
 * \Omega}\f$.
 *
 */
template <class GridType, class Field = double >
class U0 : public Dune::Fem::Function<
            Dune::FunctionSpace<typename GridType :: ctype, Field,
                                GridType::dimensionworld, 1>, U0<GridType, Field> > {
 public:
  enum { ConstantVelocity = true };
  enum { dimDomain = GridType::dimensionworld };  
  enum { dimRange = 1 };  
  typedef typename GridType :: ctype DomainFieldType;
  typedef Field RangeFieldType;

  typedef Dune::FieldVector<DomainFieldType,dimDomain> DomainType;
  typedef Dune::FieldVector<RangeFieldType ,dimRange>  RangeType;


  /**
   * @brief define problem parameters
   */
  U0():
    velocity_(0.8),
    startTime_(Dune::Parameter::getValue<double>("femhowto.startTime",0.0))
  {
    velocity_[0]=1.25;
  }

  /**
   * @brief getter for the velocity
   */
  const DomainType &velocity() const {
    return velocity_;
  }

  /**
   * @brief the velocity method
   */
  void velocity(const DomainType& x, DomainType& velocity ) const {
    velocity = velocity_;
  }
  
  /**
   * @brief evaluates \f$ u_0(x) \f$
   */
  void evaluate(const DomainType& arg, RangeType& res) const {             /*@LST0S@@LST0E@*/
    evaluate(arg, startTime_, res);
  }
  
  /**
   * @brief old version of the exact solution
   * 
   * old version of evaluate(const DomainType& arg, double t, RangeType& res),
   * which is still needed by the DataWriter
   */
  inline void evaluate(double t,  const DomainType& arg, RangeType& res) const { 
    evaluate(arg, t, res);
  }
    
  /**
   * @brief evaluate exact solution
   */ 
  void evaluate(const DomainType& arg, double t, RangeType& res) const {  /*@LST0S@@LST0E@*/
    DomainType x(arg);
    x.axpy( -t, velocity_ );
    res[0] = sin ( 2. * M_PI * (x*x) );
  }
  
  /**
   * @brief latex output for EocOutput
   */
  std::string description()
  {
    std::ostringstream ofs;
    ofs << "Problem: " << myName 
        << " Velocity: " << velocity_
        << "\n\n";
    return ofs.str();
  }
 private:    
  DomainType velocity_;
  double startTime_;
 public:
  std::string myName;
};

#endif  /*DUNE_PROBLEM_HH__*/
