// Question: should we introduce Parameter class in this example?

// include configurations options

#include <config.h>

// Includes from the IOStream Library
// ----------------------------------

#include <iostream>
#include <sstream>

// Includes from DUNE core modules
// ----------------------

#include <dune/common/misc.hh>
#include <dune/common/fvector.hh>

// Includes from DUNE-FEM
// ----------------------

// include basic grid parts
#include <dune/fem/gridpart/common/gridpart.hh>

// include Lagrange discrete function space
#include <dune/fem/space/lagrangespace.hh>

#include <dune/fem/function/adaptivefunction.hh>

#include <dune/fem/operator/lagrangeinterpolation.hh>

#include <dune/fem/misc/l2norm.hh>
#include <dune/fem/misc/h1norm.hh>

#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/misc/utility.hh>
#include <dune/fem/io/file/dataoutput.hh>

#include <dune/fem/space/common/adaptmanager.hh>

#include <dune/fem/io/parameter.hh>

// GridType 
// --------
typedef Dune::GridSelector::GridType HGridType;

// Global Type Definitions
// -----------------------
#if defined POLORDER
const int polOrder = POLORDER;
#else
const int polOrder = 1;
#endif

// Function to Project
// -------------------

template< class FunctionSpace >
struct Function
{
  typedef FunctionSpace FunctionSpaceType;

  typedef typename FunctionSpaceType::DomainFieldType DomainFieldType;
  typedef typename FunctionSpaceType::RangeFieldType RangeFieldType;

  typedef typename FunctionSpaceType::DomainType DomainType;
  typedef typename FunctionSpaceType::RangeType RangeType;
  typedef typename FunctionSpaceType::JacobianRangeType JacobianRangeType;

  void evaluate ( const DomainType &x, RangeType &value ) const
  {
    value = 1.;
    for( int i = 0; i < FunctionSpace::dimDomain; ++i )
      value *= sin( M_PI * x[ i ] );
  }

  void jacobian ( const DomainType &x, JacobianRangeType &jacobian ) const
  {
    jacobian[0] = M_PI;
    for( int i = 0; i < FunctionSpace::dimDomain; ++i )
    {
      double y = M_PI * x[ i ];
      for( int j = 0; j < FunctionSpace::dimDomain; ++j )
        jacobian[0][ j ] *= (i == j ? cos( y ) : sin( y ));
    }
  }
};

// Algorithm
// ---------

struct Algorithm
{

  typedef Dune::Fem::LeafGridPart< HGridType > GridPartType;
  typedef Dune::Fem::FunctionSpace< double, double, HGridType::dimensionworld, 1 > SpaceType;
  typedef Dune::Fem::LagrangeDiscreteFunctionSpace< SpaceType, GridPartType, polOrder > DiscreteSpaceType;
  typedef Dune::Fem::AdaptiveDiscreteFunction< DiscreteSpaceType > DiscreteFunctionType;

  typedef Function< SpaceType > FunctionType;

  explicit Algorithm ( HGridType &grid );

private:
  GridPartType gridPart_;
  DiscreteSpaceType dfSpace_;
  FunctionType function_;
};

inline Algorithm::Algorithm ( HGridType &grid )
: gridPart_( grid ),
  dfSpace_( gridPart_ )
{
  DiscreteFunctionType solution( "solution", dfSpace_ );
  solution.clear();
  // perform Lagrange interpolation
  Dune::Fem::LagrangeInterpolation< FunctionType, DiscreteFunctionType >
    :: interpolateFunction( function_, solution );

  // Initialize the DataOutput that writes the solution on the harddisk in a
  // format readable by e.g. Paraview
  // in each loop for the eoc computation the results at
  // the final time is stored
  typedef Dune::tuple<DiscreteFunctionType*> IOTupleType;
  typedef Dune::DataOutput<HGridType, IOTupleType> DataOutputType;
  IOTupleType dataTup ( &solution );
  DataOutputType dataOutput( grid, dataTup );
  dataOutput.write();

  // compute errors
  const int order = DiscreteSpaceType::polynomialOrder+1;

  typedef Dune::GridFunctionAdapter< FunctionType, GridPartType > GridFunctionType;
  GridFunctionType gridFunction( "exact solution", function_, gridPart_, order );

  Dune::L2Norm< GridPartType > l2norm( gridPart_ );
  Dune::H1Norm< GridPartType > h1norm( gridPart_ );

  std::cout << "L2-error: " << l2norm.distance( gridFunction, solution ) << std::endl;
  std::cout << "H1-error: " << h1norm.distance( gridFunction, solution ) << std::endl;
}

// Main Program
// ------------

int main ( int argc, char **argv )
try
{
  // initialize MPI manager
  Dune::MPIManager::initialize( argc, argv );
  // add command line parameters to global parameter table
  Dune::Parameter::append( argc, argv );
  // append parameters from the parameter file
  Dune::Parameter::append( (argc < 2) ? "parameter" : argv[ 1 ] );

  std::string filename;
  Dune::Parameter::get("fem.io.macroGridFile", filename);
  int startLevel = Dune::Parameter::getValue<int>("femhowto.startLevel", 0);

  // create the grid
  Dune::GridPtr< HGridType > gridptr( filename );
  Dune::GlobalRefine::apply(*gridptr, startLevel );

  Algorithm algorithm( *gridptr );

  return 0;
}
catch( const Dune::Exception &exception )
{
  // display the exception message on the console
  std::cerr << exception << std::endl;
  return 1;
}
