// include host specific macros set by configure script   /*@LST0S@*/
#include <config.h>

// include std libs
#include <iostream>
#include <string>

// Dune includes
#include <dune/fem/misc/l2error.hh>
#include <dune/fem/operator/projection/l2projection.hh>
#include <dune/fem/gridpart/common/gridpart.hh>
#include <dune/fem/solver/odesolver.hh>

// include local header files
#include <dune/fem-howto/baseevolution.hh>
#include "models.hh"

// approximation order
const int order   = POLORDER;
const int rkSteps = POLORDER + 1;                           /*@LST0E@*/

template <class GridImp, class InitialDataType>             /*@LST1S@*/  
struct StepperTraits {
  // type of Grid
  typedef GridImp                                                    GridType;

  // Choose a suitable GridView
  typedef Dune::DGAdaptiveLeafGridPart< GridType >                         GridPartType;

  // An analytical version of our model
  typedef AdvectionDiffusionModel< GridPartType, InitialDataType >   ModelType;

  // The flux for the discretization of advection terms
  typedef UpwindFlux< ModelType >                                    FluxType;

  // The DG Operator (using 2 Passes)
  typedef Dune::DGAdvectionDiffusionOperator< ModelType, UpwindFlux,
                                              order >                      DgType; 

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef typename DgType :: DestinationType                         DiscreteFunctionType;
  // ... as well as the Space type
  typedef typename DgType :: SpaceType                               DiscreteSpaceType;

  // The ODE Solvers                                                         
  typedef DuneODE :: OdeSolverInterface< DiscreteFunctionType >      OdeSolverInterfaceType; /*@\label{dg:interfaceode}@*/
  typedef DuneODE :: ImplicitOdeSolver< DiscreteFunctionType >       ImplicitOdeSolverType;  /*@\label{dg:implode}@*/
  typedef DuneODE :: ExplicitOdeSolver< DiscreteFunctionType >       ExplicitOdeSolverType;  /*@\label{dg:explode}@*/

  // type of restriction/prolongation projection for adaptive simulations 
  typedef Dune :: RestrictProlongDefault< DiscreteFunctionType > RestrictionProlongationType;
};

template <class GridImp, class InitialDataType>               
struct Stepper : public AlgorithmBase< StepperTraits< GridImp, InitialDataType> >
{
  // my traits class 
  typedef StepperTraits< GridImp, InitialDataType> Traits ;

  // my base class 
  typedef AlgorithmBase < Traits > BaseType;

  // type of Grid
  typedef typename Traits :: GridType                  GridType;

  // Choose a suitable GridView
  typedef typename Traits :: GridPartType              GridPartType;

  // An analytical version of our model
  typedef typename Traits :: ModelType                 ModelType;

  // The flux for the discretization of advection terms
  typedef typename Traits :: FluxType                  FluxType;

  // The DG Operator (using 2 Passes)
  typedef typename Traits :: DgType                    DgType;

  // The discrete function for the unknown solution is defined in the DgOperator
  typedef typename Traits :: DiscreteFunctionType      DiscreteFunctionType;

  // ... as well as the Space type
  typedef typename Traits :: DiscreteSpaceType         DiscreteSpaceType;

  // The ODE Solvers
  typedef typename Traits :: OdeSolverInterfaceType    OdeSolverInterfaceType;
  typedef typename Traits :: ImplicitOdeSolverType     ImplicitOdeSolverType;
  typedef typename Traits :: ExplicitOdeSolverType     ExplicitOdeSolverType; 

  typedef typename BaseType :: TimeProviderType   TimeProviderType;

  using BaseType :: grid_;

  Stepper(GridType& grid, const InitialDataType& problem) :
    BaseType ( grid ),
    problem_(problem),
    model_(problem_),
    convectionFlux_(model_),
    dgOperator_(grid_, convectionFlux_),
    eocId_( Dune::FemEoc::addEntry(std::string("$L^2$-error")) ),
    odeSolver_(0)
  {
  }                                                                        

  ~Stepper()
  {
    delete odeSolver_;
    odeSolver_ = 0;
  }

  // before first step, do data initialization 
  void initializeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    // choice of explicit or implicit ode solver
    static const std::string odeSolver[]
      = { "explicit", "implicit" };
    int odeSolve = Dune::Parameter::getEnum( "femhowto.odesolver", odeSolver, 0 );

    // create implicit or explicit ode solver
    if( odeSolve == 0 )
      odeSolver_ = new ExplicitOdeSolverType(dgOperator_, tp, rkSteps);   /*@\label{dg:odeexpldef}@*/
    else
      odeSolver_ = new ImplicitOdeSolverType(dgOperator_, tp, rkSteps);   /*@\label{dg:odeimpldef}@*/

    assert( odeSolver_ );

   
    Dune::Fem::L2Projection< InitialDataType, DiscreteFunctionType > l2pro;  /*@\label{dg:l2pro0}@*/
    l2pro(problem_, u);                                                      /*@\label{dg:l2pro1}@*/

    // odeSolver_->initialize applies the DG Operator once to get an initial
    // estimate on the time step. This does not change the initial data u.
    odeSolver_->initialize(u);                               /*@\label{dg:odeinit}@*/
  }

  // solve ODE for one time step 
  void step(TimeProviderType& tp, DiscreteFunctionType& u) 
  {
    assert(odeSolver_);
    odeSolver_->solve(u);                                      /*@\label{dg:odesolve}@*/
  }

  // after last step, do EOC calculation 
  void finalizeStep(TimeProviderType& tp, DiscreteFunctionType& u)
  {
    Dune::L2Error<DiscreteFunctionType> L2err;
    // Compute L2 error of discretized solution ...
    Dune::FieldVector<double,ModelType::dimRange> error;
    error = L2err.norm(problem_, u, tp.time());
    // ... and print the statistics out to the eocOutputPath file
    Dune::FemEoc::setErrors(eocId_,error.two_norm());

    // delete ode solver
    delete odeSolver_;
    odeSolver_ = 0;
  }                                                       /*@LST1E@*/
private:
  // InitialDataType is a Dune::Operator that evaluates to $u_0$ and also has a
  // method that gives you the exact solution.
  const InitialDataType  &problem_;
  ModelType               model_;
  // Initial flux for advection discretization (UpwindFlux)
  FluxType                convectionFlux_;
  DgType                  dgOperator_;
  const unsigned int      eocId_;
  OdeSolverInterfaceType *odeSolver_;
};

/**
 * @brief main function for the LocalDG Advection-Diffusion application
 *
 * \c main starts the Simulation of an advection-diffusion pde with
 * the localdg method with EOC analysis and visual output to grape, paraview or
 * gnuplot.
 * \attention The localdg implementation uses the \c Dune::Pass
 * concept.
 *
 * @param argc number of arguments from command line
 * @param argv array of arguments from command line
 * @param envp array of environmental variables
 * @return 0 we don't program bugs. :)
 */
int main(int argc, char ** argv, char ** envp) {          /*@LST0S@*/
  typedef Dune::GridSelector :: GridType GridType;

  /* Initialize MPI (always do this even if you are not using MPI) */
  Dune::MPIManager :: initialize( argc, argv );

  try {

  // *** Initialization
  Dune::Parameter::append(argc,argv);                           /*@\label{dg:param0}@*/
  if (argc == 2) {
    Dune::Parameter::append(argv[1]);
  } 
  else 
  {
    Dune::Parameter::append("parameter");                       /*@\label{dg:paramfile}@*/
  }                                                       /*@\label{dg:param1}@*/

  // ProblemType is a Dune::Function that evaluates to $u_0$ and also has a
  // method that gives you the exact solution.
  typedef Dune::U0< GridType > ProblemType;
  ProblemType problem;

  // Note to me: problem description is for FemEOC
  Dune::GridPtr<GridType> gridptr = initialize< GridType >( std::string("Local DG scheme") + /*@\label{dg:gridinit}@*/
                                          problem.description() );
  // get grid reference 
  GridType & grid = *gridptr;

  /*********************************************
   * EOC Loop                                  *
   ********************************************/
  Stepper<GridType, ProblemType :: BaseType> stepper(grid, problem);  /*@\label{dg:stepper}@*/

  compute(stepper);                                                   /*@\label{dg:evolve}@*/

  Dune::Parameter::write("parameter.log");

  }
  catch (Dune::Exception &e) {                            /*@\label{dg:catch0}@*/
    std::cerr << e << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Generic exception!" << std::endl;
    return 2;
  }                                                      /*@\label{dg:catch1}@*/

  return 0;
}

