#ifndef DUNE_EXAMPLEDISCRETEMODELS_HH
#define DUNE_EXAMPLEDISCRETEMODELS_HH

// Dune includes
#include <dune/fem/gridpart/common/gridpart.hh>
#include <dune/fem/gridpart/adaptiveleafgridpart.hh>

// Dune-Fem includes
#include <dune/fem/space/dgspace.hh>
#include <dune/fem/pass/dgpass.hh>
#include <dune/fem/pass/dgdiscretemodel.hh>
#include <dune/fem/pass/selection.hh>
#include <dune/fem/solver/timeprovider.hh>
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/quadrature/cachingquadrature.hh>

//*************************************************************
namespace Dune {

  // forward declaration of Models
  template <class Model, class NumFlux, int polOrd,
           int passUId >
  class AdvDiffDModel1;

  template <class Model, class NumFlux, int polOrd,
           int passUId, int passGradId >
  class AdvDiffDModel2; // with advection and diffusion


  // General Traits class
  template <class Model,int dimRange,int polOrd>
  class PassTraits
  {
  public:
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridPartType                     GridPartType;
    typedef typename GridPartType :: GridType                        GridType;
    static const int dimDomain = Model :: Traits :: dimDomain;

    typedef CachingQuadrature< GridPartType, 0 >                     VolumeQuadratureType;
    typedef CachingQuadrature< GridPartType, 1 >                     FaceQuadratureType;

    typedef FunctionSpace< double, double, dimDomain, dimRange >     FunctionSpaceType;
    typedef DiscontinuousGalerkinSpace< FunctionSpaceType,
                                        GridPartType, polOrd,
                                        CachingStorage >             DiscreteFunctionSpaceType;
    typedef AdaptiveDiscreteFunction< DiscreteFunctionSpaceType >    DestinationType;
  };

  /*********************************************
   * Discrete model traits                     *
   ********************************************/
  template <class Model, class NumFlux, int polOrd, int passUId >
  struct AdvDiffTraits1
  {
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridType                         GridType;

    enum { dimRange = ModelTraits::dimGradRange };
    enum { dimDomain = ModelTraits::dimDomain };

    typedef PassTraits< Model, dimRange, polOrd >                    Traits;
    typedef typename Traits :: FunctionSpaceType                     FunctionSpaceType;

    typedef typename Traits :: VolumeQuadratureType                  VolumeQuadratureType;
    typedef typename Traits :: FaceQuadratureType                    FaceQuadratureType;
    typedef typename Traits :: GridPartType                          GridPartType;

    typedef typename Traits :: DiscreteFunctionSpaceType             DiscreteFunctionSpaceType;
    typedef typename Traits :: DestinationType                       DestinationType;
    typedef DestinationType                                          DiscreteFunctionType;

    typedef typename DestinationType :: DomainType                   DomainType;
    typedef typename ModelTraits :: FaceLocalCoordinate              FaceLocalCoordinate;
    typedef typename ModelTraits :: LocalCoordinate                  LocalCoordinate;
    typedef typename DestinationType :: RangeType                    RangeType;
    typedef typename DestinationType :: JacobianRangeType            JacobianRangeType;

    typedef AdvDiffDModel1< Model, NumFlux, polOrd, passUId >        DGDiscreteModelType;
  };

  template <class Model, class NumFlux, int polOrd, int passUId, int passGradId >
  struct AdvDiffTraits2
  {
    typedef typename Model :: Traits                                 ModelTraits;
    typedef typename ModelTraits :: GridType                         GridType;

    enum { dimRange = ModelTraits::dimRange };
    enum { dimDomain = ModelTraits::dimDomain };

    typedef PassTraits< Model, dimRange, polOrd >                    Traits;
    typedef typename Traits :: FunctionSpaceType                     FunctionSpaceType;

    typedef typename Traits :: VolumeQuadratureType                  VolumeQuadratureType;
    typedef typename Traits :: FaceQuadratureType                    FaceQuadratureType;
    typedef typename Traits :: GridPartType                          GridPartType;
    typedef typename Traits :: DiscreteFunctionSpaceType             DiscreteFunctionSpaceType;
    typedef typename Traits :: DestinationType                       DestinationType;
    typedef DestinationType                                          DiscreteFunctionType;

    typedef typename DestinationType :: DomainType                   DomainType;
    typedef typename ModelTraits :: FaceLocalCoordinate              FaceLocalCoordinate;
    typedef typename ModelTraits :: LocalCoordinate                  LocalCoordinate;
    typedef typename DestinationType :: RangeType                    RangeType;
    typedef typename DestinationType :: JacobianRangeType            JacobianRangeType;

    typedef AdvDiffDModel2< Model, NumFlux, polOrd, passUId, passGradId > DGDiscreteModelType;
  };

  /**********************************************
   * Discrete Models for LDG Passes             *
   *********************************************/
  template <class Model, class NumFlux, int polOrd, int passUId> /*@LST1S@*/
  class AdvDiffDModel1 :
    public DGDiscreteModelDefaultWithInsideOutside
      <AdvDiffTraits1<Model, NumFlux, polOrd, passUId>, passUId >
  {
    typedef DGDiscreteModelDefaultWithInsideOutside
              < AdvDiffTraits1< Model, NumFlux, polOrd, passUId >,
                passUId > BaseType;

    using BaseType :: numericalFlux;
    using BaseType :: boundaryFlux;

    // This type definition allows a convenient access to arguments of passes.
#if DUNE_VERSION_NEWER_REV(DUNE_COMMON,2,1,0)
    integral_constant<int,passUId> uVar;
#else
    Int2Type<passUId> uVar;
#endif
  public:
    typedef AdvDiffTraits1< Model, NumFlux, polOrd, passUId >        Traits; /*@LST1E@*/

    typedef typename Traits :: DomainType                            DomainType;
    typedef typename Traits :: LocalCoordinate                       LocalCoordinate;
    typedef typename Traits :: FaceLocalCoordinate                   FaceLocalCoordinate;
    typedef typename Traits :: RangeType                             RangeType;
    typedef typename Traits :: GridType                              GridType;
    typedef typename Traits :: JacobianRangeType                     JacobianRangeType;
    typedef typename Traits :: GridPartType :: IntersectionType      Intersection;
    typedef typename Traits :: GridPartType
              :: IntersectionIteratorType                            IntersectionIterator;
    typedef typename GridType :: template Codim< 0 > :: Entity       EntityType;

  public:
    /**
     * @brief constructor
     */
    AdvDiffDModel1(const Model& mod,        /*@LST1S@*/
                   const NumFlux& numf) :
      model_(mod),
      numflux_(numf),
      // Set CFL number for diffusion
      cflDiffinv_(2.*(polOrd+1.))
    {}

    bool hasSource() const { return false; }
    bool hasFlux() const { return true; }  /*@LST1E@*/

    /**
     * @brief method required by LocalDGPass
     *
     * @param it intersection
     * @param time current time given by TimeProvider
     * @param x coordinate of required evaluation local to \c it
     * @param uLeft DOF evaluation on this side of \c it
     * @param uRight DOF evaluation on the other side of \c it
     * @param gLeft result for this side of \c it
     * @param gRight result for the other side of \c it
     * @return wave speed estimate (multiplied with the integration element of the intersection).
     *         To estimate the time step |T|/wave is used
     */
    template <class ArgumentTuple>   /*@LST1S@*/
    double numericalFlux(const Intersection& it,
                         double time, const FaceLocalCoordinate& x,
                         const ArgumentTuple& uLeft,
                         const ArgumentTuple& uRight,
                         RangeType& gLeft,
                         RangeType& gRight)
    {
      const DomainType normal = it.integrationOuterNormal(x);

      JacobianRangeType diffmatrix(0.0);
      RangeType diffflux(0.);
      /* central differences (might be suboptimal) */
      model_.diffusion1(this->inside(),  /* get entity */
                        time,          /* for time dependent diffusion */
                        it.geometryInInside().global(x),
                                       /* midpoint of intersection */
                        uLeft[uVar],   /* { u_(x^-) } */
                        diffmatrix     /* return diffusion tensor */
                        );

      diffmatrix.umv(normal,diffflux);
      model_.diffusion1(this->outside(),  /* get entity */
                        time,          /* for time dependent diffusion */
                        it.geometryInOutside().global(x),
                                       /* midpoint of intersection */
                        uRight[uVar],  /* { u_(x^+) } */
                        diffmatrix     /* return diffusion tensor */
                        );
      diffmatrix.umv(normal,diffflux);
      diffflux*=0.5;

      gLeft = diffflux;
      gRight = diffflux;

      // upper bound for the next time step length
      return model_.diffusionTimeStep()*cflDiffinv_;
    }

    /**
     * @brief same as numericalFlux() but for the boundary
     */
    template <class ArgumentTuple, class FaceDom>
    double boundaryFlux(const Intersection& it,
                        double time, 
                        //const FaceLocalCoordinate& x,
                        const FaceDom& x,
                        const ArgumentTuple& uLeft,
                        RangeType& gLeft)    /*@LST1E@*/
    {
      const DomainType normal = it.integrationOuterNormal(x);

      typedef typename ArgumentTuple::template Get<passUId>::Type UType;

      JacobianRangeType diffmatrix;
      gLeft = 0.;
      if (model_.hasBoundaryValue(it,time,x))
      {
        // get the boundary value and use it to compute the flux
        UType uRight;
        // uRight == boundary Value
        model_.boundaryValue(it, time, x, uLeft[uVar], uRight);
        // diffmatrix == sqrt(eps) * uRight
        model_.diffusion1(this->inside(), time,
                          it.geometryInInside().global(x),
                          uRight, diffmatrix);
      }
      else
      {
        model_.diffusion1(this->inside(), time,
                          it.geometryInInside().global(x),
                          uLeft[uVar], diffmatrix);
      }
      diffmatrix.umv(normal,gLeft);

      return model_.diffusionTimeStep()*cflDiffinv_/2;
    }

    /**
     * @brief method required by LocalDGPass
     */
    template <class ArgumentTuple>    /*@LST1S@*/
    void analyticalFlux(const EntityType& en,
                        double time, 
                        const LocalCoordinate& x,
                        const ArgumentTuple& u, JacobianRangeType& f)
    {
      typedef typename ElementType<0, ArgumentTuple>::Type UType;
      model_.diffusion1(en, time, x, u[uVar], f);
    }                                /*@LST1E@*/

  private:
    const Model& model_;
    const NumFlux& numflux_;
    const double cflDiffinv_;
  };

  // Discrete Model for Pass2                             /*@LST0S@*/
  template <class Model, class NumFlux, int polOrd, int passUId, int passGradId>
  class AdvDiffDModel2 :
    public DGDiscreteModelDefaultWithInsideOutside
      <AdvDiffTraits2<Model, NumFlux, polOrd, passUId, passGradId >, 
            passUId , passGradId>
  {
    typedef DGDiscreteModelDefaultWithInsideOutside
              < AdvDiffTraits2< Model, NumFlux, polOrd, passUId, passGradId >,
                passUId, passGradId >                        BaseType;

    using BaseType :: numericalFlux;
    using BaseType :: boundaryFlux;

    // These type definitions allow a convenient access to arguments of pass.
#if DUNE_VERSION_NEWER_REV(DUNE_COMMON,2,1,0)
    integral_constant<int,passUId>    uVar;     /*@\label{dm:int2type0}@*/
    integral_constant<int,passGradId> sigmaVar; /*@\label{dm:int2type1}@*/
#else
    Int2Type<passUId>    uVar;     /*@\label{dm:int2type0}@*/
    Int2Type<passGradId> sigmaVar; /*@\label{dm:int2type1}@*/
#endif
  public:
    typedef AdvDiffTraits2< Model, NumFlux, polOrd, passUId,
                            passGradId >                             Traits; /*@LST0E@*/

    typedef typename Traits :: DomainType                            DomainType;
    typedef typename Traits :: LocalCoordinate                       LocalCoordinate;
    typedef typename Traits :: FaceLocalCoordinate                   FaceLocalCoordinate;

    typedef typename Traits :: RangeType                             RangeType;
    typedef typename Traits :: GridType                              GridType;
    typedef typename Traits :: JacobianRangeType                     JacobianRangeType;

    typedef typename Traits :: GridPartType :: IntersectionType      Intersection;
    typedef typename Traits :: GridPartType
              :: IntersectionIteratorType                            IntersectionIterator;
    typedef typename GridType :: template Codim< 0 > :: Entity       EntityType;

  public:
    /**
     * @brief constructor
     */
    AdvDiffDModel2(const Model& mod,const NumFlux& numf) : /*@LST0S@*/
      model_(mod),
      numflux_(numf),
      penalty_(Parameter::getValue<double>("femhowto.penalty")),
      // Set CFL number for penalty term (compare diffusion in first pass)
      cflDiffinv_(2.*(polOrd+1.))
    {}

    bool hasSource() const { return false; }  /*@\label{dm:hasSource}@*/
    bool hasFlux() const { return true; }     /*@LST0E@*/

    /**
     * @brief flux function on interfaces between cells
     *
     * @param it intersection
     * @param time current time given by TimeProvider
     * @param x coordinate of required evaluation local to \c it
     * @param uLeft DOF evaluation on this side of \c it
     * @param uRight DOF evaluation on the other side of \c it
     * @param gLeft result for this side of \c it
     * @param gRight result for the other side of \c it
     * @return wave speed estimate (multiplied with the integration element of the intersection).
     *         To estimate the time step |T|/wave is used
     */
    template <class ArgumentTuple>          /*@LST0S@*/
    double numericalFlux(const Intersection& it,
                         double time, const FaceLocalCoordinate& x,
                         const ArgumentTuple& uLeft,
                         const ArgumentTuple& uRight,
                         RangeType& gLeft,
                         RangeType& gRight)
    {
      const DomainType normal = it.integrationOuterNormal(x);


      /*****************************
       * Advection                 *
       ****************************/
      // delegated to numflux_
      double wave = numflux_.                       /*@\label{dm:upwind}@*/
        numericalFlux(it, time, x, uLeft[uVar], uRight[uVar], gLeft, gRight);

      /*****************************
       * Diffusion                 *
       ****************************/
      JacobianRangeType diffmatrix;
      RangeType diffflux(0.);
      /* Central differences */
      model_.                                      /*@\label{dm:int2typeusage}@*/
        diffusion2(this->inside(), time, it.geometryInInside().global(x),
                   uLeft[uVar], uLeft[sigmaVar], diffmatrix);
      diffmatrix.umv(normal, diffflux);
      model_.
        diffusion2(this->outside(), time, it.geometryInOutside().global(x),
                   uRight[uVar], uRight[sigmaVar], diffmatrix);
      diffmatrix.umv(normal, diffflux);
      diffflux*=0.5;

      // add penalty term ( enVolume() is available since we derive from
      //    DGDiscreteModelDefaultWithInsideOutside)
      double factor = penalty_ * model_.diffusionCoefficient() *
                      normal.two_norm2() /
                      ( 0.5 * ( this->enVolume()+this->nbVolume() ) );

      RangeType jump( uLeft[uVar] );
      jump -= uRight[uVar];
      diffflux.axpy(factor,jump);

      gLeft  += diffflux;
      gRight += diffflux;
      return std::max( wave, factor*cflDiffinv_ );
    }

    /**
     * @brief same as numericalFlux() but for fluxes over boundary interfaces
     */
    template <class ArgumentTuple>
    double boundaryFlux(const Intersection& it,
                        const double time, 
                        const FaceLocalCoordinate& x,
                        const ArgumentTuple& uLeft,
                        RangeType& gLeft)    /*@LST0E@*/
    {
      const DomainType normal = it.integrationOuterNormal(x);

      typedef typename ArgumentTuple::template Get<passUId>::Type UType;

      double wave=0.0;
      if (model_.hasBoundaryValue(it,time,x))
      {
        /*****************************
         * Advection                 *
         ****************************/
        UType     uRight;
        RangeType gRight;

        // get the boundary value for upwind discretization
        model_.boundaryValue(it, time, x, uLeft[uVar], uRight);
        wave = numflux_.numericalFlux(it, time, x, uLeft[uVar], uRight, gLeft, gRight);

        /****************************/
        /* Diffusion                 *
         ****************************/
        JacobianRangeType diffmatrix;
        model_.diffusion2(this->inside(), time,
                          it.geometryInInside().global(x),
                          uLeft[uVar], uLeft[sigmaVar], diffmatrix);
        diffmatrix.umv(normal, gLeft);

        // add penalty term
        double factor = penalty_ * model_.diffusionCoefficient() *
                        normal.two_norm2() / this->enVolume();
        RangeType jump( uLeft[uVar] );
        jump -= uRight;
        gLeft.axpy(factor,jump);
        wave = std::max( wave, factor*cflDiffinv_ );
      }
      else
      {
        // not implemented
        assert(false);
      }
      return wave;
    }
                                                  /*@LST0S@*/
    /**
     * @brief analytical flux function $f_2(u_0,u_1)$
     */
    template <class ArgumentTuple> 
    void analyticalFlux( const EntityType& en,
                         const double time, 
                         const LocalCoordinate& x,
                         const ArgumentTuple& u, JacobianRangeType& f )
    {
      /*****************************
       * Advection                 *
       ****************************/
      model_.advection(en,time, x, u[uVar],f);
      /*****************************
       * Diffusion                 *
       ****************************/
      JacobianRangeType diffmatrix;
      model_.diffusion2(en, time, x, u[uVar], u[sigmaVar], diffmatrix);
      f += diffmatrix;
    }
  private:
    const Model   &model_;
    const NumFlux &numflux_;
    const double   penalty_;
    const double   cflDiffinv_;
  };                                              /*@LST0E@*/

}
#endif
