#ifndef POISSON_SERIAL_HH
#define POISSON_SERIAL_HH

//***********************************************************************
/*! Stokes problem: 

  This is an example solving the poisson problem
  \f{displaymath}
  \begin{array}{rcll}
  -\triangle u + \nabla p &=& f & \quad \mbox{in}\ \Omega\\
                 \nabla u &=& 0 & \quad \mbox{in}\ \Omega
                        u &=& g & \quad \mbox{on}\ \partial\Omega
  \end{array}
  \f}
  with the finite element method using Lagrangian elements. The polynomial
  order is given by POLORDER.
  
  In this example, $\Omega = ]-1,1[^{dimworld}$ and
  \f[
  f( x, y, z ) = 0 
  \f]

  The exact solution to the stokes problem is then given by
  \f[
  u( x, y, z ) = \prod_{i=1}^{dimworld} sin( 2 \pi x_i)\\
  p( x, y, z ) = 2.*exp(x[0]) * sin(x[1]).
  \f]
*/
//***********************************************************************

// this define enables the use of higher (>2) order Lagrange basis functions
#define USE_TWISTFREE_MAPPER

// System Includes
// ---------------
#include <iostream>
#include <sstream>


// DUNE Core Includes
// ------------------
#include <dune/common/version.hh>

// if this value is not defined, then we have version 1.1.1
#ifndef DUNE_VERSION_HH
#define OLD_DUNE_GRID_VERSION
#endif

#include <dune/common/stdstreams.hh>
#include <dune/common/timer.hh>



// DUNE-FEM includes
// -----------------

// grid parts 
#include <dune/fem/gridpart/common/gridpart.hh>

// adaptation classes 
#include <dune/fem/space/common/adaptmanager.hh>
// lagrange space 
#include <dune/fem/space/lagrangespace.hh>

// discrete functions 
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/function/blockvectorfunction.hh>
#include <dune/fem/function/vectorfunction.hh>

#include <dune/fem/operator/discreteoperatorimp.hh>

// matrix implementations 
#include <dune/fem/operator/matrix/spmatrix.hh>
#include <dune/fem/operator/matrix/blockmatrix.hh>
#include <dune/fem/operator/matrix/ontheflymatrix.hh>
#include <dune/fem/operator/matrix/istlmatrix.hh>

// linear solvers 
#include <dune/fem/solver/inverseoperators.hh>
#include <dune/fem/solver/oemsolver/oemsolver.hh>
#include <dune/fem/solver/istlsolver.hh>
#include "uzawa.hh"

// l2 norm and h1 norm 
#include <dune/fem/misc/l2norm.hh>
#include <dune/fem/misc/h1norm.hh>

// parameter reading facilities 
#include <dune/fem/io/parameter.hh>

// header for data output for grape and VTK 
#include <dune/fem/io/file/datawriter.hh>

#include <dune/fem/misc/femeoc.hh>

// base.hh for compute function
//#include  <dune/fem-howto/base.hh>

#include "combinedop.hh"

// Local Includes
// --------------
// #define USE_COMBINED_FCT

// matrix assembler 
#include "laplaceoperator.hh"
#include "stokesoperator.hh"


using namespace Dune;

// Algorithm
// --------- 
template <class GridImp, int velocityPolynomialOrder, 
          int pressurePolynomialOrder,
          class GridPartImp = LeafGridPart<GridImp> >
class Algorithm                                                       /*@LST0S@*/
{
public:                                                               /*@LST0E@*/
  typedef GridImp GridType;

  /** choose the grid partition (and hence the index set) to use
   *
   *  \note Not all index sets are continuous. The LeafIndexSet for AlbertaGrid,
   *        for example, is not. If you want to use OEM solvers, the index set
   *        must be continuous. In such a case use AdaptiveLeafGridPart.
   */
  //---- GridParts -----------------------------------------------------------
  typedef GridPartImp  GridPartType;
                                                                    
  //---- FunctionSpace -------------------------------------------------------
  //! define the function space, \f[ \R^n \rightarrow \R \f]
  // see dune/common/functionspace.hh
  typedef FunctionSpace< double, double, GridType::dimensionworld, GridType :: dimension > VelocityFunctionSpaceType;
  typedef FunctionSpace< double, double, GridType::dimensionworld, 1 > PressureFunctionSpaceType;

  // set dimension of the Range for the Velocity field
  enum{ dimRange = GridType :: dimension };

  //! to be revised 
  typedef ProblemInterface< VelocityFunctionSpaceType > ProblemType;

  // The data functions (as defined in problem.hh)
  //---- Right Hand Side, Exact Solution, and Stiffness tensor ---------------
  typedef typename ProblemType ::  ExactSolutionType   ExactVelocitySolutionType;

  //---- Adapter for exact solution ------------------------------------------
  typedef GridFunctionAdapter< ExactVelocitySolutionType, GridPartType >
    GridExactVelocitySolutionType;


  //---- DiscreteFunctionSpace -----------------------------------------------
  //! define the discrete function space our unkown belongs to
#ifdef USE_COMBINED_FCT
  typedef LagrangeDiscreteFunctionSpace
    < PressureFunctionSpaceType, GridPartType, velocityPolynomialOrder, CachingStorage >
    SingleVelocitySpaceType;
  //typedef AdaptiveDiscreteFunction< SingleVelocitySpaceType > SingleVelocityFunctionType;
  
  typedef BlockVectorDiscreteFunction< SingleVelocitySpaceType > SingleVelocityFunctionType;
  typedef CombinedDiscreteFunction< SingleVelocityFunctionType, GridType::dimensionworld > DiscreteVelocityFunctionType;

  typedef typename DiscreteVelocityFunctionType :: DiscreteFunctionSpaceType
    DiscreteVelocitySpaceType;

  typedef Fem::SparseRowMatrixTraits < SingleVelocitySpaceType, SingleVelocitySpaceType > 
      MatrixObjectTraits;
  //typedef ISTLMatrixTraits < SingleVelocitySpaceType, SingleVelocitySpaceType > 
  //    MatrixObjectTraits;

  typedef LaplaceFEOp< SingleVelocityFunctionType, MatrixObjectTraits >
    SingleLaplaceOperatorType;                                            

  //typedef CombinedOp< SingleLaplaceOperatorType, GridType::dimensionworld >
  //  LaplaceOperatorType;
  typedef SingleLaplaceOperatorType  LaplaceOperatorType;

  typedef CGInverseOperator< SingleVelocityFunctionType > InverseSingleLaplaceOperatorType;
  //typedef ISTLCGOp< SingleVelocityFunctionType, SingleLaplaceOperatorType > InverseSingleLaplaceOperatorType;
  //typedef ISTLBICGSTABOp< SingleVelocityFunctionType, SingleLaplaceOperatorType > InverseSingleLaplaceOperatorType;

  typedef CombinedOp< InverseSingleLaplaceOperatorType, GridType::dimensionworld >
    InverseLaplaceOperatorType;
#else 
  typedef LagrangeDiscreteFunctionSpace
    < VelocityFunctionSpaceType, GridPartType, velocityPolynomialOrder, CachingStorage >
    DiscreteVelocitySpaceType;
  typedef AdaptiveDiscreteFunction< DiscreteVelocitySpaceType > DiscreteVelocityFunctionType;
  //typedef BlockVectorDiscreteFunction< DiscreteVelocitySpaceType > DiscreteVelocityFunctionType;

  typedef Fem::SparseRowMatrixTraits < DiscreteVelocitySpaceType, DiscreteVelocitySpaceType > 
      MatrixObjectTraits;
  //typedef ISTLMatrixTraits < DiscreteVelocitySpaceType, DiscreteVelocitySpaceType > 
  //    MatrixObjectTraits;

  typedef LaplaceFEOp< DiscreteVelocityFunctionType, MatrixObjectTraits >
    LaplaceOperatorType;                                            

  typedef CGInverseOperator< DiscreteVelocityFunctionType > InverseLaplaceOperatorType;
  //typedef ISTLCGOp< DiscreteVelocityFunctionType, LaplaceOperatorType > InverseLaplaceOperatorType;
#endif

  typedef LagrangeDiscreteFunctionSpace
    < PressureFunctionSpaceType, GridPartType, pressurePolynomialOrder, CachingStorage >
    DiscretePressureSpaceType;
    

  //---- DiscreteFunction ----------------------------------------------------
  //---- good choice for adaptive simulations using OEM solver ---------------
  //! define the type of discrete function we are using


  typedef AdaptiveDiscreteFunction< DiscretePressureSpaceType > DiscretePressureFunctionType;

  //----- Forward definitions 
  //----- in order to keap the structures of base.hh
  typedef DiscreteVelocitySpaceType    DiscreteSpaceType;
  typedef DiscreteVelocityFunctionType DiscreteFunctionType;

  //---- other possible choices, use BlockVectorDiscreteFunction for ISTL ----
  //typedef BlockVectorDiscreteFunction< DiscreteSpaceType > DiscreteFunctionType;
  //typedef ManagedDiscreteFunction< VectorDiscreteFunction< DiscreteSpaceType, DynamicVector< double > > > DiscreteFunctionType;

  //---- MatrixObjects -------------------------------------------------------
  //---- good choice for build in solvers ------------------------------------
  //! define the type of the system matrix object

  typedef Fem::SparseRowMatrixTraits < DiscreteVelocitySpaceType, DiscretePressureSpaceType >
      DivergenceMatrixObjectTraits;

  //---- other choices, ISTLMatrixTraits for BCRSMatrix from DUNE-ISTL -------
  //typedef ISTLMatrixTraits < DiscreteSpaceType, DiscreteSpaceType > MatrixObjectTraits;
  //typedef OnTheFlyMatrixTraits < DiscreteSpaceType, DiscreteSpaceType > MatrixObjectTraits;
      
  //! define the discrete laplace operator, see ./laplaceoperator.hh 
  //typedef LaplaceFEOp< DiscreteVelocityFunctionType, MatrixObjectTraits >
  //  LaplaceOperatorType;                                            

  typedef DivergenceFEOp< DiscreteVelocityFunctionType, DiscretePressureFunctionType, 
          DivergenceMatrixObjectTraits> 
            DivergenceOperatorType;

  //---- InverseOperator ----------------------------------------------------
  //---- good choice for build in CG solver ---------------------------------
  //! define the inverse operator we are using to solve the system 
  typedef UzawaInverseOp< DiscreteVelocityFunctionType, 
                          DiscretePressureFunctionType, 
                          InverseLaplaceOperatorType, 
                          DivergenceOperatorType >
     InverseOperatorType;

    //---- other choices ------------------------------------------------------ 
    //typedef ISTLBICGSTABOp< DiscreteFunctionType, LaplaceOperatorType >
    //typedef ISTLCGOp< DiscreteFunctionType, LaplaceOperatorType >
    //typedef OEMCGOp<DiscreteFunctionType,LaplaceOperatorType>
    //typedef OEMBICGSTABOp<DiscreteFunctionType,LaplaceOperatorType>
    //typedef OEMBICGSQOp<DiscreteFunctionType,LaplaceOperatorType>
    //typedef OEMGMRESOp<DiscreteFunctionType,LaplaceOperatorType>
    //    InverseOperatorType;


public:
  //! constructor                                                  /*@LST0S@*/
  Algorithm(GridType & grid, const ProblemType& problem)
  : grid_( grid ),
    gridPart_( grid_ ),
    velocitySpace_( gridPart_ ),            /*@\label{stokes:setvelocitySpace}@*/
    pressureSpace_( gridPart_ ),            /*@\label{stokes:setPressureSpace}@*/
    problem_( problem )
  {
    // add entries to eoc calculation 
    std::vector<std::string> femEocHeaders;
    // we want to calculate L2 and H1 error (and EOC)
    femEocHeaders.push_back("$L^2$-error");
    femEocHeaders.push_back("$H^1$-error");                        /*@LST0E@*/

    // get eoc id 
    eocId_ = FemEoc::addEntry(femEocHeaders);

    // distribute grid (only for parallel runs, in serial nothing is done here)
    grid.loadBalance();
  }                                                                /*@LST0S@*/

  //! setup and solve the linear system
  void operator()(DiscreteVelocityFunctionType & solution)
  {                                                                /*@LST0E@*/

    DiscretePressureFunctionType pressure("pressure", pressureSpace_);

    // type of grid iterator 
    typedef typename DiscreteVelocitySpaceType :: IteratorType IteratorType; 
    // type of entity 
    typedef typename IteratorType :: Entity EntityType;

    // in verbose mode some info 
    if( Parameter :: verbose () ) 
    {
      std::cout << std :: endl << "Solving for " << velocitySpace_.size()
        << " unkowns and polynomial order "
        << DiscreteVelocitySpaceType :: polynomialOrder << "." 
        << std :: endl << std :: endl;
    }

    // create adapter (for visualization with grape)
    GridExactVelocitySolutionType ugrid( "exact velocity", problem_.exactSolution(), gridPart_,
                                          DiscreteVelocitySpaceType :: polynomialOrder + 1 );

    // initialize solution with zero                                 /*@LST0S@*/
    solution.clear();

    // create laplace assembler (is assembled on call of systemMatrix by solver)
    LaplaceOperatorType laplace( velocitySpace_, problem_ );               /*@\label{sto:laplaceDef}@*/

    // create divergence assambler 
    DivergenceOperatorType div( velocitySpace_, pressureSpace_);            /*@\label{sto:divergenceDef}@*/

    // create right hand side
    DiscreteVelocityFunctionType rhs( "rhs", velocitySpace_ );                      /*@\label{sto:rhsInit0}@*/ 

    // create pressure right hand side
    DiscretePressureFunctionType g("rhs for pressure", pressureSpace_);         /*@\label{sto:rhsInit1}@*/
         
    // initialize as zero 
    g.clear();

    // setup right hand side 
    const int quadratureOrder = 2 * velocitySpace_.order() + 1;
    RightHandSideAssembler< DiscreteVelocityFunctionType >
      ::assemble ( problem_ , rhs , quadratureOrder ); 

    // set Dirichlet Boundary to exact solution  
    const IteratorType endit = velocitySpace_.end();
    for( IteratorType it = velocitySpace_.begin(); it != endit; ++it )
    {
      const EntityType &entity = *it;
      // in entity has intersections with the boundary adjust dirichlet nodes 
      if( entity.hasBoundaryIntersections() )
        boundaryTreatment( entity, problem_, rhs, solution );              /*@\label{sto:getBndFunction}@*/
    }                                                              

    // adjust discrete pressure right hand side to fix the divergence free condition on
    // boundary, first apply divergence to bnd data in solution
    div(solution, pressure);            /*@\label{sto:getbndinfluence}@*/

    // apply the changes in the divergence matrix to the right hande side
    g -= pressure;                    /*@\label{sto:correctrhs}@*/
 
    pressure.clear();

    // set correct bnd vals to divergence operator
    div.boundaryTreatment();            /*@\label{sto:correctdiv}@*/


    // solve the linear system
    solve( laplace, div, rhs, g, solution, pressure );                                /*@\label{sto:solve}@*/ 
  }

  //! finalize computation by calculating errors and EOCs 
  void finalize(DiscreteVelocityFunctionType & solution)
  {
    // create exact solution 
    ExactVelocitySolutionType uexact( problem_ ); 

    // create grid function adapter 
    GridExactVelocitySolutionType ugrid( "exact velocity", uexact, gridPart_,
                                          DiscreteVelocitySpaceType :: polynomialOrder + 1 );

    // create L2 - Norm 
    L2Norm< GridPartType > l2norm( gridPart_ );
    // calculate L2 - Norm 
    const double l2errorVelo = l2norm.distance( ugrid, solution );      /*@LST0E@*/

    // create H1 - Norm 
    H1Norm< GridPartType > h1norm( gridPart_ );
    // calculate H1 - Norm 
    const double h1errorVelo = h1norm.distance( ugrid, solution );


    // store values 
    std::vector<double> errors;
    errors.push_back( l2errorVelo );
    errors.push_back( h1errorVelo );

    // submit error to the FEM EOC calculator 
    FemEoc :: setErrors(eocId_, errors);
  }                                                        /*@LST0S@*//*@LST0E@*/

  //! return reference to discrete space 
  DiscreteVelocitySpaceType & space() {
    return velocitySpace_;
  }


private:                                                            /*@LST0S@*/
  //! set the dirichlet points to exact values
  template< class EntityType, class DiscreteFunctionType >
  void boundaryTreatment( const EntityType &entity,
                          const ProblemType& problem,  
                          DiscreteFunctionType &rhs,
                          DiscreteFunctionType &solution)
  {                                                                 /*@LST0E@*/
    typedef typename DiscreteFunctionType :: DiscreteFunctionSpaceType
      DiscreteSpaceType;
    typedef typename DiscreteFunctionType :: LocalFunctionType LocalFunctionType;

    typedef typename DiscreteSpaceType :: LagrangePointSetType
      LagrangePointSetType;
    typedef typename DiscreteSpaceType :: GridPartType GridPartType;

    const int faceCodim = 1;
    typedef typename GridPartType :: IntersectionIteratorType
      IntersectionIteratorType;
    typedef typename LagrangePointSetType
      :: template Codim< faceCodim > :: SubEntityIteratorType
      FaceDofIteratorType;

    const DiscreteSpaceType &dfSpace = solution.space();
    const GridPartType &gridPart = dfSpace.gridPart();

    // get local functions of data 
    LocalFunctionType rhsLocal = rhs.localFunction( entity );
    LocalFunctionType solutionLocal = solution.localFunction( entity );

    typedef typename EntityType :: Geometry Geometry; 
    const Geometry& geo = entity.geometry();

    const LagrangePointSetType &lagrangePointSet = dfSpace.lagrangePointSet( entity );
    
    IntersectionIteratorType it = gridPart.ibegin( entity );
    const IntersectionIteratorType endit = gridPart.iend( entity );
    for( ; it != endit; ++it )
    {
      typedef typename IntersectionIteratorType :: Intersection IntersectionType;
      const IntersectionType& intersection = *it;

      // if intersection is with boundary, adjust data  
      if( !intersection.boundary() )
        continue;

      //typedef typename IntersectionType :: Geometry GlobalGeometryType;
      //const GlobalGeometryType& interGeometry = intersection.intersectionGlobal();

      // get face number of boundary intersection 
      const int face = intersection.indexInInside();

      // get dof iterators 
      FaceDofIteratorType faceIt
        = lagrangePointSet.template beginSubEntity< faceCodim >( face );
      const FaceDofIteratorType faceEndIt
        = lagrangePointSet.template endSubEntity< faceCodim >( face );
      for( ; faceIt != faceEndIt; ++faceIt )
      {
        typedef typename VelocityFunctionSpaceType :: DomainType DomainType;
        //get local and global Lagrange points
        const DomainType &local = lagrangePointSet.point( *faceIt );
        const DomainType global = geo.global( local );
       
        for(int i=0; (i<dimRange); ++i)
        {
          // get local dof number 
          const int localDof = (*faceIt) * dimRange  + i ;

          // check whether Dirichlet boundary or not 
          if( ! problem.dirichletBoundary( intersection.boundaryId(), global ) )
          {
            continue;
          }
          
          // evaluate boundary data                                     /*@LST0S@*/
          typedef typename VelocityFunctionSpaceType :: RangeType RangeType;
          RangeType phi, phiLocal;
          problem.g( global, phi );

          const double bndValue = phi[ i ];       /*@\label{sto:projectbnddata}@*/

          // adjust right hand side and solution data 
          rhsLocal[ localDof ] = bndValue;                           /*@\label{sto:assgnval}@*/   
          solutionLocal[ localDof ] = bndValue;                        
                                                                     /*@LST0E@*/
        }          
      }
    }
  }                                                                 /*@LST0S@*/

  //! solve the resulting linear system
  void solve ( LaplaceOperatorType &laplace,
               DivergenceOperatorType &div,
               const DiscreteVelocityFunctionType &rhs,
               const DiscretePressureFunctionType &g,
               DiscreteVelocityFunctionType &velocity,
               DiscretePressureFunctionType &pressure)
  { 
    /*@LST0E@*/
    // create timer (also stops time) 
    Timer timer;

    // solve the linear system (with CG)
    const double dummy = 12345.67890;
    double solverEps = 1e-8 ;
    solverEps = Parameter :: getValue( "femhowto.solvereps", solverEps );

    int maxIterations = 10000;
    maxIterations = Parameter :: getValue ("fem.solver.maxIterations", maxIterations);

    // get verbosity information from Parameter class 
    const bool verbose = Parameter :: verbose();

    // create inverse laplace operator 
    InverseLaplaceOperatorType inverseLaplace( laplace, solverEps, solverEps );

    // create inverse operator                                     /*@LST0S@*/
    InverseOperatorType uzawa( inverseLaplace, div, dummy, solverEps, maxIterations, verbose);       /*@\label{sto:uzawaInit}@*/
  
    // solve the system 
    uzawa( rhs, g, velocity, pressure );                                            /*@\label{sto:uzawaExec}@*//*@LST0E@*/

    // get elapsed time 
    const double solveTime = timer.elapsed();

    // output in only in verbose mode (parameter fem.verbose)
    if( verbose ) 
      std :: cout << "Time needed by solver: " << solveTime << "s" << std :: endl;
  }                                                                /*@LST0S@*/


protected:
  GridType& grid_;
  GridPartType gridPart_;
  DiscreteVelocitySpaceType velocitySpace_;
  DiscretePressureSpaceType pressureSpace_;
  const ProblemType& problem_;
  int eocId_;
};                                                                /*@LST0E@*/

#endif // POISSONSERIAL_HH
