#ifndef DUNE_COMBINEDOPERATOR_HH
#define DUNE_COMBINEDOPERATOR_HH

#include <limits>

#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/io/parameter.hh>
#include <dune/fem/function/combinedfunction.hh>

#include <dune/fem/operator/2order/lagrangematrixtraits.hh>

namespace Dune
{

  template <class Op, int N>
  struct CombinedOpTraits
  {
    typedef typename Op::DestinationType ContainedDestinationType ;
    // use combined function as destination type 
    typedef CombinedDiscreteFunction< ContainedDestinationType, N >  DestinationType ;
    typedef typename DestinationType :: RangeFieldType   RangeFieldType ;
    typedef typename DestinationType :: DomainFieldType  DomainFieldType ;
  };




  template< class Op, int N >
  class CombinedOp  
  : public Operator< typename CombinedOpTraits<Op, N>::DomainFieldType, 
                     typename CombinedOpTraits<Op, N>::RangeFieldType,
                     typename CombinedOpTraits<Op, N>::DestinationType, 
                     typename CombinedOpTraits<Op, N>::DestinationType > 
  {
    typedef CombinedOpTraits<Op, N>  Traits ;

  public:
    typedef typename Traits :: DestinationType  DiscreteFunctionType;
    typedef typename Traits :: ContainedDestinationType ContainedDestinationType;
    typedef typename DiscreteFunctionType :: DiscreteFunctionSpaceType
      DiscreteFunctionSpaceType;
    typedef Op OperatorType;

    /** \brief constructor of UMFPACKOp
        \param[in] op Operator to invert 
        \param[in] redEps realative tolerance for residual (not used here)
        \param[in] absLimit absolut solving tolerance for residual (not used here)
        \param[in] maxIter maximal number of iterations performed (not used here)
        \param[in] verbose verbosity 
    */
    template <class OPType>
    CombinedOp( const OPType &op, const double eps = 1e-8, const double absEps = 1e-8 ) 
    : op_( op, eps, absEps ),
      numbers_( 2, double(0.0) )
    {
    }

    template <class DFSpaceType>
    CombinedOp( const DFSpaceType &spc ) 
    : op_( spc.containedSpace() ),
      numbers_( 2, double(0.0) )
    {
    }

    //CombinedOp( const DiscreteFunctionSpaceType &spc ) 
    //: op_( spc.containedSpace() ),
    //  numbers_( 2, double(0.0) )
    //{
    //}

    void prepare ( const DiscreteFunctionType &, DiscreteFunctionType & ) const
    {}

    void finalize () const
    {}

    /** \brief solve the system 
        \param[in] arg right hand side 
        \param[out] dest solution 
    */
    void apply ( const DiscreteFunctionType &rhs, DiscreteFunctionType &solution) const
    {
      prepare( rhs, solution );

      numbers_[ 0 ] = 0;
      numbers_[ 1 ] = 0;

      // apply operator for each sub function 
      for( int i=0; i<N; ++i ) 
      {
        // apply internal operator 
        op_( rhs.subFunction( i ), solution.subFunction( i ) );

        // get numbers 
        //numbers_[ 0 ] += op_.averageCommTime();
        //numbers_[ 1 ] += op_.iterations();
      }

      // finalize operator  
      finalize ();
    }

    /** \brief solve the system 
        \param[in] arg right hand side 
        \param[out] dest solution 
    */
    void operator() ( const DiscreteFunctionType &arg, DiscreteFunctionType &dest ) const
    {
      apply( arg, dest );
    }

    void printTexInfo(std::ostream& out) const
    {
      out << "Solver: Combined Solver ";
      out  << "\\\\ \n";
    }

    double averageCommTime() const 
    {
      if( numbers_.size() > 0 )
        return numbers_[ 0 ];
      else 
        return 0.0;
    }

    int iterations() const 
    {
      if( numbers_.size() > 1 )
        return numbers_[ 1 ];
      else 
        return 0;
    }

  protected:
    OperatorType op_;
    mutable std::vector< double > numbers_; 
  };

}

#endif // #ifndef DUNE_UMFPACKSOLVER_HH
