#ifndef DUNE_LAPLACE_HH
#define DUNE_LAPLACE_HH

//- Dune includes
#include <dune/common/fmatrix.hh>
#include <dune/common/timer.hh>

#include <dune/fem/storage/array.hh>
#include <dune/fem/quadrature/quadrature.hh>
#include <dune/fem/function/common/scalarproducts.hh>
#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/operator/2order/lagrangematrixsetup.hh>

#include "dune/fem-howto/probleminterfaces.hh"

namespace Dune
{

  //! \brief The Laplace operator
  template< class DiscreteFunction, class MatrixTraits >             
  class LaplaceFEOp
  : public Operator< typename DiscreteFunction :: RangeFieldType,
                     typename DiscreteFunction :: RangeFieldType,
                     DiscreteFunction,
                     DiscreteFunction >,
    public OEMSolver::PreconditionInterface                         
  {                                                                
    // needs to be friend for conversion check 
    friend class Conversion<ThisType,OEMSolver::PreconditionInterface>;
    
  public:
    //! type of discrete functions
    typedef DiscreteFunction DiscreteFunctionType;

    //! type of this LaplaceFEOp
    typedef LaplaceFEOp< DiscreteFunctionType, MatrixTraits > LaplaceFEOpType;

  private:
    typedef LaplaceFEOpType ThisType;

  public:
    //! type of discrete function space
    typedef typename DiscreteFunctionType :: DiscreteFunctionSpaceType
      DiscreteFunctionSpaceType;

    //! type of problem 
    typedef ProblemInterface< typename DiscreteFunctionSpaceType :: FunctionSpaceType > 
        ProblemType; 

    //! field type of range
    typedef typename DiscreteFunctionSpaceType :: RangeFieldType
      RangeFieldType;
       
  protected:
    //! type of jacobian
    typedef typename DiscreteFunctionSpaceType :: JacobianRangeType
      JacobianRangeType;
    //! type of the base function set
    typedef typename DiscreteFunctionSpaceType :: BaseFunctionSetType
      BaseFunctionSetType;

  public:
    //! type of grid partition
    typedef typename DiscreteFunctionSpaceType :: GridPartType GridPartType;
    //! type of grid
    typedef typename DiscreteFunctionSpaceType :: GridType GridType;

    //! polynomial order of base functions
    enum { polynomialOrder = DiscreteFunctionSpaceType :: polynomialOrder };

    //! The grid's dimension
    enum { dimension = GridType :: dimension };

    //! The dimension of the Rangefield
    enum{ dimRange = DiscreteFunctionSpaceType :: FunctionSpaceType :: dimRange };
        
    //! type of quadrature to be used
    typedef CachingQuadrature< GridPartType, 0 > QuadratureType;

    typedef typename MatrixTraits
      :: template  MatrixObject< LagrangeMatrixTraits< MatrixTraits > >
      :: MatrixObjectType LinearOperatorType;

    //! get important types from the MatrixObject 
    typedef typename LinearOperatorType :: LocalMatrixType LocalMatrixType;
    typedef typename LinearOperatorType :: PreconditionMatrixType PreconditionMatrixType;
    typedef typename LinearOperatorType :: MatrixType MatrixType;

  public:
    // types for boundary treatment
    // ----------------------------
    typedef typename DiscreteFunctionSpaceType :: MapperType MapperType;
    typedef SlaveDofs< DiscreteFunctionSpaceType, MapperType > SlaveDofsType;
    typedef typename SlaveDofsType :: SingletonKey SlaveDofsKeyType;    
    typedef SingletonList< SlaveDofsKeyType, SlaveDofsType >
      SlaveDofsProviderType;
    
  protected:
    // type of DofManager
    typedef DofManager< GridType > DofManagerType;

   
  public:
    //! constructor 
    explicit LaplaceFEOp( const DiscreteFunctionSpaceType &dfSpace,
                          const ProblemType& problem)
    : discreteFunctionSpace_( dfSpace ),
      dofManager_( DofManagerType :: instance( dfSpace.grid() ) ),
      linearOperator_( discreteFunctionSpace_, discreteFunctionSpace_ ),
      sequence_( -1 ),
      problem_( problem ),
      slaveDofs_( getSlaveDofs( discreteFunctionSpace_ ) ),
      gradCache_( discreteFunctionSpace_.mapper().maxNumDofs() ),
      gradDiffusion_( discreteFunctionSpace_.mapper().maxNumDofs() )
    {
    }
        
  private:
    // prohibit copying
    LaplaceFEOp ( const ThisType & );

  public:                                                           
    //! apply the operator
    virtual void operator() ( const DiscreteFunctionType &u, 
                              DiscreteFunctionType &w ) const 
    {
      systemMatrix().apply( u, w );                                
    }                                                               

    //! apply umfpack to solve linear System
    virtual void solve ( const DiscreteFunctionType &b,
                         DiscreteFunctionType &x ) const
    {
      systemMatrix().solveUMF( b, x);
    }
  
    //! return reference to preconditioning matrix, used by OEM-Solver
    const PreconditionMatrixType &preconditionMatrix () const
    {
      return systemMatrix().preconditionMatrix();
    }

    //! return true if preconditioning is enabled
    bool hasPreconditionMatrix () const
    {
      return linearOperator_.hasPreconditionMatrix();
    }

    //! print the system matrix into a stream
    void print ( std :: ostream & out = std :: cout ) const 
    {
      systemMatrix().matrix().print( out );
    }

    //! return reference to discreteFunctionSpace
    const DiscreteFunctionSpaceType &discreteFunctionSpace () const
    {
      return discreteFunctionSpace_;
    }

    //! return reference to problem
    const ProblemType& problem() const
    {
      return problem_;
    } 

    /*! \brief obtain a reference to the system matrix
     *
     *  The assembled matrix is returned. If the system matrix has not been
     *  assembled, yet, the assembly is performed.
     *
     *  \returns a reference to the system matrix
     */
    LinearOperatorType &systemMatrix () const                        
    {
      // if stored sequence number it not equal to the one of the
      // dofManager (or space) then the grid has been changed
      // and matrix has to be assembled new
      if( sequence_ != dofManager_.sequence() )                     
        assemble();

      return linearOperator_;
    }

    /** \brief perform a grid walkthrough and assemble the global matrix */
    void assemble () const 
    {
      const DiscreteFunctionSpaceType &space = discreteFunctionSpace();

      // reserve memory for matrix 
      linearOperator_.reserve();                                   

      // create timer (also stops time)
      Timer timer;

      // clear matrix                                              
      linearOperator_.clear();

      // apply local matrix assembler on each element
      typedef typename DiscreteFunctionSpaceType :: IteratorType IteratorType;
      IteratorType end = space.end();
      for(IteratorType it = space.begin(); it != end; ++it)
      {
        assembleLocal( *it );                                       
      }

      // correct matrix due to boundary conditions 
      boundaryCorrectOnGrid();              
      
      // get elapsed time 
      const double assemblyTime = timer.elapsed();
      // in verbose mode print times 
      if ( Parameter :: verbose () )
        std :: cout << "Time to assemble laplace matrix: " << assemblyTime << "s" << std :: endl;

      // get grid sequence number from space (for adaptive runs)    
      sequence_ = dofManager_.sequence();
    }

  protected:
    //! assemble local matrix for given entity             /*@LST0S@*/
    template< class EntityType >
    void assembleLocal( const EntityType &entity ) const   
    {                                                       /*@LST0E@*/        
      typedef typename EntityType :: Geometry GeometryType;

      // assert that matrix is not build on ghost elements 
      assert( entity.partitionType() != GhostEntity );

      // cache geometry of entity 
      const GeometryType geometry = entity.geometry();

      // get local matrix from matrix object
      LocalMatrixType localMatrix                                   
        = linearOperator_.localMatrix( entity, entity );
      
      // get base function set 
      const BaseFunctionSetType &baseSet = localMatrix.domainBaseFunctionSet();

      // get number of local base functions 
      const size_t numBaseFunctions = baseSet.size();
            
      // create quadrature of appropriate order 
      QuadratureType quadrature( entity, 2 * (polynomialOrder - 1) );

      // loop over all quadrature points
      const size_t numQuadraturePoints = quadrature.nop();
      for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
      {
        // get local coordinate of quadrature point 
        const typename QuadratureType :: CoordinateType &x
          = quadrature.point( pt );

        // get jacobian inverse transposed 
        const FieldMatrix< double, dimension, dimension > &inv
          = geometry.jacobianInverseTransposed( x );

        typedef typename ProblemType :: DiffusionMatrixType DiffusionMatrixType;
        DiffusionMatrixType K;
        // evaluate diffusion matrix 
        problem().K( geometry.global( x ), K );

        // for all base functions evaluate the gradient       /*@LST0S@*/
        // on quadrature point pt and apply jacobian inverse 
        baseSet.jacobianAll( quadrature[ pt ], inv, gradCache_ );

        // apply diffusion tensor  
        for( size_t i = 0; i < numBaseFunctions; ++i )
        {
          for(int k=0;k<dimRange;++k)
            K.mv( gradCache_[ i ][ k ], gradDiffusion_[ i ][ k ] );
        }                                   /*@LST0E@*/       

        // evaluate integration weight 
        weight_ = quadrature.weight( pt ) * geometry.integrationElement( x );
        
        // add scalar product of gradients to local matrix 
        updateLocalMatrix( localMatrix );
      }
    }                           /*@LST0S@*//*@LST0E@*/    
  
    //! add scalar product of cached gradients to local matrix
    void updateLocalMatrix ( LocalMatrixType &localMatrix ) const       /*@LST0S@*/
    {
      const size_t rows    = localMatrix.rows();
      const size_t columns = localMatrix.columns();
      
      for( size_t i = 0; i < rows; ++i )
      {
        for ( size_t j = 0; j < columns; ++j )
        {
          RangeFieldType value =0.;
          for(int k=0;k<dimRange;++k)                                  /*@\label{sto:laplacemulitI}@*/
            value += (gradCache_[ i ][ k ] * gradDiffusion_[ j ][ k ]);          
          
          value *= weight_;
          localMatrix.add( i, j, value );
        }
      }
    }                                                                 /*@LST0E@*/
  
    //! adjust matrix due to boundary conditions
    void boundaryCorrectOnGrid () const
    {                                                                
      typedef typename DiscreteFunctionSpaceType :: IteratorType IteratorType;
      typedef typename IteratorType :: Entity EntityType;

      const DiscreteFunctionSpaceType &dfSpace = discreteFunctionSpace();

      const IteratorType end = dfSpace.end();
      for( IteratorType it = dfSpace.begin(); it != end; ++it )
      {
        const EntityType &entity = *it;
        // if entity has boundary intersections 
        if( entity.hasBoundaryIntersections() )
          boundaryCorrectOnEntity( entity );
      }
    }                                                      

    /*! treatment of Dirichlet-DoFs for one entity
     *
     *   delete rows for dirichlet-DoFs, setting diagonal element to 1.
     *
     *   \note A LagrangeDiscreteFunctionSpace is implicitly assumed.
     *
     *   \param[in]  entity  entity to perform Dirichlet treatment on
     */
    template< class EntityType >
    void boundaryCorrectOnEntity ( const EntityType &entity ) const 
    {                                                               
      typedef typename DiscreteFunctionSpaceType :: LagrangePointSetType
        LagrangePointSetType;

      const int faceCodim = 1;
      typedef typename GridPartType :: IntersectionIteratorType
        IntersectionIteratorType;
      typedef typename LagrangePointSetType
        :: template Codim< faceCodim > :: SubEntityIteratorType
        FaceDofIteratorType;

      const DiscreteFunctionSpaceType &dfSpace = discreteFunctionSpace();
      const GridPartType &gridPart = dfSpace.gridPart();

      const LagrangePointSetType &lagrangePointSet                  
        = dfSpace.lagrangePointSet( entity );

      SlaveDofsType &slaveDofs = this->slaveDofs();
 
      // get local matrix from matrix object 
      LocalMatrixType localMatrix = linearOperator_.localMatrix( entity, entity );

      // loop over all intersections and find dirichlet 
      // boundaries 
      const IntersectionIteratorType endit = gridPart.iend( entity );
      for( IntersectionIteratorType it = gridPart.ibegin( entity );
           it != endit ; ++it )
      {
        // get intersection 
        typedef typename IntersectionIteratorType :: Intersection IntersectionType;
        const IntersectionType& intersection = *it;
    
        // skip non-boundary elements
        if( ! intersection.boundary() )                                       
          continue;
       
        // get local face number of boundary intersection 
        const int face = intersection.indexInInside();

        // get iterator over all local dofs on this face 
        FaceDofIteratorType faceIt
          = lagrangePointSet.template beginSubEntity< faceCodim >( face );
        const FaceDofIteratorType faceEndIt
          = lagrangePointSet.template endSubEntity< faceCodim >( face );

        // iterate over face dofs and set unit row 
        for( ; faceIt != faceEndIt; ++faceIt )
        {
          for(int i=0;i<dimRange;++i)
          {
            // write 1 to the diagonal if dof is no slaveDof
            const int localDof = (*faceIt) * dimRange + i;          
            localMatrix.clearRow( localDof );

            // get global dof to check wether its a slave dof
            const int globalDof = dfSpace.mapper().mapToGlobal( entity, localDof );         

            if( !slaveDofs.isSlave( globalDof ) )
              localMatrix.set( localDof, localDof, 1.);
          }
        }
      }
    }                                                                 

    // return slave dofs                          
    static SlaveDofsType *getSlaveDofs ( const DiscreteFunctionSpaceType &space )
    {
      SlaveDofsKeyType key( space, space.mapper() );
      return &(SlaveDofsProviderType :: getObject( key ));
    }

    // return reference to slave dofs 
    SlaveDofsType &slaveDofs () const
    {
      slaveDofs_->rebuild();
      return *slaveDofs_;
    }                                                    

  protected:
    const DiscreteFunctionSpaceType &discreteFunctionSpace_;
    const DofManagerType &dofManager_;

    //! pointer to the system matrix
    mutable LinearOperatorType linearOperator_;
 
    //! flag indicating whether the system matrix has been assembled
    mutable int sequence_;
      
    //! the diffusion tensor 
    const ProblemType& problem_;

    //! pointer to slave dofs 
    SlaveDofsType *const slaveDofs_;

    mutable Fem :: DynamicArray< JacobianRangeType > gradCache_;
    mutable Fem :: DynamicArray< JacobianRangeType > gradDiffusion_;
    mutable RangeFieldType weight_;    
  };                                                                  



  // Assembler for right hand side
  // note: this is not an L2-Projection
  template< class DiscreteFunction >
  class RightHandSideAssembler
  {
  public:
    typedef DiscreteFunction DiscreteFunctionType;
    
    typedef typename DiscreteFunctionType :: DiscreteFunctionSpaceType
      DiscreteFunctionSpaceType;
    typedef typename DiscreteFunctionType :: LocalFunctionType
      LocalFunctionType;
  
    typedef typename DiscreteFunctionSpaceType :: BaseFunctionSetType
      BaseFunctionSetType;
    typedef typename DiscreteFunctionSpaceType :: RangeType RangeType;
    typedef typename DiscreteFunctionSpaceType :: GridPartType GridPartType;
    typedef typename GridPartType :: GridType GridType;

    enum { dimension = GridType :: dimension };

  public:
    //! discreteFunction is an output parameter (kind of return value)
    template< class FunctionType >
    static void assemble( const FunctionType &function,
                          DiscreteFunctionType &discreteFunction,
                          const int polOrd)
    {
      typedef typename DiscreteFunctionSpaceType :: IteratorType IteratorType;
      typedef typename IteratorType :: Entity EntityType;
      typedef typename EntityType :: Geometry GeometryType;

      // We use a caching quadrature for codimension 0 entities
      typedef CachingQuadrature< GridPartType, 0 > QuadratureType;
      
      const DiscreteFunctionSpaceType &discreteFunctionSpace
        = discreteFunction.space();
  
      // set discreteFunction to zero
      discreteFunction.clear(); 

      // start grid traversal 
      const IteratorType endit = discreteFunctionSpace.end();
      for( IteratorType it = discreteFunctionSpace.begin(); 
           it != endit; ++it )
      {
        // *it gives a reference to the current entity
        const EntityType &entity = *it;

        // obtain the entity's geometry
        const GeometryType geometry = entity.geometry();
      
        // obtain BaseFunctionSet for the entity
        // note that base functions are always defined on the reference geometry
        LocalFunctionType localFunction = discreteFunction.localFunction( entity ); 
        const BaseFunctionSetType baseFunctionSet
          = discreteFunctionSpace.baseFunctionSet( entity ); 

        QuadratureType quadrature( entity, polOrd );
        const size_t numQuadraturePoints = quadrature.nop();
        for( size_t qP = 0; qP < numQuadraturePoints; ++qP )
        {
          // get integration element multiplied with quadrature weight
          const double factor = 
            geometry.integrationElement( quadrature.point( qP ) ) * 
            quadrature.weight( qP );

          RangeType phi;
          // evaluate right hand side function
          function.evaluate( geometry.global( quadrature.point( qP ) ), phi );

          // multiply with integration element
          phi *= factor;

          // 
          localFunction.axpy( quadrature[ qP ], phi ); 
        }
      }

      // communicate data (for parallel runs)
      discreteFunction.communicate();
    }
  };

} // end namespace 
#endif
