/**************************************************************************
**       Title: main.cc
** Description: File demonstrating a simple numerics problem on arbitrary
**              grids: stokes-problem with known solution is given
**              and compared with numerical solution (EOC)
**              To simulate other problems without know solutions, the 
**              caraman vortex street and the driven cavity problem are added.
**              The carman vortex street wroks only in 2D with GRIDTYPE=ALUGRID_CONFORM
**              Dune grid parser is used.
**
**              For changing grid-types, compile with
**
**              make clean
**
**              and one of the following
**           
**              make                         or
**              make GRIDTYPE=ALUGRID_CONFORM       (default)
**                    -> compiles and works correctly 
**              make GRIDTYPE=ALBERTAGRID
**                    -> compiles and works correctly
**              make GRIDTYPE=SGRID
**                    -> compiles and works correctly
**              make GRIDTYPE=YASPGRID
**                    -> compiles and works correctly
**                       - only problem driven_cavity can run on YASPGRID
**
**
**              Similarly, the polynomial order can be specified by
**
**              make POLORDER=2
**              
**              here we need polynomial order higher then 2, for the stability of
**              nuemrical scheme.
**              ! polynomial order higher than 2 needs a twisted free grid !
**              This not jet a parallel runnig programm, this feature will be comming
**              soon.
**
**************************************************************************/
#include <config.h>

// definition of algorithm
#include "algorithm.hh"

// definition of problem
#include "problemdata.hh"

// mpi manager calling MPI_Init (addon to mpihelper from DUNE)
#include <dune/fem/misc/mpimanager.hh>

#ifndef POLORDER 
const int velocityPolynomialOrder = 2;
const int preasurePolynomialOrder = 1;
#else 
const int velocityPolynomialOrder = POLORDER;
const int preasurePolynomialOrder = velocityPolynomialOrder - 1;
#endif

#include <dune/fem-howto/base.hh>

using namespace Dune;


// Main Program
// ------------
int main( int argc, char **argv )                                   /*@LST0S@*/
{
  typedef GridSelector::GridType GridType;

  typedef ProblemInterface<FunctionSpace< double, double, GridType::dimension, GridType::dimension > > ProblemType;

  // initialize MPI 
  MPIManager :: initialize ( argc, argv );
  const int rank = MPIManager :: rank ();
  
  try
  {
    // append parameters from the comand line 
    Parameter::append( argc, argv );

    // append parameters from the parameter file  
    Parameter::append( (argc < 2) ? "parameter" : argv[ 1 ] );

    // generate GridPointer holding grid instance
    GridPtr< GridType > gridptr = initialize< GridType >(std::string("Stokes problem"));/*@\label{sto:gridinit}@*/

    // get grid reference 
    GridType& grid = *gridptr ;
      
    // create problem 
    ProblemType* problem = createProblem<GridType> ();              /*@\label{sto:probleminit}@*/
    assert( problem );

    // create stepper class 
    Algorithm<GridType, velocityPolynomialOrder, preasurePolynomialOrder > algorithm(grid, *problem);  /*@\label{sto:stepper}@*/
    // compute solution
    compute(algorithm);                                       /*@\label{sto:compute}@*/

    // write parameter logfile 
    Parameter :: write("parameter.log");

    // remove problem 
    delete problem;

    return 0;
  }
  catch( Exception &exception )                                     /*@\label{sto:catch0}@*/
  {
    if( rank == 0 )
      std :: cerr << exception << std :: endl;
    return 1;
  }                                                                 /*@\label{sto:catch1}@*/
}                                                                   /*@LST0E@*/
