#ifndef DUNE_PROBLEM_HH
#define DUNE_PROBLEM_HH

#include <cassert>
#include <cmath>

#include <dune/fem-howto/probleminterfaces.hh>

namespace Dune
{
  //! an exact problem
  template <class GridImp>
  class ExactProblem 
    : public ProblemInterface<FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > >
  {
  public:
    typedef FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > FunctionSpaceType ;
    typedef ProblemInterface<FunctionSpaceType> BaseType;

    enum { dimDomain = FunctionSpaceType :: dimDomain };

    typedef typename BaseType :: DomainType DomainType;
    typedef typename BaseType :: RangeType  RangeType;
    typedef typename BaseType :: JacobianRangeType  JacobianRangeType;
    typedef typename BaseType :: DomainFieldType DomainFieldType;
    typedef typename BaseType :: RangeFieldType  RangeFieldType;

    typedef typename BaseType :: DiffusionMatrixType DiffusionMatrixType;

  public:
    ExactProblem () {}

    //! the right hand side 
    void f(const DomainType& x, RangeType& ret) const 
    {
      ret =0;
    }

    //! the exact solution 
    void u(const DomainType& x, RangeType& ret) const 
    {
      ret[0] = -exp(x[0]) * (x[1] * cos(x[1]) +  sin(x[1]));
      ret[1] = exp(x[0]) * x[1] * sin(x[1]);
    }

    //! the diffusion matrix 
    void K(const DomainType& x, DiffusionMatrixType& m) const 
    {
      m = 0;
      for( int i=0; i<dimDomain; ++i) m[ i ][ i ] = 1; 
    }

    bool constantK() const { return true; }

    //! the gradient of the exact solution 
    virtual void gradient(const DomainType& x,
                          JacobianRangeType& grad) const 
    {
      grad[0][0] = -exp(x[0]) *(x[1] * cos(x[1]) +  sin(x[1])) ;
      grad[0][1] = -exp(x[0]) *( cos(x[1]) + cos(x[1]) - x[1]*sin(x[1]));
      grad[1][0] = exp(x[0]) * x[1] * sin(x[1]);
      grad[1][1] = exp(x[0]) * (x[1] * cos(x[1]) +  sin(x[1]));
    }
  };

  //! state of the art test problem driven cavity
  template <class GridImp>
  class DrivenCavityProblem 
    : public ProblemInterface<FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > >
  {
  public:
    typedef FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > FunctionSpaceType ;
    typedef ProblemInterface<FunctionSpaceType> BaseType;

    enum { dimDomain = FunctionSpaceType :: dimDomain };

    typedef typename BaseType :: DomainType DomainType;
    typedef typename BaseType :: RangeType  RangeType;
    typedef typename BaseType :: JacobianRangeType  JacobianRangeType;
    typedef typename BaseType :: DomainFieldType DomainFieldType;
    typedef typename BaseType :: RangeFieldType  RangeFieldType;

    typedef typename BaseType :: DiffusionMatrixType DiffusionMatrixType;

  public:
    DrivenCavityProblem () {}

    //! the right hand side 
    void f(const DomainType& x, RangeType& ret) const 
    {
      ret =0;
    }

    //! the exact solution 
    void u(const DomainType& x, RangeType& ret) const 
    {
      ret =0;
    }

    //! since there is no exact solution we need boundary values
    void g(const DomainType& x, RangeType&ret) const
    {
      ret =0;
      if( x[1]  == 1 )
        ret[0] = 1.;
    }

    //! the diffusion matrix 
    void K(const DomainType& x, DiffusionMatrixType& m) const 
    {
      m=0;
      for( int i=0; i<dimDomain; ++i) m[ i ][ i ] = 1; 
    }

    bool constantK() const { return true; }

    //! the gradient of the exact solution 
    virtual void gradient(const DomainType& x,
                          JacobianRangeType& grad) const 
    {
      grad =0;
    }
  };

  //! test problem for the Carman vortex street
  template <class GridImp>
  class CarmanProblem 
    : public ProblemInterface<FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > >
  {
  public:
    typedef FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > FunctionSpaceType ;
    typedef ProblemInterface<FunctionSpaceType> BaseType;

    enum { dimDomain = FunctionSpaceType :: dimDomain };

    typedef typename BaseType :: DomainType DomainType;
    typedef typename BaseType :: RangeType  RangeType;
    typedef typename BaseType :: JacobianRangeType  JacobianRangeType;
    typedef typename BaseType :: DomainFieldType DomainFieldType;
    typedef typename BaseType :: RangeFieldType  RangeFieldType;

    typedef typename BaseType :: DiffusionMatrixType DiffusionMatrixType;

  public:
    CarmanProblem () {}

    //! the right hand side 
    void f(const DomainType& x, RangeType& ret) const 
    {
      ret = 0;
    }

    //! the exact solution 
    void u(const DomainType& x, RangeType& ret) const 
    {
      ret = 0;
    }

    void g(const DomainType& x, RangeType& ret) const
    {
      ret =0;
      if(std::fabs(x[0]-1.1 ) == 1.1 )
        ret[0] = std::pow(0.41,-2.)* ( 6.* x[1]*( 0.41 - x[1]) );
    }

    //! the diffusion matrix 
    void K(const DomainType& x, DiffusionMatrixType& m) const 
    {
      m=0;
      for( int i=0; i<dimDomain; ++i) m[ i ][ i ] = 1; 
    }

    bool constantK() const { return true; }

    //! the gradient of the exact solution 
    virtual void gradient(const DomainType& x,
                          JacobianRangeType& grad) const 
    {
      grad = 0;
    }
  };


  template <class GridImp> 
  static ProblemInterface<FunctionSpace< double, double, GridImp :: dimension, GridImp :: dimension > >* 
  createProblem()
  {
    int problem = 0; // default value 
    const std::string problemTable[] = { "exact", "driven_cavity", "carman" };
    // read problem from parameter file 
    problem =  Parameter::getEnum( "femhowto.stokesproblem", problemTable, problem );

    // see problem.hh for available problems 
    switch (problem)
    {
      case 0: return new ExactProblem< GridImp > ();
      case 1: return new DrivenCavityProblem< GridImp > ();
      case 2: return new CarmanProblem< GridImp > ();
      default: std::cerr << "Wrong problem value, bye, bye!" << std::endl;
               abort();
    }
    return 0;
  }

}
#endif
