#ifndef DUNE_DIVERGENCE_HH
#define DUNE_DIVERGENCE_HH

//- Dune includes
#include <dune/common/fmatrix.hh>
#include <dune/common/timer.hh>

#include <dune/fem/storage/array.hh>
#include <dune/fem/quadrature/quadrature.hh>
#include <dune/fem/operator/common/operator.hh>
#include <dune/fem/operator/2order/lagrangematrixsetup.hh>


namespace Dune
{

  //! \brief The Laplace operator
  template< class DiscreteVelocityFunction, class DiscretePressureFunction, class MatrixTraits >             /*@LST0S@*/
  class DivergenceFEOp 
  : public Operator< typename DiscreteVelocityFunction :: RangeFieldType,
                     typename DiscretePressureFunction :: RangeFieldType,
                     DiscreteVelocityFunction,
                     DiscretePressureFunction >,
    public OEMSolver::PreconditionInterface                         
  {                                                                 /*@LST0E@*/
    // needs to be friend for conversion check 
    friend class Conversion<ThisType,OEMSolver::PreconditionInterface>;
    
  public:
    //! type of discrete functions
    typedef DiscreteVelocityFunction DiscreteVelocityFunctionType;
    typedef DiscretePressureFunction DiscretePressureFunctionType;

    //! type of this LaplaceFEOp
    typedef DivergenceFEOp< DiscreteVelocityFunctionType, DiscretePressureFunctionType, MatrixTraits > DivergenceFEOpType;

  private:
    typedef DivergenceFEOpType ThisType;

  public:
    //! type of discrete function space
    typedef typename DiscreteVelocityFunctionType :: DiscreteFunctionSpaceType
      DiscreteVelocityFunctionSpaceType;
    typedef typename DiscretePressureFunctionType :: DiscreteFunctionSpaceType
      DiscretePressureFunctionSpaceType;
    
    //! field type of range
    typedef typename DiscreteVelocityFunctionSpaceType :: RangeFieldType
      RangeFieldType;
       
    //! type of grid partition
    typedef typename DiscreteVelocityFunctionSpaceType :: GridPartType GridPartType;
    //! type of grid
    typedef typename DiscreteVelocityFunctionSpaceType :: GridType GridType;

  protected:
    //! type of jacobian
    typedef typename DiscreteVelocityFunctionSpaceType :: JacobianRangeType
      JacobianRangeType;
    //! type of the base function set
    typedef typename DiscreteVelocityFunctionSpaceType :: BaseFunctionSetType
      VelocityBaseFunctionSetType;

    typedef typename DiscretePressureFunctionSpaceType :: BaseFunctionSetType
      PressureBaseFunctionSetType;
    typedef typename DiscretePressureFunctionSpaceType :: RangeType PressureRangeType;

    enum{dimRange = GridType::dimension};

  public:
    //! polynomial order of base functions
    enum { velocityPolynomialOrder = DiscreteVelocityFunctionSpaceType :: polynomialOrder,
           pressurePolynomialOrder = DiscretePressureFunctionSpaceType :: polynomialOrder };

    //! The grid's dimension
    enum { dimension = GridType :: dimension };
        
    //! type of quadrature to be used
    typedef CachingQuadrature< GridPartType, 0 > QuadratureType;

    typedef typename MatrixTraits
      :: template  MatrixObject< LagrangeMatrixTraits< MatrixTraits > >
      :: MatrixObjectType LinearOperatorType;

    //! get important types from the MatrixObject 
    typedef typename LinearOperatorType :: LocalMatrixType LocalMatrixType;
    typedef typename LinearOperatorType :: PreconditionMatrixType PreconditionMatrixType;
    typedef typename LinearOperatorType :: MatrixType MatrixType;

  public:
    // types for boundary treatment
    // ----------------------------
    typedef typename DiscreteVelocityFunctionSpaceType :: MapperType VelocityMapperType;
    typedef typename DiscretePressureFunctionSpaceType :: MapperType PressureMapperType;
  protected:
    // type of DofManager
    typedef DofManager< GridType > DofManagerType;

   
  public:
    //! constructor 
    explicit DivergenceFEOp( const DiscreteVelocityFunctionSpaceType &velocitySpace,
                             const DiscretePressureFunctionSpaceType &pressureSpace)
    : discreteVelocityFunctionSpace_( velocitySpace ),
      discretePressureFunctionSpace_( pressureSpace ),
      dofManager_( DofManagerType :: instance( velocitySpace.grid() ) ),
      linearOperator_( discreteVelocityFunctionSpace_, discretePressureFunctionSpace_ ),
      sequence_( -1 ),
      gradCache_( discreteVelocityFunctionSpace_.mapper().maxNumDofs() ),
      values_( discretePressureFunctionSpace_.mapper().maxNumDofs() )
    {
    }
        
  private:
    // prohibit copying
    DivergenceFEOp ( const ThisType & );

  public:                                                           /*@LST0S@*/
    //! apply the operator
    virtual void operator() ( const DiscreteVelocityFunctionType &u, 
                              DiscretePressureFunctionType &w ) const 
    {
      systemMatrix().apply( u, w );                                 /*@\label{sto:matrixEval}@*/
    }                                                               /*@LST0E@*/
  
    //! return reference to preconditioning matrix, used by OEM-Solver
    const PreconditionMatrixType &preconditionMatrix () const
    {
      return systemMatrix().preconditionMatrix();
    }
                                                                    /*@LST0S@*/
    virtual void applyTransposed ( const DiscretePressureFunctionType &u,
                                   DiscreteVelocityFunctionType &w) const 
    {
      systemMatrix().apply_t(u,w);                /*@\label{sto:applytransposed}@*/
    }                                                             /*@LST0E@*/

    //! return true if preconditioning is enabled
    bool hasPreconditionMatrix () const
    {
      return linearOperator_.hasPreconditionMatrix();
    }

    //! print the system matrix into a stream
    void print ( std :: ostream & out = std :: cout ) const 
    {
      systemMatrix().matrix().print( out );
    }

    //! return reference to discreteFunctionSpace
    const DiscreteVelocityFunctionSpaceType &discreteFunctionSpace () const
    {
      return discreteVelocityFunctionSpace_;
    }

    /*! \brief obtain a reference to the system matrix
     *
     *  The assembled matrix is returned. If the system matrix has not been
     *  assembled, yet, the assembly is performed.
     *
     *  \returns a reference to the system matrix
     */
    LinearOperatorType &systemMatrix () const                        /*@LST0S@*/
    {
      // if stored sequence number it not equal to the one of the
      // dofManager (or space) then the grid has been changed
      // and matrix has to be assembled new
      if( sequence_ != dofManager_.sequence() )                     /*@\label{sto:sequence}@*/
        assemble();

      return linearOperator_;
    }    

    /** \brief perform a grid walkthrough and assemble the global matrix */
    void assemble ()  const
    {
      const DiscreteVelocityFunctionSpaceType &velocitySpace = discreteFunctionSpace();

      // reserve memory for matrix 
      linearOperator_.reserve();                                   /*@LST0E@*/

      // create timer (also stops time)
      Timer timer;

      // clear matrix                                             /*@LST0S@*/ 
      linearOperator_.clear();

      // apply local matrix assembler on each element
      typedef typename DiscreteVelocityFunctionSpaceType :: IteratorType IteratorType;
      IteratorType end = velocitySpace.end();
      for(IteratorType it = velocitySpace.begin(); it != end; ++it)
      {
        assembleLocal( *it );                                       /*@\label{sto:applAss}@*/
      }

      // get elapsed time 
      const double assemblyTime = timer.elapsed();
      // in verbose mode print times 
      if ( Parameter :: verbose () )
        std :: cout << "Time to assemble div matrix: " << assemblyTime << "s" << std :: endl;

      // get grid sequence number from space (for adaptive runs)    /*@LST0S@*/
      sequence_ = dofManager_.sequence();
    }

    // make boundry treament
    void boundaryTreatment () const
    {
      typedef typename DiscreteVelocityFunctionSpaceType :: IteratorType IteratorType;
      typedef typename IteratorType :: Entity EntityType;

      const DiscreteVelocityFunctionSpaceType &dfSpace = discreteFunctionSpace();

      const IteratorType end = dfSpace.end();
      for( IteratorType it = dfSpace.begin(); it != end; ++it )
      {
        const EntityType &entity = *it;
        // if entity has boundary intersections 
        if( entity.hasBoundaryIntersections() )
          boundaryCorrectOnEntity( entity );
      }
    }

  protected:
    //! assemble local matrix for given entity
    template< class EntityType >
    void assembleLocal( const EntityType &entity ) const 
    {
      typedef typename EntityType :: Geometry GeometryType;

      // assert that matrix is not build on ghost elements 
      assert( entity.partitionType() != GhostEntity );

      // cache geometry of entity 
      const GeometryType geometry = entity.geometry();

      // get local matrix from matrix object
      LocalMatrixType localMatrix                                   /*@\label{sto:localMInit}@*/
        = linearOperator_.localMatrix( entity, entity );
      
      // get base function set 
      const VelocityBaseFunctionSetType &velocityBaseSet = localMatrix.domainBaseFunctionSet();/*@\label{sto:baseSetInit}@*/
      const PressureBaseFunctionSetType &pressureBaseSet = localMatrix.rangeBaseFunctionSet(); /*@\label{sto:baseSetInitI}@*/

      // create quadrature of appropriate order 
      QuadratureType quadrature( entity, 2 * (velocityPolynomialOrder - 1) );/*@\label{sto:quadraInit}@*/

      // loop over all quadrature points
      const size_t numQuadraturePoints = quadrature.nop();
      for( size_t pt = 0; pt < numQuadraturePoints; ++pt )
      {
        // get local coordinate of quadrature point 
        const typename QuadratureType :: CoordinateType &x
          = quadrature.point( pt );

        // get jacobian inverse transposed 
        const FieldMatrix< double, dimension, dimension > &inv
          = geometry.jacobianInverseTransposed( x );


        // for all base functions evaluate the gradient
        // on quadrature point pt and apply jacobian inverse 
        velocityBaseSet.jacobianAll( quadrature[ pt ], inv, gradCache_ );

        // evaluate each base function
        pressureBaseSet.evaluateAll( quadrature[ pt ], values_ );          /*@\label{sto:cacheII}@*/

        // evaluate integration weight 
        weight_ = quadrature.weight( pt ) * geometry.integrationElement( x );
        
        // add scalar product of gradients to local matrix 
        updateLocalMatrix( localMatrix );
      }
    }
  
    //! add scalar product of cached gradients to local matrix
    void updateLocalMatrix ( LocalMatrixType &localMatrix ) const
    {
      const size_t rows    = localMatrix.rows();
      const size_t columns = localMatrix.columns();
      for( size_t i = 0; i < rows; ++i )
      {
        for ( size_t j = 0; j < columns; ++j )
        {
          RangeFieldType value =0.;
          for(size_t k = 0;k<dimRange;++k)
            value += gradCache_[ j ][ k ][ k ];
          value *= -1.* weight_ *values_[ i ];
          localMatrix.add( i, j, value );           /*@\label{sto:write}@*/
        }
      }
    }                                                         /*@LST0E@*/

    template< class EntityType >
    void boundaryCorrectOnEntity ( const EntityType &entity ) const 
    {                                                               
      typedef typename DiscreteVelocityFunctionSpaceType :: LagrangePointSetType
        LagrangePointSetType;

      const int faceCodim = 1;
      typedef typename GridPartType :: IntersectionIteratorType
        IntersectionIteratorType;
      typedef typename LagrangePointSetType
        :: template Codim< faceCodim > :: SubEntityIteratorType
        FaceDofIteratorType;

      const DiscreteVelocityFunctionSpaceType &dfSpace = discreteFunctionSpace();
      const GridPartType &gridPart = dfSpace.gridPart();

      const LagrangePointSetType &lagrangePointSet                  
        = dfSpace.lagrangePointSet( entity );

      // get local matrix from matrix object 
      LocalMatrixType localMatrix = linearOperator_.localMatrix( entity, entity );

      // loop over all intersections and find dirichlet 
      // boundaries 
      const IntersectionIteratorType endit = gridPart.iend( entity );
      for( IntersectionIteratorType it = gridPart.ibegin( entity );
           it != endit ; ++it )
      {
        // get intersection 
        typedef typename IntersectionIteratorType :: Intersection IntersectionType;
        const IntersectionType& intersection = *it;

        // skip non-boundary elements
        if( ! intersection.boundary() )                                      
          continue;

        // get local face number of boundary intersection 
        const int face = intersection.indexInInside();

        // get iterator over all local dofs on this face 
        FaceDofIteratorType faceIt
          = lagrangePointSet.template beginSubEntity< faceCodim >( face );
        const FaceDofIteratorType faceEndIt
          = lagrangePointSet.template endSubEntity< faceCodim >( face );

        // iterate over face dofs and set unit row 
        for( ; faceIt != faceEndIt; ++faceIt )
        {
          for(int i =0;i<dimRange;++i)
          {
            const int localDof = (*faceIt) * dimRange + i;
            localMatrix.clearCol( localDof );
          }
        }
      }
    }


  protected:
    const DiscreteVelocityFunctionSpaceType &discreteVelocityFunctionSpace_;
    const DiscretePressureFunctionSpaceType &discretePressureFunctionSpace_;
    const DofManagerType &dofManager_;

    //! pointer to the system matrix
    mutable LinearOperatorType linearOperator_;
 
    //! flag indicating whether the system matrix has been assembled
    mutable int sequence_;
      
    mutable Fem :: DynamicArray< JacobianRangeType > gradCache_;
    mutable Fem :: DynamicArray< PressureRangeType > values_;
    mutable RangeFieldType weight_;    
  };                                                                  /*@LST0S@*//*@LST0E@*/

} // end namespace 
#endif
